// SPDX-License-Identifier: MIT
// Copyright (C) 2018-present iced project and contributors

using System;

namespace Generator.Enums {
	[Enum("Code", Documentation = "x86 instruction code", Public = true)]
	enum Code {
		[Comment("It's an invalid instruction, eg. it's a new unknown instruction, garbage or there's not enough bytes to decode the instruction etc.")]
		INVALID,
		[Comment("A #(c:db)#/#(c:.byte)# asm directive that can store 1-16 bytes")]
		DeclareByte,
		[Comment("A #(c:dw)#/#(c:.word)# asm directive that can store 1-8 words")]
		DeclareWord,
		[Comment("A #(c:dd)#/#(c:.int)# asm directive that can store 1-4 dwords")]
		DeclareDword,
		[Comment("A #(c:dq)#/#(c:.quad)# asm directive that can store 1-2 qwords")]
		DeclareQword,
		Add_rm8_r8,
		Add_rm16_r16,
		Add_rm32_r32,
		Add_rm64_r64,
		Add_r8_rm8,
		Add_r16_rm16,
		Add_r32_rm32,
		Add_r64_rm64,
		Add_AL_imm8,
		Add_AX_imm16,
		Add_EAX_imm32,
		Add_RAX_imm32,
		Pushw_ES,
		Pushd_ES,
		Popw_ES,
		Popd_ES,
		Or_rm8_r8,
		Or_rm16_r16,
		Or_rm32_r32,
		Or_rm64_r64,
		Or_r8_rm8,
		Or_r16_rm16,
		Or_r32_rm32,
		Or_r64_rm64,
		Or_AL_imm8,
		Or_AX_imm16,
		Or_EAX_imm32,
		Or_RAX_imm32,
		Pushw_CS,
		Pushd_CS,
		Popw_CS,
		Adc_rm8_r8,
		Adc_rm16_r16,
		Adc_rm32_r32,
		Adc_rm64_r64,
		Adc_r8_rm8,
		Adc_r16_rm16,
		Adc_r32_rm32,
		Adc_r64_rm64,
		Adc_AL_imm8,
		Adc_AX_imm16,
		Adc_EAX_imm32,
		Adc_RAX_imm32,
		Pushw_SS,
		Pushd_SS,
		Popw_SS,
		Popd_SS,
		Sbb_rm8_r8,
		Sbb_rm16_r16,
		Sbb_rm32_r32,
		Sbb_rm64_r64,
		Sbb_r8_rm8,
		Sbb_r16_rm16,
		Sbb_r32_rm32,
		Sbb_r64_rm64,
		Sbb_AL_imm8,
		Sbb_AX_imm16,
		Sbb_EAX_imm32,
		Sbb_RAX_imm32,
		Pushw_DS,
		Pushd_DS,
		Popw_DS,
		Popd_DS,
		And_rm8_r8,
		And_rm16_r16,
		And_rm32_r32,
		And_rm64_r64,
		And_r8_rm8,
		And_r16_rm16,
		And_r32_rm32,
		And_r64_rm64,
		And_AL_imm8,
		And_AX_imm16,
		And_EAX_imm32,
		And_RAX_imm32,
		Daa,
		Sub_rm8_r8,
		Sub_rm16_r16,
		Sub_rm32_r32,
		Sub_rm64_r64,
		Sub_r8_rm8,
		Sub_r16_rm16,
		Sub_r32_rm32,
		Sub_r64_rm64,
		Sub_AL_imm8,
		Sub_AX_imm16,
		Sub_EAX_imm32,
		Sub_RAX_imm32,
		Das,
		Xor_rm8_r8,
		Xor_rm16_r16,
		Xor_rm32_r32,
		Xor_rm64_r64,
		Xor_r8_rm8,
		Xor_r16_rm16,
		Xor_r32_rm32,
		Xor_r64_rm64,
		Xor_AL_imm8,
		Xor_AX_imm16,
		Xor_EAX_imm32,
		Xor_RAX_imm32,
		Aaa,
		Cmp_rm8_r8,
		Cmp_rm16_r16,
		Cmp_rm32_r32,
		Cmp_rm64_r64,
		Cmp_r8_rm8,
		Cmp_r16_rm16,
		Cmp_r32_rm32,
		Cmp_r64_rm64,
		Cmp_AL_imm8,
		Cmp_AX_imm16,
		Cmp_EAX_imm32,
		Cmp_RAX_imm32,
		Aas,
		Inc_r16,
		Inc_r32,
		Dec_r16,
		Dec_r32,
		Push_r16,
		Push_r32,
		Push_r64,
		Pop_r16,
		Pop_r32,
		Pop_r64,
		Pushaw,
		Pushad,
		Popaw,
		Popad,
		Bound_r16_m1616,
		Bound_r32_m3232,
		Arpl_rm16_r16,
		Arpl_r32m16_r32,
		Movsxd_r16_rm16,
		Movsxd_r32_rm32,
		Movsxd_r64_rm32,
		Push_imm16,
		Pushd_imm32,
		Pushq_imm32,
		Imul_r16_rm16_imm16,
		Imul_r32_rm32_imm32,
		Imul_r64_rm64_imm32,
		Pushw_imm8,
		Pushd_imm8,
		Pushq_imm8,
		Imul_r16_rm16_imm8,
		Imul_r32_rm32_imm8,
		Imul_r64_rm64_imm8,
		Insb_m8_DX,
		Insw_m16_DX,
		Insd_m32_DX,
		Outsb_DX_m8,
		Outsw_DX_m16,
		Outsd_DX_m32,
		Jo_rel8_16,
		Jo_rel8_32,
		Jo_rel8_64,
		Jno_rel8_16,
		Jno_rel8_32,
		Jno_rel8_64,
		Jb_rel8_16,
		Jb_rel8_32,
		Jb_rel8_64,
		Jae_rel8_16,
		Jae_rel8_32,
		Jae_rel8_64,
		Je_rel8_16,
		Je_rel8_32,
		Je_rel8_64,
		Jne_rel8_16,
		Jne_rel8_32,
		Jne_rel8_64,
		Jbe_rel8_16,
		Jbe_rel8_32,
		Jbe_rel8_64,
		Ja_rel8_16,
		Ja_rel8_32,
		Ja_rel8_64,
		Js_rel8_16,
		Js_rel8_32,
		Js_rel8_64,
		Jns_rel8_16,
		Jns_rel8_32,
		Jns_rel8_64,
		Jp_rel8_16,
		Jp_rel8_32,
		Jp_rel8_64,
		Jnp_rel8_16,
		Jnp_rel8_32,
		Jnp_rel8_64,
		Jl_rel8_16,
		Jl_rel8_32,
		Jl_rel8_64,
		Jge_rel8_16,
		Jge_rel8_32,
		Jge_rel8_64,
		Jle_rel8_16,
		Jle_rel8_32,
		Jle_rel8_64,
		Jg_rel8_16,
		Jg_rel8_32,
		Jg_rel8_64,
		Add_rm8_imm8,
		Or_rm8_imm8,
		Adc_rm8_imm8,
		Sbb_rm8_imm8,
		And_rm8_imm8,
		Sub_rm8_imm8,
		Xor_rm8_imm8,
		Cmp_rm8_imm8,
		Add_rm16_imm16,
		Add_rm32_imm32,
		Add_rm64_imm32,
		Or_rm16_imm16,
		Or_rm32_imm32,
		Or_rm64_imm32,
		Adc_rm16_imm16,
		Adc_rm32_imm32,
		Adc_rm64_imm32,
		Sbb_rm16_imm16,
		Sbb_rm32_imm32,
		Sbb_rm64_imm32,
		And_rm16_imm16,
		And_rm32_imm32,
		And_rm64_imm32,
		Sub_rm16_imm16,
		Sub_rm32_imm32,
		Sub_rm64_imm32,
		Xor_rm16_imm16,
		Xor_rm32_imm32,
		Xor_rm64_imm32,
		Cmp_rm16_imm16,
		Cmp_rm32_imm32,
		Cmp_rm64_imm32,
		Add_rm8_imm8_82,
		Or_rm8_imm8_82,
		Adc_rm8_imm8_82,
		Sbb_rm8_imm8_82,
		And_rm8_imm8_82,
		Sub_rm8_imm8_82,
		Xor_rm8_imm8_82,
		Cmp_rm8_imm8_82,
		Add_rm16_imm8,
		Add_rm32_imm8,
		Add_rm64_imm8,
		Or_rm16_imm8,
		Or_rm32_imm8,
		Or_rm64_imm8,
		Adc_rm16_imm8,
		Adc_rm32_imm8,
		Adc_rm64_imm8,
		Sbb_rm16_imm8,
		Sbb_rm32_imm8,
		Sbb_rm64_imm8,
		And_rm16_imm8,
		And_rm32_imm8,
		And_rm64_imm8,
		Sub_rm16_imm8,
		Sub_rm32_imm8,
		Sub_rm64_imm8,
		Xor_rm16_imm8,
		Xor_rm32_imm8,
		Xor_rm64_imm8,
		Cmp_rm16_imm8,
		Cmp_rm32_imm8,
		Cmp_rm64_imm8,
		Test_rm8_r8,
		Test_rm16_r16,
		Test_rm32_r32,
		Test_rm64_r64,
		Xchg_rm8_r8,
		Xchg_rm16_r16,
		Xchg_rm32_r32,
		Xchg_rm64_r64,
		Mov_rm8_r8,
		Mov_rm16_r16,
		Mov_rm32_r32,
		Mov_rm64_r64,
		Mov_r8_rm8,
		Mov_r16_rm16,
		Mov_r32_rm32,
		Mov_r64_rm64,
		Mov_rm16_Sreg,
		Mov_r32m16_Sreg,
		Mov_r64m16_Sreg,
		Lea_r16_m,
		Lea_r32_m,
		Lea_r64_m,
		Mov_Sreg_rm16,
		Mov_Sreg_r32m16,
		Mov_Sreg_r64m16,
		Pop_rm16,
		Pop_rm32,
		Pop_rm64,
		Nopw,
		Nopd,
		Nopq,
		Xchg_r16_AX,
		Xchg_r32_EAX,
		Xchg_r64_RAX,
		Pause,
		Cbw,
		Cwde,
		Cdqe,
		Cwd,
		Cdq,
		Cqo,
		Call_ptr1616,
		Call_ptr1632,
		Wait,
		Pushfw,
		Pushfd,
		Pushfq,
		Popfw,
		Popfd,
		Popfq,
		Sahf,
		Lahf,
		Mov_AL_moffs8,
		Mov_AX_moffs16,
		Mov_EAX_moffs32,
		Mov_RAX_moffs64,
		Mov_moffs8_AL,
		Mov_moffs16_AX,
		Mov_moffs32_EAX,
		Mov_moffs64_RAX,
		Movsb_m8_m8,
		Movsw_m16_m16,
		Movsd_m32_m32,
		Movsq_m64_m64,
		Cmpsb_m8_m8,
		Cmpsw_m16_m16,
		Cmpsd_m32_m32,
		Cmpsq_m64_m64,
		Test_AL_imm8,
		Test_AX_imm16,
		Test_EAX_imm32,
		Test_RAX_imm32,
		Stosb_m8_AL,
		Stosw_m16_AX,
		Stosd_m32_EAX,
		Stosq_m64_RAX,
		Lodsb_AL_m8,
		Lodsw_AX_m16,
		Lodsd_EAX_m32,
		Lodsq_RAX_m64,
		Scasb_AL_m8,
		Scasw_AX_m16,
		Scasd_EAX_m32,
		Scasq_RAX_m64,
		Mov_r8_imm8,
		Mov_r16_imm16,
		Mov_r32_imm32,
		Mov_r64_imm64,
		Rol_rm8_imm8,
		Ror_rm8_imm8,
		Rcl_rm8_imm8,
		Rcr_rm8_imm8,
		Shl_rm8_imm8,
		Shr_rm8_imm8,
		Sal_rm8_imm8,
		Sar_rm8_imm8,
		Rol_rm16_imm8,
		Rol_rm32_imm8,
		Rol_rm64_imm8,
		Ror_rm16_imm8,
		Ror_rm32_imm8,
		Ror_rm64_imm8,
		Rcl_rm16_imm8,
		Rcl_rm32_imm8,
		Rcl_rm64_imm8,
		Rcr_rm16_imm8,
		Rcr_rm32_imm8,
		Rcr_rm64_imm8,
		Shl_rm16_imm8,
		Shl_rm32_imm8,
		Shl_rm64_imm8,
		Shr_rm16_imm8,
		Shr_rm32_imm8,
		Shr_rm64_imm8,
		Sal_rm16_imm8,
		Sal_rm32_imm8,
		Sal_rm64_imm8,
		Sar_rm16_imm8,
		Sar_rm32_imm8,
		Sar_rm64_imm8,
		Retnw_imm16,
		Retnd_imm16,
		Retnq_imm16,
		Retnw,
		Retnd,
		Retnq,
		Les_r16_m1616,
		Les_r32_m1632,
		Lds_r16_m1616,
		Lds_r32_m1632,
		Mov_rm8_imm8,
		Xabort_imm8,
		Mov_rm16_imm16,
		Mov_rm32_imm32,
		Mov_rm64_imm32,
		Xbegin_rel16,
		Xbegin_rel32,
		Enterw_imm16_imm8,
		Enterd_imm16_imm8,
		Enterq_imm16_imm8,
		Leavew,
		Leaved,
		Leaveq,
		Retfw_imm16,
		Retfd_imm16,
		Retfq_imm16,
		Retfw,
		Retfd,
		Retfq,
		Int3,
		Int_imm8,
		Into,
		Iretw,
		Iretd,
		Iretq,
		Rol_rm8_1,
		Ror_rm8_1,
		Rcl_rm8_1,
		Rcr_rm8_1,
		Shl_rm8_1,
		Shr_rm8_1,
		Sal_rm8_1,
		Sar_rm8_1,
		Rol_rm16_1,
		Rol_rm32_1,
		Rol_rm64_1,
		Ror_rm16_1,
		Ror_rm32_1,
		Ror_rm64_1,
		Rcl_rm16_1,
		Rcl_rm32_1,
		Rcl_rm64_1,
		Rcr_rm16_1,
		Rcr_rm32_1,
		Rcr_rm64_1,
		Shl_rm16_1,
		Shl_rm32_1,
		Shl_rm64_1,
		Shr_rm16_1,
		Shr_rm32_1,
		Shr_rm64_1,
		Sal_rm16_1,
		Sal_rm32_1,
		Sal_rm64_1,
		Sar_rm16_1,
		Sar_rm32_1,
		Sar_rm64_1,
		Rol_rm8_CL,
		Ror_rm8_CL,
		Rcl_rm8_CL,
		Rcr_rm8_CL,
		Shl_rm8_CL,
		Shr_rm8_CL,
		Sal_rm8_CL,
		Sar_rm8_CL,
		Rol_rm16_CL,
		Rol_rm32_CL,
		Rol_rm64_CL,
		Ror_rm16_CL,
		Ror_rm32_CL,
		Ror_rm64_CL,
		Rcl_rm16_CL,
		Rcl_rm32_CL,
		Rcl_rm64_CL,
		Rcr_rm16_CL,
		Rcr_rm32_CL,
		Rcr_rm64_CL,
		Shl_rm16_CL,
		Shl_rm32_CL,
		Shl_rm64_CL,
		Shr_rm16_CL,
		Shr_rm32_CL,
		Shr_rm64_CL,
		Sal_rm16_CL,
		Sal_rm32_CL,
		Sal_rm64_CL,
		Sar_rm16_CL,
		Sar_rm32_CL,
		Sar_rm64_CL,
		Aam_imm8,
		Aad_imm8,
		Salc,
		Xlat_m8,
		Fadd_m32fp,
		Fmul_m32fp,
		Fcom_m32fp,
		Fcomp_m32fp,
		Fsub_m32fp,
		Fsubr_m32fp,
		Fdiv_m32fp,
		Fdivr_m32fp,
		Fadd_st0_sti,
		Fmul_st0_sti,
		Fcom_st0_sti,
		Fcomp_st0_sti,
		Fsub_st0_sti,
		Fsubr_st0_sti,
		Fdiv_st0_sti,
		Fdivr_st0_sti,
		Fld_m32fp,
		Fst_m32fp,
		Fstp_m32fp,
		Fldenv_m14byte,
		Fldenv_m28byte,
		Fldcw_m2byte,
		Fnstenv_m14byte,
		Fstenv_m14byte,
		Fnstenv_m28byte,
		Fstenv_m28byte,
		Fnstcw_m2byte,
		Fstcw_m2byte,
		Fld_sti,
		Fxch_st0_sti,
		Fnop,
		Fstpnce_sti,
		Fchs,
		Fabs,
		Ftst,
		Fxam,
		Fld1,
		Fldl2t,
		Fldl2e,
		Fldpi,
		Fldlg2,
		Fldln2,
		Fldz,
		F2xm1,
		Fyl2x,
		Fptan,
		Fpatan,
		Fxtract,
		Fprem1,
		Fdecstp,
		Fincstp,
		Fprem,
		Fyl2xp1,
		Fsqrt,
		Fsincos,
		Frndint,
		Fscale,
		Fsin,
		Fcos,
		Fiadd_m32int,
		Fimul_m32int,
		Ficom_m32int,
		Ficomp_m32int,
		Fisub_m32int,
		Fisubr_m32int,
		Fidiv_m32int,
		Fidivr_m32int,
		Fcmovb_st0_sti,
		Fcmove_st0_sti,
		Fcmovbe_st0_sti,
		Fcmovu_st0_sti,
		Fucompp,
		Fild_m32int,
		Fisttp_m32int,
		Fist_m32int,
		Fistp_m32int,
		Fld_m80fp,
		Fstp_m80fp,
		Fcmovnb_st0_sti,
		Fcmovne_st0_sti,
		Fcmovnbe_st0_sti,
		Fcmovnu_st0_sti,
		Fneni,
		Feni,
		Fndisi,
		Fdisi,
		Fnclex,
		Fclex,
		Fninit,
		Finit,
		Fnsetpm,
		Fsetpm,
		Frstpm,
		Fucomi_st0_sti,
		Fcomi_st0_sti,
		Fadd_m64fp,
		Fmul_m64fp,
		Fcom_m64fp,
		Fcomp_m64fp,
		Fsub_m64fp,
		Fsubr_m64fp,
		Fdiv_m64fp,
		Fdivr_m64fp,
		Fadd_sti_st0,
		Fmul_sti_st0,
		Fcom_st0_sti_DCD0,
		Fcomp_st0_sti_DCD8,
		Fsubr_sti_st0,
		Fsub_sti_st0,
		Fdivr_sti_st0,
		Fdiv_sti_st0,
		Fld_m64fp,
		Fisttp_m64int,
		Fst_m64fp,
		Fstp_m64fp,
		Frstor_m94byte,
		Frstor_m108byte,
		Fnsave_m94byte,
		Fsave_m94byte,
		Fnsave_m108byte,
		Fsave_m108byte,
		Fnstsw_m2byte,
		Fstsw_m2byte,
		Ffree_sti,
		Fxch_st0_sti_DDC8,
		Fst_sti,
		Fstp_sti,
		Fucom_st0_sti,
		Fucomp_st0_sti,
		Fiadd_m16int,
		Fimul_m16int,
		Ficom_m16int,
		Ficomp_m16int,
		Fisub_m16int,
		Fisubr_m16int,
		Fidiv_m16int,
		Fidivr_m16int,
		Faddp_sti_st0,
		Fmulp_sti_st0,
		Fcomp_st0_sti_DED0,
		Fcompp,
		Fsubrp_sti_st0,
		Fsubp_sti_st0,
		Fdivrp_sti_st0,
		Fdivp_sti_st0,
		Fild_m16int,
		Fisttp_m16int,
		Fist_m16int,
		Fistp_m16int,
		Fbld_m80bcd,
		Fild_m64int,
		Fbstp_m80bcd,
		Fistp_m64int,
		Ffreep_sti,
		Fxch_st0_sti_DFC8,
		Fstp_sti_DFD0,
		Fstp_sti_DFD8,
		Fnstsw_AX,
		Fstsw_AX,
		Fstdw_AX,
		Fstsg_AX,
		Fucomip_st0_sti,
		Fcomip_st0_sti,
		Loopne_rel8_16_CX,
		Loopne_rel8_32_CX,
		Loopne_rel8_16_ECX,
		Loopne_rel8_32_ECX,
		Loopne_rel8_64_ECX,
		Loopne_rel8_16_RCX,
		Loopne_rel8_64_RCX,
		Loope_rel8_16_CX,
		Loope_rel8_32_CX,
		Loope_rel8_16_ECX,
		Loope_rel8_32_ECX,
		Loope_rel8_64_ECX,
		Loope_rel8_16_RCX,
		Loope_rel8_64_RCX,
		Loop_rel8_16_CX,
		Loop_rel8_32_CX,
		Loop_rel8_16_ECX,
		Loop_rel8_32_ECX,
		Loop_rel8_64_ECX,
		Loop_rel8_16_RCX,
		Loop_rel8_64_RCX,
		Jcxz_rel8_16,
		Jcxz_rel8_32,
		Jecxz_rel8_16,
		Jecxz_rel8_32,
		Jecxz_rel8_64,
		Jrcxz_rel8_16,
		Jrcxz_rel8_64,
		In_AL_imm8,
		In_AX_imm8,
		In_EAX_imm8,
		Out_imm8_AL,
		Out_imm8_AX,
		Out_imm8_EAX,
		Call_rel16,
		Call_rel32_32,
		Call_rel32_64,
		Jmp_rel16,
		Jmp_rel32_32,
		Jmp_rel32_64,
		Jmp_ptr1616,
		Jmp_ptr1632,
		Jmp_rel8_16,
		Jmp_rel8_32,
		Jmp_rel8_64,
		In_AL_DX,
		In_AX_DX,
		In_EAX_DX,
		Out_DX_AL,
		Out_DX_AX,
		Out_DX_EAX,
		Int1,
		Hlt,
		Cmc,
		Test_rm8_imm8,
		Test_rm8_imm8_F6r1,
		Not_rm8,
		Neg_rm8,
		Mul_rm8,
		Imul_rm8,
		Div_rm8,
		Idiv_rm8,
		Test_rm16_imm16,
		Test_rm32_imm32,
		Test_rm64_imm32,
		Test_rm16_imm16_F7r1,
		Test_rm32_imm32_F7r1,
		Test_rm64_imm32_F7r1,
		Not_rm16,
		Not_rm32,
		Not_rm64,
		Neg_rm16,
		Neg_rm32,
		Neg_rm64,
		Mul_rm16,
		Mul_rm32,
		Mul_rm64,
		Imul_rm16,
		Imul_rm32,
		Imul_rm64,
		Div_rm16,
		Div_rm32,
		Div_rm64,
		Idiv_rm16,
		Idiv_rm32,
		Idiv_rm64,
		Clc,
		Stc,
		Cli,
		Sti,
		Cld,
		Std,
		Inc_rm8,
		Dec_rm8,
		Inc_rm16,
		Inc_rm32,
		Inc_rm64,
		Dec_rm16,
		Dec_rm32,
		Dec_rm64,
		Call_rm16,
		Call_rm32,
		Call_rm64,
		Call_m1616,
		Call_m1632,
		Call_m1664,
		Jmp_rm16,
		Jmp_rm32,
		Jmp_rm64,
		Jmp_m1616,
		Jmp_m1632,
		Jmp_m1664,
		Push_rm16,
		Push_rm32,
		Push_rm64,
		Sldt_rm16,
		Sldt_r32m16,
		Sldt_r64m16,
		Str_rm16,
		Str_r32m16,
		Str_r64m16,
		Lldt_rm16,
		Lldt_r32m16,
		Lldt_r64m16,
		Ltr_rm16,
		Ltr_r32m16,
		Ltr_r64m16,
		Verr_rm16,
		Verr_r32m16,
		Verr_r64m16,
		Verw_rm16,
		Verw_r32m16,
		Verw_r64m16,
		Jmpe_rm16,
		Jmpe_rm32,
		Sgdt_m1632_16,
		Sgdt_m1632,
		Sgdt_m1664,
		Sidt_m1632_16,
		Sidt_m1632,
		Sidt_m1664,
		Lgdt_m1632_16,
		Lgdt_m1632,
		Lgdt_m1664,
		Lidt_m1632_16,
		Lidt_m1632,
		Lidt_m1664,
		Smsw_rm16,
		Smsw_r32m16,
		Smsw_r64m16,
		Rstorssp_m64,
		Lmsw_rm16,
		Lmsw_r32m16,
		Lmsw_r64m16,
		Invlpg_m,
		Enclv,
		Vmcall,
		Vmlaunch,
		Vmresume,
		Vmxoff,
		Pconfig,
		Monitorw,
		Monitord,
		Monitorq,
		Mwait,
		Clac,
		Stac,
		Encls,
		Xgetbv,
		Xsetbv,
		Vmfunc,
		Xend,
		Xtest,
		Enclu,
		Vmrunw,
		Vmrund,
		Vmrunq,
		Vmmcall,
		Vmloadw,
		Vmloadd,
		Vmloadq,
		Vmsavew,
		Vmsaved,
		Vmsaveq,
		Stgi,
		Clgi,
		Skinit,
		Invlpgaw,
		Invlpgad,
		Invlpgaq,
		Setssbsy,
		Saveprevssp,
		Rdpkru,
		Wrpkru,
		Swapgs,
		Rdtscp,
		Monitorxw,
		Monitorxd,
		Monitorxq,
		Mcommit,
		Mwaitx,
		Clzerow,
		Clzerod,
		Clzeroq,
		Rdpru,
		Lar_r16_rm16,
		Lar_r32_r32m16,
		Lar_r64_r64m16,
		Lsl_r16_rm16,
		Lsl_r32_r32m16,
		Lsl_r64_r64m16,
		Storeall,
		Loadall286,
		Syscall,
		Clts,
		Loadall386,
		Sysretd,
		Sysretq,
		Invd,
		Wbinvd,
		Wbnoinvd,
		Cl1invmb,
		Ud2,
		Reservednop_rm16_r16_0F0D,
		Reservednop_rm32_r32_0F0D,
		Reservednop_rm64_r64_0F0D,
		Prefetch_m8,
		Prefetchw_m8,
		Prefetchwt1_m8,
		Femms,
		Umov_rm8_r8,
		Umov_rm16_r16,
		Umov_rm32_r32,
		Umov_r8_rm8,
		Umov_r16_rm16,
		Umov_r32_rm32,
		Movups_xmm_xmmm128,
		VEX_Vmovups_xmm_xmmm128,
		VEX_Vmovups_ymm_ymmm256,
		EVEX_Vmovups_xmm_k1z_xmmm128,
		EVEX_Vmovups_ymm_k1z_ymmm256,
		EVEX_Vmovups_zmm_k1z_zmmm512,
		Movupd_xmm_xmmm128,
		VEX_Vmovupd_xmm_xmmm128,
		VEX_Vmovupd_ymm_ymmm256,
		EVEX_Vmovupd_xmm_k1z_xmmm128,
		EVEX_Vmovupd_ymm_k1z_ymmm256,
		EVEX_Vmovupd_zmm_k1z_zmmm512,
		Movss_xmm_xmmm32,
		VEX_Vmovss_xmm_xmm_xmm,
		VEX_Vmovss_xmm_m32,
		EVEX_Vmovss_xmm_k1z_xmm_xmm,
		EVEX_Vmovss_xmm_k1z_m32,
		Movsd_xmm_xmmm64,
		VEX_Vmovsd_xmm_xmm_xmm,
		VEX_Vmovsd_xmm_m64,
		EVEX_Vmovsd_xmm_k1z_xmm_xmm,
		EVEX_Vmovsd_xmm_k1z_m64,
		Movups_xmmm128_xmm,
		VEX_Vmovups_xmmm128_xmm,
		VEX_Vmovups_ymmm256_ymm,
		EVEX_Vmovups_xmmm128_k1z_xmm,
		EVEX_Vmovups_ymmm256_k1z_ymm,
		EVEX_Vmovups_zmmm512_k1z_zmm,
		Movupd_xmmm128_xmm,
		VEX_Vmovupd_xmmm128_xmm,
		VEX_Vmovupd_ymmm256_ymm,
		EVEX_Vmovupd_xmmm128_k1z_xmm,
		EVEX_Vmovupd_ymmm256_k1z_ymm,
		EVEX_Vmovupd_zmmm512_k1z_zmm,
		Movss_xmmm32_xmm,
		VEX_Vmovss_xmm_xmm_xmm_0F11,
		VEX_Vmovss_m32_xmm,
		EVEX_Vmovss_xmm_k1z_xmm_xmm_0F11,
		EVEX_Vmovss_m32_k1_xmm,
		Movsd_xmmm64_xmm,
		VEX_Vmovsd_xmm_xmm_xmm_0F11,
		VEX_Vmovsd_m64_xmm,
		EVEX_Vmovsd_xmm_k1z_xmm_xmm_0F11,
		EVEX_Vmovsd_m64_k1_xmm,
		Movhlps_xmm_xmm,
		Movlps_xmm_m64,
		VEX_Vmovhlps_xmm_xmm_xmm,
		VEX_Vmovlps_xmm_xmm_m64,
		EVEX_Vmovhlps_xmm_xmm_xmm,
		EVEX_Vmovlps_xmm_xmm_m64,
		Movlpd_xmm_m64,
		VEX_Vmovlpd_xmm_xmm_m64,
		EVEX_Vmovlpd_xmm_xmm_m64,
		Movsldup_xmm_xmmm128,
		VEX_Vmovsldup_xmm_xmmm128,
		VEX_Vmovsldup_ymm_ymmm256,
		EVEX_Vmovsldup_xmm_k1z_xmmm128,
		EVEX_Vmovsldup_ymm_k1z_ymmm256,
		EVEX_Vmovsldup_zmm_k1z_zmmm512,
		Movddup_xmm_xmmm64,
		VEX_Vmovddup_xmm_xmmm64,
		VEX_Vmovddup_ymm_ymmm256,
		EVEX_Vmovddup_xmm_k1z_xmmm64,
		EVEX_Vmovddup_ymm_k1z_ymmm256,
		EVEX_Vmovddup_zmm_k1z_zmmm512,
		Movlps_m64_xmm,
		VEX_Vmovlps_m64_xmm,
		EVEX_Vmovlps_m64_xmm,
		Movlpd_m64_xmm,
		VEX_Vmovlpd_m64_xmm,
		EVEX_Vmovlpd_m64_xmm,
		Unpcklps_xmm_xmmm128,
		VEX_Vunpcklps_xmm_xmm_xmmm128,
		VEX_Vunpcklps_ymm_ymm_ymmm256,
		EVEX_Vunpcklps_xmm_k1z_xmm_xmmm128b32,
		EVEX_Vunpcklps_ymm_k1z_ymm_ymmm256b32,
		EVEX_Vunpcklps_zmm_k1z_zmm_zmmm512b32,
		Unpcklpd_xmm_xmmm128,
		VEX_Vunpcklpd_xmm_xmm_xmmm128,
		VEX_Vunpcklpd_ymm_ymm_ymmm256,
		EVEX_Vunpcklpd_xmm_k1z_xmm_xmmm128b64,
		EVEX_Vunpcklpd_ymm_k1z_ymm_ymmm256b64,
		EVEX_Vunpcklpd_zmm_k1z_zmm_zmmm512b64,
		Unpckhps_xmm_xmmm128,
		VEX_Vunpckhps_xmm_xmm_xmmm128,
		VEX_Vunpckhps_ymm_ymm_ymmm256,
		EVEX_Vunpckhps_xmm_k1z_xmm_xmmm128b32,
		EVEX_Vunpckhps_ymm_k1z_ymm_ymmm256b32,
		EVEX_Vunpckhps_zmm_k1z_zmm_zmmm512b32,
		Unpckhpd_xmm_xmmm128,
		VEX_Vunpckhpd_xmm_xmm_xmmm128,
		VEX_Vunpckhpd_ymm_ymm_ymmm256,
		EVEX_Vunpckhpd_xmm_k1z_xmm_xmmm128b64,
		EVEX_Vunpckhpd_ymm_k1z_ymm_ymmm256b64,
		EVEX_Vunpckhpd_zmm_k1z_zmm_zmmm512b64,
		Movlhps_xmm_xmm,
		VEX_Vmovlhps_xmm_xmm_xmm,
		EVEX_Vmovlhps_xmm_xmm_xmm,
		Movhps_xmm_m64,
		VEX_Vmovhps_xmm_xmm_m64,
		EVEX_Vmovhps_xmm_xmm_m64,
		Movhpd_xmm_m64,
		VEX_Vmovhpd_xmm_xmm_m64,
		EVEX_Vmovhpd_xmm_xmm_m64,
		Movshdup_xmm_xmmm128,
		VEX_Vmovshdup_xmm_xmmm128,
		VEX_Vmovshdup_ymm_ymmm256,
		EVEX_Vmovshdup_xmm_k1z_xmmm128,
		EVEX_Vmovshdup_ymm_k1z_ymmm256,
		EVEX_Vmovshdup_zmm_k1z_zmmm512,
		Movhps_m64_xmm,
		VEX_Vmovhps_m64_xmm,
		EVEX_Vmovhps_m64_xmm,
		Movhpd_m64_xmm,
		VEX_Vmovhpd_m64_xmm,
		EVEX_Vmovhpd_m64_xmm,
		Reservednop_rm16_r16_0F18,
		Reservednop_rm32_r32_0F18,
		Reservednop_rm64_r64_0F18,
		Reservednop_rm16_r16_0F19,
		Reservednop_rm32_r32_0F19,
		Reservednop_rm64_r64_0F19,
		Reservednop_rm16_r16_0F1A,
		Reservednop_rm32_r32_0F1A,
		Reservednop_rm64_r64_0F1A,
		Reservednop_rm16_r16_0F1B,
		Reservednop_rm32_r32_0F1B,
		Reservednop_rm64_r64_0F1B,
		Reservednop_rm16_r16_0F1C,
		Reservednop_rm32_r32_0F1C,
		Reservednop_rm64_r64_0F1C,
		Reservednop_rm16_r16_0F1D,
		Reservednop_rm32_r32_0F1D,
		Reservednop_rm64_r64_0F1D,
		Reservednop_rm16_r16_0F1E,
		Reservednop_rm32_r32_0F1E,
		Reservednop_rm64_r64_0F1E,
		Reservednop_rm16_r16_0F1F,
		Reservednop_rm32_r32_0F1F,
		Reservednop_rm64_r64_0F1F,
		Prefetchnta_m8,
		Prefetcht0_m8,
		Prefetcht1_m8,
		Prefetcht2_m8,
		Bndldx_bnd_mib,
		Bndmov_bnd_bndm64,
		Bndmov_bnd_bndm128,
		Bndcl_bnd_rm32,
		Bndcl_bnd_rm64,
		Bndcu_bnd_rm32,
		Bndcu_bnd_rm64,
		Bndstx_mib_bnd,
		Bndmov_bndm64_bnd,
		Bndmov_bndm128_bnd,
		Bndmk_bnd_m32,
		Bndmk_bnd_m64,
		Bndcn_bnd_rm32,
		Bndcn_bnd_rm64,
		Cldemote_m8,
		Rdsspd_r32,
		Rdsspq_r64,
		Endbr64,
		Endbr32,
		Nop_rm16,
		Nop_rm32,
		Nop_rm64,
		Mov_r32_cr,
		Mov_r64_cr,
		Mov_r32_dr,
		Mov_r64_dr,
		Mov_cr_r32,
		Mov_cr_r64,
		Mov_dr_r32,
		Mov_dr_r64,
		Mov_r32_tr,
		Mov_tr_r32,
		Movaps_xmm_xmmm128,
		VEX_Vmovaps_xmm_xmmm128,
		VEX_Vmovaps_ymm_ymmm256,
		EVEX_Vmovaps_xmm_k1z_xmmm128,
		EVEX_Vmovaps_ymm_k1z_ymmm256,
		EVEX_Vmovaps_zmm_k1z_zmmm512,
		Movapd_xmm_xmmm128,
		VEX_Vmovapd_xmm_xmmm128,
		VEX_Vmovapd_ymm_ymmm256,
		EVEX_Vmovapd_xmm_k1z_xmmm128,
		EVEX_Vmovapd_ymm_k1z_ymmm256,
		EVEX_Vmovapd_zmm_k1z_zmmm512,
		Movaps_xmmm128_xmm,
		VEX_Vmovaps_xmmm128_xmm,
		VEX_Vmovaps_ymmm256_ymm,
		EVEX_Vmovaps_xmmm128_k1z_xmm,
		EVEX_Vmovaps_ymmm256_k1z_ymm,
		EVEX_Vmovaps_zmmm512_k1z_zmm,
		Movapd_xmmm128_xmm,
		VEX_Vmovapd_xmmm128_xmm,
		VEX_Vmovapd_ymmm256_ymm,
		EVEX_Vmovapd_xmmm128_k1z_xmm,
		EVEX_Vmovapd_ymmm256_k1z_ymm,
		EVEX_Vmovapd_zmmm512_k1z_zmm,
		Cvtpi2ps_xmm_mmm64,
		Cvtpi2pd_xmm_mmm64,
		Cvtsi2ss_xmm_rm32,
		Cvtsi2ss_xmm_rm64,
		VEX_Vcvtsi2ss_xmm_xmm_rm32,
		VEX_Vcvtsi2ss_xmm_xmm_rm64,
		EVEX_Vcvtsi2ss_xmm_xmm_rm32_er,
		EVEX_Vcvtsi2ss_xmm_xmm_rm64_er,
		Cvtsi2sd_xmm_rm32,
		Cvtsi2sd_xmm_rm64,
		VEX_Vcvtsi2sd_xmm_xmm_rm32,
		VEX_Vcvtsi2sd_xmm_xmm_rm64,
		EVEX_Vcvtsi2sd_xmm_xmm_rm32_er,
		EVEX_Vcvtsi2sd_xmm_xmm_rm64_er,
		Movntps_m128_xmm,
		VEX_Vmovntps_m128_xmm,
		VEX_Vmovntps_m256_ymm,
		EVEX_Vmovntps_m128_xmm,
		EVEX_Vmovntps_m256_ymm,
		EVEX_Vmovntps_m512_zmm,
		Movntpd_m128_xmm,
		VEX_Vmovntpd_m128_xmm,
		VEX_Vmovntpd_m256_ymm,
		EVEX_Vmovntpd_m128_xmm,
		EVEX_Vmovntpd_m256_ymm,
		EVEX_Vmovntpd_m512_zmm,
		Movntss_m32_xmm,
		Movntsd_m64_xmm,
		Cvttps2pi_mm_xmmm64,
		Cvttpd2pi_mm_xmmm128,
		Cvttss2si_r32_xmmm32,
		Cvttss2si_r64_xmmm32,
		VEX_Vcvttss2si_r32_xmmm32,
		VEX_Vcvttss2si_r64_xmmm32,
		EVEX_Vcvttss2si_r32_xmmm32_sae,
		EVEX_Vcvttss2si_r64_xmmm32_sae,
		Cvttsd2si_r32_xmmm64,
		Cvttsd2si_r64_xmmm64,
		VEX_Vcvttsd2si_r32_xmmm64,
		VEX_Vcvttsd2si_r64_xmmm64,
		EVEX_Vcvttsd2si_r32_xmmm64_sae,
		EVEX_Vcvttsd2si_r64_xmmm64_sae,
		Cvtps2pi_mm_xmmm64,
		Cvtpd2pi_mm_xmmm128,
		Cvtss2si_r32_xmmm32,
		Cvtss2si_r64_xmmm32,
		VEX_Vcvtss2si_r32_xmmm32,
		VEX_Vcvtss2si_r64_xmmm32,
		EVEX_Vcvtss2si_r32_xmmm32_er,
		EVEX_Vcvtss2si_r64_xmmm32_er,
		Cvtsd2si_r32_xmmm64,
		Cvtsd2si_r64_xmmm64,
		VEX_Vcvtsd2si_r32_xmmm64,
		VEX_Vcvtsd2si_r64_xmmm64,
		EVEX_Vcvtsd2si_r32_xmmm64_er,
		EVEX_Vcvtsd2si_r64_xmmm64_er,
		Ucomiss_xmm_xmmm32,
		VEX_Vucomiss_xmm_xmmm32,
		EVEX_Vucomiss_xmm_xmmm32_sae,
		Ucomisd_xmm_xmmm64,
		VEX_Vucomisd_xmm_xmmm64,
		EVEX_Vucomisd_xmm_xmmm64_sae,
		Comiss_xmm_xmmm32,
		Comisd_xmm_xmmm64,
		VEX_Vcomiss_xmm_xmmm32,
		VEX_Vcomisd_xmm_xmmm64,
		EVEX_Vcomiss_xmm_xmmm32_sae,
		EVEX_Vcomisd_xmm_xmmm64_sae,
		Wrmsr,
		Rdtsc,
		Rdmsr,
		Rdpmc,
		Sysenter,
		Sysexitd,
		Sysexitq,
		Getsecd,
		Cmovo_r16_rm16,
		Cmovo_r32_rm32,
		Cmovo_r64_rm64,
		Cmovno_r16_rm16,
		Cmovno_r32_rm32,
		Cmovno_r64_rm64,
		Cmovb_r16_rm16,
		Cmovb_r32_rm32,
		Cmovb_r64_rm64,
		Cmovae_r16_rm16,
		Cmovae_r32_rm32,
		Cmovae_r64_rm64,
		Cmove_r16_rm16,
		Cmove_r32_rm32,
		Cmove_r64_rm64,
		Cmovne_r16_rm16,
		Cmovne_r32_rm32,
		Cmovne_r64_rm64,
		Cmovbe_r16_rm16,
		Cmovbe_r32_rm32,
		Cmovbe_r64_rm64,
		Cmova_r16_rm16,
		Cmova_r32_rm32,
		Cmova_r64_rm64,
		Cmovs_r16_rm16,
		Cmovs_r32_rm32,
		Cmovs_r64_rm64,
		Cmovns_r16_rm16,
		Cmovns_r32_rm32,
		Cmovns_r64_rm64,
		Cmovp_r16_rm16,
		Cmovp_r32_rm32,
		Cmovp_r64_rm64,
		Cmovnp_r16_rm16,
		Cmovnp_r32_rm32,
		Cmovnp_r64_rm64,
		Cmovl_r16_rm16,
		Cmovl_r32_rm32,
		Cmovl_r64_rm64,
		Cmovge_r16_rm16,
		Cmovge_r32_rm32,
		Cmovge_r64_rm64,
		Cmovle_r16_rm16,
		Cmovle_r32_rm32,
		Cmovle_r64_rm64,
		Cmovg_r16_rm16,
		Cmovg_r32_rm32,
		Cmovg_r64_rm64,
		VEX_Kandw_kr_kr_kr,
		VEX_Kandq_kr_kr_kr,
		VEX_Kandb_kr_kr_kr,
		VEX_Kandd_kr_kr_kr,
		VEX_Kandnw_kr_kr_kr,
		VEX_Kandnq_kr_kr_kr,
		VEX_Kandnb_kr_kr_kr,
		VEX_Kandnd_kr_kr_kr,
		VEX_Knotw_kr_kr,
		VEX_Knotq_kr_kr,
		VEX_Knotb_kr_kr,
		VEX_Knotd_kr_kr,
		VEX_Korw_kr_kr_kr,
		VEX_Korq_kr_kr_kr,
		VEX_Korb_kr_kr_kr,
		VEX_Kord_kr_kr_kr,
		VEX_Kxnorw_kr_kr_kr,
		VEX_Kxnorq_kr_kr_kr,
		VEX_Kxnorb_kr_kr_kr,
		VEX_Kxnord_kr_kr_kr,
		VEX_Kxorw_kr_kr_kr,
		VEX_Kxorq_kr_kr_kr,
		VEX_Kxorb_kr_kr_kr,
		VEX_Kxord_kr_kr_kr,
		VEX_Kaddw_kr_kr_kr,
		VEX_Kaddq_kr_kr_kr,
		VEX_Kaddb_kr_kr_kr,
		VEX_Kaddd_kr_kr_kr,
		VEX_Kunpckwd_kr_kr_kr,
		VEX_Kunpckdq_kr_kr_kr,
		VEX_Kunpckbw_kr_kr_kr,
		Movmskps_r32_xmm,
		Movmskps_r64_xmm,
		VEX_Vmovmskps_r32_xmm,
		VEX_Vmovmskps_r64_xmm,
		VEX_Vmovmskps_r32_ymm,
		VEX_Vmovmskps_r64_ymm,
		Movmskpd_r32_xmm,
		Movmskpd_r64_xmm,
		VEX_Vmovmskpd_r32_xmm,
		VEX_Vmovmskpd_r64_xmm,
		VEX_Vmovmskpd_r32_ymm,
		VEX_Vmovmskpd_r64_ymm,
		Sqrtps_xmm_xmmm128,
		VEX_Vsqrtps_xmm_xmmm128,
		VEX_Vsqrtps_ymm_ymmm256,
		EVEX_Vsqrtps_xmm_k1z_xmmm128b32,
		EVEX_Vsqrtps_ymm_k1z_ymmm256b32,
		EVEX_Vsqrtps_zmm_k1z_zmmm512b32_er,
		Sqrtpd_xmm_xmmm128,
		VEX_Vsqrtpd_xmm_xmmm128,
		VEX_Vsqrtpd_ymm_ymmm256,
		EVEX_Vsqrtpd_xmm_k1z_xmmm128b64,
		EVEX_Vsqrtpd_ymm_k1z_ymmm256b64,
		EVEX_Vsqrtpd_zmm_k1z_zmmm512b64_er,
		Sqrtss_xmm_xmmm32,
		VEX_Vsqrtss_xmm_xmm_xmmm32,
		EVEX_Vsqrtss_xmm_k1z_xmm_xmmm32_er,
		Sqrtsd_xmm_xmmm64,
		VEX_Vsqrtsd_xmm_xmm_xmmm64,
		EVEX_Vsqrtsd_xmm_k1z_xmm_xmmm64_er,
		Rsqrtps_xmm_xmmm128,
		VEX_Vrsqrtps_xmm_xmmm128,
		VEX_Vrsqrtps_ymm_ymmm256,
		Rsqrtss_xmm_xmmm32,
		VEX_Vrsqrtss_xmm_xmm_xmmm32,
		Rcpps_xmm_xmmm128,
		VEX_Vrcpps_xmm_xmmm128,
		VEX_Vrcpps_ymm_ymmm256,
		Rcpss_xmm_xmmm32,
		VEX_Vrcpss_xmm_xmm_xmmm32,
		Andps_xmm_xmmm128,
		VEX_Vandps_xmm_xmm_xmmm128,
		VEX_Vandps_ymm_ymm_ymmm256,
		EVEX_Vandps_xmm_k1z_xmm_xmmm128b32,
		EVEX_Vandps_ymm_k1z_ymm_ymmm256b32,
		EVEX_Vandps_zmm_k1z_zmm_zmmm512b32,
		Andpd_xmm_xmmm128,
		VEX_Vandpd_xmm_xmm_xmmm128,
		VEX_Vandpd_ymm_ymm_ymmm256,
		EVEX_Vandpd_xmm_k1z_xmm_xmmm128b64,
		EVEX_Vandpd_ymm_k1z_ymm_ymmm256b64,
		EVEX_Vandpd_zmm_k1z_zmm_zmmm512b64,
		Andnps_xmm_xmmm128,
		VEX_Vandnps_xmm_xmm_xmmm128,
		VEX_Vandnps_ymm_ymm_ymmm256,
		EVEX_Vandnps_xmm_k1z_xmm_xmmm128b32,
		EVEX_Vandnps_ymm_k1z_ymm_ymmm256b32,
		EVEX_Vandnps_zmm_k1z_zmm_zmmm512b32,
		Andnpd_xmm_xmmm128,
		VEX_Vandnpd_xmm_xmm_xmmm128,
		VEX_Vandnpd_ymm_ymm_ymmm256,
		EVEX_Vandnpd_xmm_k1z_xmm_xmmm128b64,
		EVEX_Vandnpd_ymm_k1z_ymm_ymmm256b64,
		EVEX_Vandnpd_zmm_k1z_zmm_zmmm512b64,
		Orps_xmm_xmmm128,
		VEX_Vorps_xmm_xmm_xmmm128,
		VEX_Vorps_ymm_ymm_ymmm256,
		EVEX_Vorps_xmm_k1z_xmm_xmmm128b32,
		EVEX_Vorps_ymm_k1z_ymm_ymmm256b32,
		EVEX_Vorps_zmm_k1z_zmm_zmmm512b32,
		Orpd_xmm_xmmm128,
		VEX_Vorpd_xmm_xmm_xmmm128,
		VEX_Vorpd_ymm_ymm_ymmm256,
		EVEX_Vorpd_xmm_k1z_xmm_xmmm128b64,
		EVEX_Vorpd_ymm_k1z_ymm_ymmm256b64,
		EVEX_Vorpd_zmm_k1z_zmm_zmmm512b64,
		Xorps_xmm_xmmm128,
		VEX_Vxorps_xmm_xmm_xmmm128,
		VEX_Vxorps_ymm_ymm_ymmm256,
		EVEX_Vxorps_xmm_k1z_xmm_xmmm128b32,
		EVEX_Vxorps_ymm_k1z_ymm_ymmm256b32,
		EVEX_Vxorps_zmm_k1z_zmm_zmmm512b32,
		Xorpd_xmm_xmmm128,
		VEX_Vxorpd_xmm_xmm_xmmm128,
		VEX_Vxorpd_ymm_ymm_ymmm256,
		EVEX_Vxorpd_xmm_k1z_xmm_xmmm128b64,
		EVEX_Vxorpd_ymm_k1z_ymm_ymmm256b64,
		EVEX_Vxorpd_zmm_k1z_zmm_zmmm512b64,
		Addps_xmm_xmmm128,
		VEX_Vaddps_xmm_xmm_xmmm128,
		VEX_Vaddps_ymm_ymm_ymmm256,
		EVEX_Vaddps_xmm_k1z_xmm_xmmm128b32,
		EVEX_Vaddps_ymm_k1z_ymm_ymmm256b32,
		EVEX_Vaddps_zmm_k1z_zmm_zmmm512b32_er,
		Addpd_xmm_xmmm128,
		VEX_Vaddpd_xmm_xmm_xmmm128,
		VEX_Vaddpd_ymm_ymm_ymmm256,
		EVEX_Vaddpd_xmm_k1z_xmm_xmmm128b64,
		EVEX_Vaddpd_ymm_k1z_ymm_ymmm256b64,
		EVEX_Vaddpd_zmm_k1z_zmm_zmmm512b64_er,
		Addss_xmm_xmmm32,
		VEX_Vaddss_xmm_xmm_xmmm32,
		EVEX_Vaddss_xmm_k1z_xmm_xmmm32_er,
		Addsd_xmm_xmmm64,
		VEX_Vaddsd_xmm_xmm_xmmm64,
		EVEX_Vaddsd_xmm_k1z_xmm_xmmm64_er,
		Mulps_xmm_xmmm128,
		VEX_Vmulps_xmm_xmm_xmmm128,
		VEX_Vmulps_ymm_ymm_ymmm256,
		EVEX_Vmulps_xmm_k1z_xmm_xmmm128b32,
		EVEX_Vmulps_ymm_k1z_ymm_ymmm256b32,
		EVEX_Vmulps_zmm_k1z_zmm_zmmm512b32_er,
		Mulpd_xmm_xmmm128,
		VEX_Vmulpd_xmm_xmm_xmmm128,
		VEX_Vmulpd_ymm_ymm_ymmm256,
		EVEX_Vmulpd_xmm_k1z_xmm_xmmm128b64,
		EVEX_Vmulpd_ymm_k1z_ymm_ymmm256b64,
		EVEX_Vmulpd_zmm_k1z_zmm_zmmm512b64_er,
		Mulss_xmm_xmmm32,
		VEX_Vmulss_xmm_xmm_xmmm32,
		EVEX_Vmulss_xmm_k1z_xmm_xmmm32_er,
		Mulsd_xmm_xmmm64,
		VEX_Vmulsd_xmm_xmm_xmmm64,
		EVEX_Vmulsd_xmm_k1z_xmm_xmmm64_er,
		Cvtps2pd_xmm_xmmm64,
		VEX_Vcvtps2pd_xmm_xmmm64,
		VEX_Vcvtps2pd_ymm_xmmm128,
		EVEX_Vcvtps2pd_xmm_k1z_xmmm64b32,
		EVEX_Vcvtps2pd_ymm_k1z_xmmm128b32,
		EVEX_Vcvtps2pd_zmm_k1z_ymmm256b32_sae,
		Cvtpd2ps_xmm_xmmm128,
		VEX_Vcvtpd2ps_xmm_xmmm128,
		VEX_Vcvtpd2ps_xmm_ymmm256,
		EVEX_Vcvtpd2ps_xmm_k1z_xmmm128b64,
		EVEX_Vcvtpd2ps_xmm_k1z_ymmm256b64,
		EVEX_Vcvtpd2ps_ymm_k1z_zmmm512b64_er,
		Cvtss2sd_xmm_xmmm32,
		VEX_Vcvtss2sd_xmm_xmm_xmmm32,
		EVEX_Vcvtss2sd_xmm_k1z_xmm_xmmm32_sae,
		Cvtsd2ss_xmm_xmmm64,
		VEX_Vcvtsd2ss_xmm_xmm_xmmm64,
		EVEX_Vcvtsd2ss_xmm_k1z_xmm_xmmm64_er,
		Cvtdq2ps_xmm_xmmm128,
		VEX_Vcvtdq2ps_xmm_xmmm128,
		VEX_Vcvtdq2ps_ymm_ymmm256,
		EVEX_Vcvtdq2ps_xmm_k1z_xmmm128b32,
		EVEX_Vcvtdq2ps_ymm_k1z_ymmm256b32,
		EVEX_Vcvtdq2ps_zmm_k1z_zmmm512b32_er,
		EVEX_Vcvtqq2ps_xmm_k1z_xmmm128b64,
		EVEX_Vcvtqq2ps_xmm_k1z_ymmm256b64,
		EVEX_Vcvtqq2ps_ymm_k1z_zmmm512b64_er,
		Cvtps2dq_xmm_xmmm128,
		VEX_Vcvtps2dq_xmm_xmmm128,
		VEX_Vcvtps2dq_ymm_ymmm256,
		EVEX_Vcvtps2dq_xmm_k1z_xmmm128b32,
		EVEX_Vcvtps2dq_ymm_k1z_ymmm256b32,
		EVEX_Vcvtps2dq_zmm_k1z_zmmm512b32_er,
		Cvttps2dq_xmm_xmmm128,
		VEX_Vcvttps2dq_xmm_xmmm128,
		VEX_Vcvttps2dq_ymm_ymmm256,
		EVEX_Vcvttps2dq_xmm_k1z_xmmm128b32,
		EVEX_Vcvttps2dq_ymm_k1z_ymmm256b32,
		EVEX_Vcvttps2dq_zmm_k1z_zmmm512b32_sae,
		Subps_xmm_xmmm128,
		VEX_Vsubps_xmm_xmm_xmmm128,
		VEX_Vsubps_ymm_ymm_ymmm256,
		EVEX_Vsubps_xmm_k1z_xmm_xmmm128b32,
		EVEX_Vsubps_ymm_k1z_ymm_ymmm256b32,
		EVEX_Vsubps_zmm_k1z_zmm_zmmm512b32_er,
		Subpd_xmm_xmmm128,
		VEX_Vsubpd_xmm_xmm_xmmm128,
		VEX_Vsubpd_ymm_ymm_ymmm256,
		EVEX_Vsubpd_xmm_k1z_xmm_xmmm128b64,
		EVEX_Vsubpd_ymm_k1z_ymm_ymmm256b64,
		EVEX_Vsubpd_zmm_k1z_zmm_zmmm512b64_er,
		Subss_xmm_xmmm32,
		VEX_Vsubss_xmm_xmm_xmmm32,
		EVEX_Vsubss_xmm_k1z_xmm_xmmm32_er,
		Subsd_xmm_xmmm64,
		VEX_Vsubsd_xmm_xmm_xmmm64,
		EVEX_Vsubsd_xmm_k1z_xmm_xmmm64_er,
		Minps_xmm_xmmm128,
		VEX_Vminps_xmm_xmm_xmmm128,
		VEX_Vminps_ymm_ymm_ymmm256,
		EVEX_Vminps_xmm_k1z_xmm_xmmm128b32,
		EVEX_Vminps_ymm_k1z_ymm_ymmm256b32,
		EVEX_Vminps_zmm_k1z_zmm_zmmm512b32_sae,
		Minpd_xmm_xmmm128,
		VEX_Vminpd_xmm_xmm_xmmm128,
		VEX_Vminpd_ymm_ymm_ymmm256,
		EVEX_Vminpd_xmm_k1z_xmm_xmmm128b64,
		EVEX_Vminpd_ymm_k1z_ymm_ymmm256b64,
		EVEX_Vminpd_zmm_k1z_zmm_zmmm512b64_sae,
		Minss_xmm_xmmm32,
		VEX_Vminss_xmm_xmm_xmmm32,
		EVEX_Vminss_xmm_k1z_xmm_xmmm32_sae,
		Minsd_xmm_xmmm64,
		VEX_Vminsd_xmm_xmm_xmmm64,
		EVEX_Vminsd_xmm_k1z_xmm_xmmm64_sae,
		Divps_xmm_xmmm128,
		VEX_Vdivps_xmm_xmm_xmmm128,
		VEX_Vdivps_ymm_ymm_ymmm256,
		EVEX_Vdivps_xmm_k1z_xmm_xmmm128b32,
		EVEX_Vdivps_ymm_k1z_ymm_ymmm256b32,
		EVEX_Vdivps_zmm_k1z_zmm_zmmm512b32_er,
		Divpd_xmm_xmmm128,
		VEX_Vdivpd_xmm_xmm_xmmm128,
		VEX_Vdivpd_ymm_ymm_ymmm256,
		EVEX_Vdivpd_xmm_k1z_xmm_xmmm128b64,
		EVEX_Vdivpd_ymm_k1z_ymm_ymmm256b64,
		EVEX_Vdivpd_zmm_k1z_zmm_zmmm512b64_er,
		Divss_xmm_xmmm32,
		VEX_Vdivss_xmm_xmm_xmmm32,
		EVEX_Vdivss_xmm_k1z_xmm_xmmm32_er,
		Divsd_xmm_xmmm64,
		VEX_Vdivsd_xmm_xmm_xmmm64,
		EVEX_Vdivsd_xmm_k1z_xmm_xmmm64_er,
		Maxps_xmm_xmmm128,
		VEX_Vmaxps_xmm_xmm_xmmm128,
		VEX_Vmaxps_ymm_ymm_ymmm256,
		EVEX_Vmaxps_xmm_k1z_xmm_xmmm128b32,
		EVEX_Vmaxps_ymm_k1z_ymm_ymmm256b32,
		EVEX_Vmaxps_zmm_k1z_zmm_zmmm512b32_sae,
		Maxpd_xmm_xmmm128,
		VEX_Vmaxpd_xmm_xmm_xmmm128,
		VEX_Vmaxpd_ymm_ymm_ymmm256,
		EVEX_Vmaxpd_xmm_k1z_xmm_xmmm128b64,
		EVEX_Vmaxpd_ymm_k1z_ymm_ymmm256b64,
		EVEX_Vmaxpd_zmm_k1z_zmm_zmmm512b64_sae,
		Maxss_xmm_xmmm32,
		VEX_Vmaxss_xmm_xmm_xmmm32,
		EVEX_Vmaxss_xmm_k1z_xmm_xmmm32_sae,
		Maxsd_xmm_xmmm64,
		VEX_Vmaxsd_xmm_xmm_xmmm64,
		EVEX_Vmaxsd_xmm_k1z_xmm_xmmm64_sae,
		Punpcklbw_mm_mmm32,
		Punpcklbw_xmm_xmmm128,
		VEX_Vpunpcklbw_xmm_xmm_xmmm128,
		VEX_Vpunpcklbw_ymm_ymm_ymmm256,
		EVEX_Vpunpcklbw_xmm_k1z_xmm_xmmm128,
		EVEX_Vpunpcklbw_ymm_k1z_ymm_ymmm256,
		EVEX_Vpunpcklbw_zmm_k1z_zmm_zmmm512,
		Punpcklwd_mm_mmm32,
		Punpcklwd_xmm_xmmm128,
		VEX_Vpunpcklwd_xmm_xmm_xmmm128,
		VEX_Vpunpcklwd_ymm_ymm_ymmm256,
		EVEX_Vpunpcklwd_xmm_k1z_xmm_xmmm128,
		EVEX_Vpunpcklwd_ymm_k1z_ymm_ymmm256,
		EVEX_Vpunpcklwd_zmm_k1z_zmm_zmmm512,
		Punpckldq_mm_mmm32,
		Punpckldq_xmm_xmmm128,
		VEX_Vpunpckldq_xmm_xmm_xmmm128,
		VEX_Vpunpckldq_ymm_ymm_ymmm256,
		EVEX_Vpunpckldq_xmm_k1z_xmm_xmmm128b32,
		EVEX_Vpunpckldq_ymm_k1z_ymm_ymmm256b32,
		EVEX_Vpunpckldq_zmm_k1z_zmm_zmmm512b32,
		Packsswb_mm_mmm64,
		Packsswb_xmm_xmmm128,
		VEX_Vpacksswb_xmm_xmm_xmmm128,
		VEX_Vpacksswb_ymm_ymm_ymmm256,
		EVEX_Vpacksswb_xmm_k1z_xmm_xmmm128,
		EVEX_Vpacksswb_ymm_k1z_ymm_ymmm256,
		EVEX_Vpacksswb_zmm_k1z_zmm_zmmm512,
		Pcmpgtb_mm_mmm64,
		Pcmpgtb_xmm_xmmm128,
		VEX_Vpcmpgtb_xmm_xmm_xmmm128,
		VEX_Vpcmpgtb_ymm_ymm_ymmm256,
		EVEX_Vpcmpgtb_kr_k1_xmm_xmmm128,
		EVEX_Vpcmpgtb_kr_k1_ymm_ymmm256,
		EVEX_Vpcmpgtb_kr_k1_zmm_zmmm512,
		Pcmpgtw_mm_mmm64,
		Pcmpgtw_xmm_xmmm128,
		VEX_Vpcmpgtw_xmm_xmm_xmmm128,
		VEX_Vpcmpgtw_ymm_ymm_ymmm256,
		EVEX_Vpcmpgtw_kr_k1_xmm_xmmm128,
		EVEX_Vpcmpgtw_kr_k1_ymm_ymmm256,
		EVEX_Vpcmpgtw_kr_k1_zmm_zmmm512,
		Pcmpgtd_mm_mmm64,
		Pcmpgtd_xmm_xmmm128,
		VEX_Vpcmpgtd_xmm_xmm_xmmm128,
		VEX_Vpcmpgtd_ymm_ymm_ymmm256,
		EVEX_Vpcmpgtd_kr_k1_xmm_xmmm128b32,
		EVEX_Vpcmpgtd_kr_k1_ymm_ymmm256b32,
		EVEX_Vpcmpgtd_kr_k1_zmm_zmmm512b32,
		Packuswb_mm_mmm64,
		Packuswb_xmm_xmmm128,
		VEX_Vpackuswb_xmm_xmm_xmmm128,
		VEX_Vpackuswb_ymm_ymm_ymmm256,
		EVEX_Vpackuswb_xmm_k1z_xmm_xmmm128,
		EVEX_Vpackuswb_ymm_k1z_ymm_ymmm256,
		EVEX_Vpackuswb_zmm_k1z_zmm_zmmm512,
		Punpckhbw_mm_mmm64,
		Punpckhbw_xmm_xmmm128,
		VEX_Vpunpckhbw_xmm_xmm_xmmm128,
		VEX_Vpunpckhbw_ymm_ymm_ymmm256,
		EVEX_Vpunpckhbw_xmm_k1z_xmm_xmmm128,
		EVEX_Vpunpckhbw_ymm_k1z_ymm_ymmm256,
		EVEX_Vpunpckhbw_zmm_k1z_zmm_zmmm512,
		Punpckhwd_mm_mmm64,
		Punpckhwd_xmm_xmmm128,
		VEX_Vpunpckhwd_xmm_xmm_xmmm128,
		VEX_Vpunpckhwd_ymm_ymm_ymmm256,
		EVEX_Vpunpckhwd_xmm_k1z_xmm_xmmm128,
		EVEX_Vpunpckhwd_ymm_k1z_ymm_ymmm256,
		EVEX_Vpunpckhwd_zmm_k1z_zmm_zmmm512,
		Punpckhdq_mm_mmm64,
		Punpckhdq_xmm_xmmm128,
		VEX_Vpunpckhdq_xmm_xmm_xmmm128,
		VEX_Vpunpckhdq_ymm_ymm_ymmm256,
		EVEX_Vpunpckhdq_xmm_k1z_xmm_xmmm128b32,
		EVEX_Vpunpckhdq_ymm_k1z_ymm_ymmm256b32,
		EVEX_Vpunpckhdq_zmm_k1z_zmm_zmmm512b32,
		Packssdw_mm_mmm64,
		Packssdw_xmm_xmmm128,
		VEX_Vpackssdw_xmm_xmm_xmmm128,
		VEX_Vpackssdw_ymm_ymm_ymmm256,
		EVEX_Vpackssdw_xmm_k1z_xmm_xmmm128b32,
		EVEX_Vpackssdw_ymm_k1z_ymm_ymmm256b32,
		EVEX_Vpackssdw_zmm_k1z_zmm_zmmm512b32,
		Punpcklqdq_xmm_xmmm128,
		VEX_Vpunpcklqdq_xmm_xmm_xmmm128,
		VEX_Vpunpcklqdq_ymm_ymm_ymmm256,
		EVEX_Vpunpcklqdq_xmm_k1z_xmm_xmmm128b64,
		EVEX_Vpunpcklqdq_ymm_k1z_ymm_ymmm256b64,
		EVEX_Vpunpcklqdq_zmm_k1z_zmm_zmmm512b64,
		Punpckhqdq_xmm_xmmm128,
		VEX_Vpunpckhqdq_xmm_xmm_xmmm128,
		VEX_Vpunpckhqdq_ymm_ymm_ymmm256,
		EVEX_Vpunpckhqdq_xmm_k1z_xmm_xmmm128b64,
		EVEX_Vpunpckhqdq_ymm_k1z_ymm_ymmm256b64,
		EVEX_Vpunpckhqdq_zmm_k1z_zmm_zmmm512b64,
		Movd_mm_rm32,
		Movq_mm_rm64,
		Movd_xmm_rm32,
		Movq_xmm_rm64,
		VEX_Vmovd_xmm_rm32,
		VEX_Vmovq_xmm_rm64,
		EVEX_Vmovd_xmm_rm32,
		EVEX_Vmovq_xmm_rm64,
		Movq_mm_mmm64,
		Movdqa_xmm_xmmm128,
		VEX_Vmovdqa_xmm_xmmm128,
		VEX_Vmovdqa_ymm_ymmm256,
		EVEX_Vmovdqa32_xmm_k1z_xmmm128,
		EVEX_Vmovdqa32_ymm_k1z_ymmm256,
		EVEX_Vmovdqa32_zmm_k1z_zmmm512,
		EVEX_Vmovdqa64_xmm_k1z_xmmm128,
		EVEX_Vmovdqa64_ymm_k1z_ymmm256,
		EVEX_Vmovdqa64_zmm_k1z_zmmm512,
		Movdqu_xmm_xmmm128,
		VEX_Vmovdqu_xmm_xmmm128,
		VEX_Vmovdqu_ymm_ymmm256,
		EVEX_Vmovdqu32_xmm_k1z_xmmm128,
		EVEX_Vmovdqu32_ymm_k1z_ymmm256,
		EVEX_Vmovdqu32_zmm_k1z_zmmm512,
		EVEX_Vmovdqu64_xmm_k1z_xmmm128,
		EVEX_Vmovdqu64_ymm_k1z_ymmm256,
		EVEX_Vmovdqu64_zmm_k1z_zmmm512,
		EVEX_Vmovdqu8_xmm_k1z_xmmm128,
		EVEX_Vmovdqu8_ymm_k1z_ymmm256,
		EVEX_Vmovdqu8_zmm_k1z_zmmm512,
		EVEX_Vmovdqu16_xmm_k1z_xmmm128,
		EVEX_Vmovdqu16_ymm_k1z_ymmm256,
		EVEX_Vmovdqu16_zmm_k1z_zmmm512,
		Pshufw_mm_mmm64_imm8,
		Pshufd_xmm_xmmm128_imm8,
		VEX_Vpshufd_xmm_xmmm128_imm8,
		VEX_Vpshufd_ymm_ymmm256_imm8,
		EVEX_Vpshufd_xmm_k1z_xmmm128b32_imm8,
		EVEX_Vpshufd_ymm_k1z_ymmm256b32_imm8,
		EVEX_Vpshufd_zmm_k1z_zmmm512b32_imm8,
		Pshufhw_xmm_xmmm128_imm8,
		VEX_Vpshufhw_xmm_xmmm128_imm8,
		VEX_Vpshufhw_ymm_ymmm256_imm8,
		EVEX_Vpshufhw_xmm_k1z_xmmm128_imm8,
		EVEX_Vpshufhw_ymm_k1z_ymmm256_imm8,
		EVEX_Vpshufhw_zmm_k1z_zmmm512_imm8,
		Pshuflw_xmm_xmmm128_imm8,
		VEX_Vpshuflw_xmm_xmmm128_imm8,
		VEX_Vpshuflw_ymm_ymmm256_imm8,
		EVEX_Vpshuflw_xmm_k1z_xmmm128_imm8,
		EVEX_Vpshuflw_ymm_k1z_ymmm256_imm8,
		EVEX_Vpshuflw_zmm_k1z_zmmm512_imm8,
		Psrlw_mm_imm8,
		Psrlw_xmm_imm8,
		VEX_Vpsrlw_xmm_xmm_imm8,
		VEX_Vpsrlw_ymm_ymm_imm8,
		EVEX_Vpsrlw_xmm_k1z_xmmm128_imm8,
		EVEX_Vpsrlw_ymm_k1z_ymmm256_imm8,
		EVEX_Vpsrlw_zmm_k1z_zmmm512_imm8,
		Psraw_mm_imm8,
		Psraw_xmm_imm8,
		VEX_Vpsraw_xmm_xmm_imm8,
		VEX_Vpsraw_ymm_ymm_imm8,
		EVEX_Vpsraw_xmm_k1z_xmmm128_imm8,
		EVEX_Vpsraw_ymm_k1z_ymmm256_imm8,
		EVEX_Vpsraw_zmm_k1z_zmmm512_imm8,
		Psllw_mm_imm8,
		Psllw_xmm_imm8,
		VEX_Vpsllw_xmm_xmm_imm8,
		VEX_Vpsllw_ymm_ymm_imm8,
		EVEX_Vpsllw_xmm_k1z_xmmm128_imm8,
		EVEX_Vpsllw_ymm_k1z_ymmm256_imm8,
		EVEX_Vpsllw_zmm_k1z_zmmm512_imm8,
		EVEX_Vprord_xmm_k1z_xmmm128b32_imm8,
		EVEX_Vprord_ymm_k1z_ymmm256b32_imm8,
		EVEX_Vprord_zmm_k1z_zmmm512b32_imm8,
		EVEX_Vprorq_xmm_k1z_xmmm128b64_imm8,
		EVEX_Vprorq_ymm_k1z_ymmm256b64_imm8,
		EVEX_Vprorq_zmm_k1z_zmmm512b64_imm8,
		EVEX_Vprold_xmm_k1z_xmmm128b32_imm8,
		EVEX_Vprold_ymm_k1z_ymmm256b32_imm8,
		EVEX_Vprold_zmm_k1z_zmmm512b32_imm8,
		EVEX_Vprolq_xmm_k1z_xmmm128b64_imm8,
		EVEX_Vprolq_ymm_k1z_ymmm256b64_imm8,
		EVEX_Vprolq_zmm_k1z_zmmm512b64_imm8,
		Psrld_mm_imm8,
		Psrld_xmm_imm8,
		VEX_Vpsrld_xmm_xmm_imm8,
		VEX_Vpsrld_ymm_ymm_imm8,
		EVEX_Vpsrld_xmm_k1z_xmmm128b32_imm8,
		EVEX_Vpsrld_ymm_k1z_ymmm256b32_imm8,
		EVEX_Vpsrld_zmm_k1z_zmmm512b32_imm8,
		Psrad_mm_imm8,
		Psrad_xmm_imm8,
		VEX_Vpsrad_xmm_xmm_imm8,
		VEX_Vpsrad_ymm_ymm_imm8,
		EVEX_Vpsrad_xmm_k1z_xmmm128b32_imm8,
		EVEX_Vpsrad_ymm_k1z_ymmm256b32_imm8,
		EVEX_Vpsrad_zmm_k1z_zmmm512b32_imm8,
		EVEX_Vpsraq_xmm_k1z_xmmm128b64_imm8,
		EVEX_Vpsraq_ymm_k1z_ymmm256b64_imm8,
		EVEX_Vpsraq_zmm_k1z_zmmm512b64_imm8,
		Pslld_mm_imm8,
		Pslld_xmm_imm8,
		VEX_Vpslld_xmm_xmm_imm8,
		VEX_Vpslld_ymm_ymm_imm8,
		EVEX_Vpslld_xmm_k1z_xmmm128b32_imm8,
		EVEX_Vpslld_ymm_k1z_ymmm256b32_imm8,
		EVEX_Vpslld_zmm_k1z_zmmm512b32_imm8,
		Psrlq_mm_imm8,
		Psrlq_xmm_imm8,
		VEX_Vpsrlq_xmm_xmm_imm8,
		VEX_Vpsrlq_ymm_ymm_imm8,
		EVEX_Vpsrlq_xmm_k1z_xmmm128b64_imm8,
		EVEX_Vpsrlq_ymm_k1z_ymmm256b64_imm8,
		EVEX_Vpsrlq_zmm_k1z_zmmm512b64_imm8,
		Psrldq_xmm_imm8,
		VEX_Vpsrldq_xmm_xmm_imm8,
		VEX_Vpsrldq_ymm_ymm_imm8,
		EVEX_Vpsrldq_xmm_xmmm128_imm8,
		EVEX_Vpsrldq_ymm_ymmm256_imm8,
		EVEX_Vpsrldq_zmm_zmmm512_imm8,
		Psllq_mm_imm8,
		Psllq_xmm_imm8,
		VEX_Vpsllq_xmm_xmm_imm8,
		VEX_Vpsllq_ymm_ymm_imm8,
		EVEX_Vpsllq_xmm_k1z_xmmm128b64_imm8,
		EVEX_Vpsllq_ymm_k1z_ymmm256b64_imm8,
		EVEX_Vpsllq_zmm_k1z_zmmm512b64_imm8,
		Pslldq_xmm_imm8,
		VEX_Vpslldq_xmm_xmm_imm8,
		VEX_Vpslldq_ymm_ymm_imm8,
		EVEX_Vpslldq_xmm_xmmm128_imm8,
		EVEX_Vpslldq_ymm_ymmm256_imm8,
		EVEX_Vpslldq_zmm_zmmm512_imm8,
		Pcmpeqb_mm_mmm64,
		Pcmpeqb_xmm_xmmm128,
		VEX_Vpcmpeqb_xmm_xmm_xmmm128,
		VEX_Vpcmpeqb_ymm_ymm_ymmm256,
		EVEX_Vpcmpeqb_kr_k1_xmm_xmmm128,
		EVEX_Vpcmpeqb_kr_k1_ymm_ymmm256,
		EVEX_Vpcmpeqb_kr_k1_zmm_zmmm512,
		Pcmpeqw_mm_mmm64,
		Pcmpeqw_xmm_xmmm128,
		VEX_Vpcmpeqw_xmm_xmm_xmmm128,
		VEX_Vpcmpeqw_ymm_ymm_ymmm256,
		EVEX_Vpcmpeqw_kr_k1_xmm_xmmm128,
		EVEX_Vpcmpeqw_kr_k1_ymm_ymmm256,
		EVEX_Vpcmpeqw_kr_k1_zmm_zmmm512,
		Pcmpeqd_mm_mmm64,
		Pcmpeqd_xmm_xmmm128,
		VEX_Vpcmpeqd_xmm_xmm_xmmm128,
		VEX_Vpcmpeqd_ymm_ymm_ymmm256,
		EVEX_Vpcmpeqd_kr_k1_xmm_xmmm128b32,
		EVEX_Vpcmpeqd_kr_k1_ymm_ymmm256b32,
		EVEX_Vpcmpeqd_kr_k1_zmm_zmmm512b32,
		Emms,
		VEX_Vzeroupper,
		VEX_Vzeroall,
		Vmread_rm32_r32,
		Vmread_rm64_r64,
		EVEX_Vcvttps2udq_xmm_k1z_xmmm128b32,
		EVEX_Vcvttps2udq_ymm_k1z_ymmm256b32,
		EVEX_Vcvttps2udq_zmm_k1z_zmmm512b32_sae,
		EVEX_Vcvttpd2udq_xmm_k1z_xmmm128b64,
		EVEX_Vcvttpd2udq_xmm_k1z_ymmm256b64,
		EVEX_Vcvttpd2udq_ymm_k1z_zmmm512b64_sae,
		Extrq_xmm_imm8_imm8,
		EVEX_Vcvttps2uqq_xmm_k1z_xmmm64b32,
		EVEX_Vcvttps2uqq_ymm_k1z_xmmm128b32,
		EVEX_Vcvttps2uqq_zmm_k1z_ymmm256b32_sae,
		EVEX_Vcvttpd2uqq_xmm_k1z_xmmm128b64,
		EVEX_Vcvttpd2uqq_ymm_k1z_ymmm256b64,
		EVEX_Vcvttpd2uqq_zmm_k1z_zmmm512b64_sae,
		EVEX_Vcvttss2usi_r32_xmmm32_sae,
		EVEX_Vcvttss2usi_r64_xmmm32_sae,
		Insertq_xmm_xmm_imm8_imm8,
		EVEX_Vcvttsd2usi_r32_xmmm64_sae,
		EVEX_Vcvttsd2usi_r64_xmmm64_sae,
		Vmwrite_r32_rm32,
		Vmwrite_r64_rm64,
		EVEX_Vcvtps2udq_xmm_k1z_xmmm128b32,
		EVEX_Vcvtps2udq_ymm_k1z_ymmm256b32,
		EVEX_Vcvtps2udq_zmm_k1z_zmmm512b32_er,
		EVEX_Vcvtpd2udq_xmm_k1z_xmmm128b64,
		EVEX_Vcvtpd2udq_xmm_k1z_ymmm256b64,
		EVEX_Vcvtpd2udq_ymm_k1z_zmmm512b64_er,
		Extrq_xmm_xmm,
		EVEX_Vcvtps2uqq_xmm_k1z_xmmm64b32,
		EVEX_Vcvtps2uqq_ymm_k1z_xmmm128b32,
		EVEX_Vcvtps2uqq_zmm_k1z_ymmm256b32_er,
		EVEX_Vcvtpd2uqq_xmm_k1z_xmmm128b64,
		EVEX_Vcvtpd2uqq_ymm_k1z_ymmm256b64,
		EVEX_Vcvtpd2uqq_zmm_k1z_zmmm512b64_er,
		EVEX_Vcvtss2usi_r32_xmmm32_er,
		EVEX_Vcvtss2usi_r64_xmmm32_er,
		Insertq_xmm_xmm,
		EVEX_Vcvtsd2usi_r32_xmmm64_er,
		EVEX_Vcvtsd2usi_r64_xmmm64_er,
		EVEX_Vcvttps2qq_xmm_k1z_xmmm64b32,
		EVEX_Vcvttps2qq_ymm_k1z_xmmm128b32,
		EVEX_Vcvttps2qq_zmm_k1z_ymmm256b32_sae,
		EVEX_Vcvttpd2qq_xmm_k1z_xmmm128b64,
		EVEX_Vcvttpd2qq_ymm_k1z_ymmm256b64,
		EVEX_Vcvttpd2qq_zmm_k1z_zmmm512b64_sae,
		EVEX_Vcvtudq2pd_xmm_k1z_xmmm64b32,
		EVEX_Vcvtudq2pd_ymm_k1z_xmmm128b32,
		EVEX_Vcvtudq2pd_zmm_k1z_ymmm256b32_er,
		EVEX_Vcvtuqq2pd_xmm_k1z_xmmm128b64,
		EVEX_Vcvtuqq2pd_ymm_k1z_ymmm256b64,
		EVEX_Vcvtuqq2pd_zmm_k1z_zmmm512b64_er,
		EVEX_Vcvtudq2ps_xmm_k1z_xmmm128b32,
		EVEX_Vcvtudq2ps_ymm_k1z_ymmm256b32,
		EVEX_Vcvtudq2ps_zmm_k1z_zmmm512b32_er,
		EVEX_Vcvtuqq2ps_xmm_k1z_xmmm128b64,
		EVEX_Vcvtuqq2ps_xmm_k1z_ymmm256b64,
		EVEX_Vcvtuqq2ps_ymm_k1z_zmmm512b64_er,
		EVEX_Vcvtps2qq_xmm_k1z_xmmm64b32,
		EVEX_Vcvtps2qq_ymm_k1z_xmmm128b32,
		EVEX_Vcvtps2qq_zmm_k1z_ymmm256b32_er,
		EVEX_Vcvtpd2qq_xmm_k1z_xmmm128b64,
		EVEX_Vcvtpd2qq_ymm_k1z_ymmm256b64,
		EVEX_Vcvtpd2qq_zmm_k1z_zmmm512b64_er,
		EVEX_Vcvtusi2ss_xmm_xmm_rm32_er,
		EVEX_Vcvtusi2ss_xmm_xmm_rm64_er,
		EVEX_Vcvtusi2sd_xmm_xmm_rm32_er,
		EVEX_Vcvtusi2sd_xmm_xmm_rm64_er,
		Haddpd_xmm_xmmm128,
		VEX_Vhaddpd_xmm_xmm_xmmm128,
		VEX_Vhaddpd_ymm_ymm_ymmm256,
		Haddps_xmm_xmmm128,
		VEX_Vhaddps_xmm_xmm_xmmm128,
		VEX_Vhaddps_ymm_ymm_ymmm256,
		Hsubpd_xmm_xmmm128,
		VEX_Vhsubpd_xmm_xmm_xmmm128,
		VEX_Vhsubpd_ymm_ymm_ymmm256,
		Hsubps_xmm_xmmm128,
		VEX_Vhsubps_xmm_xmm_xmmm128,
		VEX_Vhsubps_ymm_ymm_ymmm256,
		Movd_rm32_mm,
		Movq_rm64_mm,
		Movd_rm32_xmm,
		Movq_rm64_xmm,
		VEX_Vmovd_rm32_xmm,
		VEX_Vmovq_rm64_xmm,
		EVEX_Vmovd_rm32_xmm,
		EVEX_Vmovq_rm64_xmm,
		Movq_xmm_xmmm64,
		VEX_Vmovq_xmm_xmmm64,
		EVEX_Vmovq_xmm_xmmm64,
		Movq_mmm64_mm,
		Movdqa_xmmm128_xmm,
		VEX_Vmovdqa_xmmm128_xmm,
		VEX_Vmovdqa_ymmm256_ymm,
		EVEX_Vmovdqa32_xmmm128_k1z_xmm,
		EVEX_Vmovdqa32_ymmm256_k1z_ymm,
		EVEX_Vmovdqa32_zmmm512_k1z_zmm,
		EVEX_Vmovdqa64_xmmm128_k1z_xmm,
		EVEX_Vmovdqa64_ymmm256_k1z_ymm,
		EVEX_Vmovdqa64_zmmm512_k1z_zmm,
		Movdqu_xmmm128_xmm,
		VEX_Vmovdqu_xmmm128_xmm,
		VEX_Vmovdqu_ymmm256_ymm,
		EVEX_Vmovdqu32_xmmm128_k1z_xmm,
		EVEX_Vmovdqu32_ymmm256_k1z_ymm,
		EVEX_Vmovdqu32_zmmm512_k1z_zmm,
		EVEX_Vmovdqu64_xmmm128_k1z_xmm,
		EVEX_Vmovdqu64_ymmm256_k1z_ymm,
		EVEX_Vmovdqu64_zmmm512_k1z_zmm,
		EVEX_Vmovdqu8_xmmm128_k1z_xmm,
		EVEX_Vmovdqu8_ymmm256_k1z_ymm,
		EVEX_Vmovdqu8_zmmm512_k1z_zmm,
		EVEX_Vmovdqu16_xmmm128_k1z_xmm,
		EVEX_Vmovdqu16_ymmm256_k1z_ymm,
		EVEX_Vmovdqu16_zmmm512_k1z_zmm,
		Jo_rel16,
		Jo_rel32_32,
		Jo_rel32_64,
		Jno_rel16,
		Jno_rel32_32,
		Jno_rel32_64,
		Jb_rel16,
		Jb_rel32_32,
		Jb_rel32_64,
		Jae_rel16,
		Jae_rel32_32,
		Jae_rel32_64,
		Je_rel16,
		Je_rel32_32,
		Je_rel32_64,
		Jne_rel16,
		Jne_rel32_32,
		Jne_rel32_64,
		Jbe_rel16,
		Jbe_rel32_32,
		Jbe_rel32_64,
		Ja_rel16,
		Ja_rel32_32,
		Ja_rel32_64,
		Js_rel16,
		Js_rel32_32,
		Js_rel32_64,
		Jns_rel16,
		Jns_rel32_32,
		Jns_rel32_64,
		Jp_rel16,
		Jp_rel32_32,
		Jp_rel32_64,
		Jnp_rel16,
		Jnp_rel32_32,
		Jnp_rel32_64,
		Jl_rel16,
		Jl_rel32_32,
		Jl_rel32_64,
		Jge_rel16,
		Jge_rel32_32,
		Jge_rel32_64,
		Jle_rel16,
		Jle_rel32_32,
		Jle_rel32_64,
		Jg_rel16,
		Jg_rel32_32,
		Jg_rel32_64,
		Seto_rm8,
		Setno_rm8,
		Setb_rm8,
		Setae_rm8,
		Sete_rm8,
		Setne_rm8,
		Setbe_rm8,
		Seta_rm8,
		Sets_rm8,
		Setns_rm8,
		Setp_rm8,
		Setnp_rm8,
		Setl_rm8,
		Setge_rm8,
		Setle_rm8,
		Setg_rm8,
		VEX_Kmovw_kr_km16,
		VEX_Kmovq_kr_km64,
		VEX_Kmovb_kr_km8,
		VEX_Kmovd_kr_km32,
		VEX_Kmovw_m16_kr,
		VEX_Kmovq_m64_kr,
		VEX_Kmovb_m8_kr,
		VEX_Kmovd_m32_kr,
		VEX_Kmovw_kr_r32,
		VEX_Kmovb_kr_r32,
		VEX_Kmovd_kr_r32,
		VEX_Kmovq_kr_r64,
		VEX_Kmovw_r32_kr,
		VEX_Kmovb_r32_kr,
		VEX_Kmovd_r32_kr,
		VEX_Kmovq_r64_kr,
		VEX_Kortestw_kr_kr,
		VEX_Kortestq_kr_kr,
		VEX_Kortestb_kr_kr,
		VEX_Kortestd_kr_kr,
		VEX_Ktestw_kr_kr,
		VEX_Ktestq_kr_kr,
		VEX_Ktestb_kr_kr,
		VEX_Ktestd_kr_kr,
		Pushw_FS,
		Pushd_FS,
		Pushq_FS,
		Popw_FS,
		Popd_FS,
		Popq_FS,
		Cpuid,
		Bt_rm16_r16,
		Bt_rm32_r32,
		Bt_rm64_r64,
		Shld_rm16_r16_imm8,
		Shld_rm32_r32_imm8,
		Shld_rm64_r64_imm8,
		Shld_rm16_r16_CL,
		Shld_rm32_r32_CL,
		Shld_rm64_r64_CL,
		Montmul_16,
		Montmul_32,
		Montmul_64,
		Xsha1_16,
		Xsha1_32,
		Xsha1_64,
		Xsha256_16,
		Xsha256_32,
		Xsha256_64,
		Xbts_r16_rm16,
		Xbts_r32_rm32,
		Xstore_16,
		Xstore_32,
		Xstore_64,
		Xcryptecb_16,
		Xcryptecb_32,
		Xcryptecb_64,
		Xcryptcbc_16,
		Xcryptcbc_32,
		Xcryptcbc_64,
		Xcryptctr_16,
		Xcryptctr_32,
		Xcryptctr_64,
		Xcryptcfb_16,
		Xcryptcfb_32,
		Xcryptcfb_64,
		Xcryptofb_16,
		Xcryptofb_32,
		Xcryptofb_64,
		Ibts_rm16_r16,
		Ibts_rm32_r32,
		Cmpxchg486_rm8_r8,
		Cmpxchg486_rm16_r16,
		Cmpxchg486_rm32_r32,
		Pushw_GS,
		Pushd_GS,
		Pushq_GS,
		Popw_GS,
		Popd_GS,
		Popq_GS,
		Rsm,
		Bts_rm16_r16,
		Bts_rm32_r32,
		Bts_rm64_r64,
		Shrd_rm16_r16_imm8,
		Shrd_rm32_r32_imm8,
		Shrd_rm64_r64_imm8,
		Shrd_rm16_r16_CL,
		Shrd_rm32_r32_CL,
		Shrd_rm64_r64_CL,
		Fxsave_m512byte,
		Fxsave64_m512byte,
		Rdfsbase_r32,
		Rdfsbase_r64,
		Fxrstor_m512byte,
		Fxrstor64_m512byte,
		Rdgsbase_r32,
		Rdgsbase_r64,
		Ldmxcsr_m32,
		Wrfsbase_r32,
		Wrfsbase_r64,
		VEX_Vldmxcsr_m32,
		Stmxcsr_m32,
		Wrgsbase_r32,
		Wrgsbase_r64,
		VEX_Vstmxcsr_m32,
		Xsave_mem,
		Xsave64_mem,
		Ptwrite_rm32,
		Ptwrite_rm64,
		Xrstor_mem,
		Xrstor64_mem,
		Incsspd_r32,
		Incsspq_r64,
		Xsaveopt_mem,
		Xsaveopt64_mem,
		Clwb_m8,
		Tpause_r32,
		Tpause_r64,
		Clrssbsy_m64,
		Umonitor_r16,
		Umonitor_r32,
		Umonitor_r64,
		Umwait_r32,
		Umwait_r64,
		Clflush_m8,
		Clflushopt_m8,
		Lfence,
		Lfence_E9,
		Lfence_EA,
		Lfence_EB,
		Lfence_EC,
		Lfence_ED,
		Lfence_EE,
		Lfence_EF,
		Mfence,
		Mfence_F1,
		Mfence_F2,
		Mfence_F3,
		Mfence_F4,
		Mfence_F5,
		Mfence_F6,
		Mfence_F7,
		Sfence,
		Sfence_F9,
		Sfence_FA,
		Sfence_FB,
		Sfence_FC,
		Sfence_FD,
		Sfence_FE,
		Sfence_FF,
		Pcommit,
		Imul_r16_rm16,
		Imul_r32_rm32,
		Imul_r64_rm64,
		Cmpxchg_rm8_r8,
		Cmpxchg_rm16_r16,
		Cmpxchg_rm32_r32,
		Cmpxchg_rm64_r64,
		Lss_r16_m1616,
		Lss_r32_m1632,
		Lss_r64_m1664,
		Btr_rm16_r16,
		Btr_rm32_r32,
		Btr_rm64_r64,
		Lfs_r16_m1616,
		Lfs_r32_m1632,
		Lfs_r64_m1664,
		Lgs_r16_m1616,
		Lgs_r32_m1632,
		Lgs_r64_m1664,
		Movzx_r16_rm8,
		Movzx_r32_rm8,
		Movzx_r64_rm8,
		Movzx_r16_rm16,
		Movzx_r32_rm16,
		Movzx_r64_rm16,
		Jmpe_disp16,
		Jmpe_disp32,
		Popcnt_r16_rm16,
		Popcnt_r32_rm32,
		Popcnt_r64_rm64,
		Ud1_r16_rm16,
		Ud1_r32_rm32,
		Ud1_r64_rm64,
		Bt_rm16_imm8,
		Bt_rm32_imm8,
		Bt_rm64_imm8,
		Bts_rm16_imm8,
		Bts_rm32_imm8,
		Bts_rm64_imm8,
		Btr_rm16_imm8,
		Btr_rm32_imm8,
		Btr_rm64_imm8,
		Btc_rm16_imm8,
		Btc_rm32_imm8,
		Btc_rm64_imm8,
		Btc_rm16_r16,
		Btc_rm32_r32,
		Btc_rm64_r64,
		Bsf_r16_rm16,
		Bsf_r32_rm32,
		Bsf_r64_rm64,
		Tzcnt_r16_rm16,
		Tzcnt_r32_rm32,
		Tzcnt_r64_rm64,
		Bsr_r16_rm16,
		Bsr_r32_rm32,
		Bsr_r64_rm64,
		Lzcnt_r16_rm16,
		Lzcnt_r32_rm32,
		Lzcnt_r64_rm64,
		Movsx_r16_rm8,
		Movsx_r32_rm8,
		Movsx_r64_rm8,
		Movsx_r16_rm16,
		Movsx_r32_rm16,
		Movsx_r64_rm16,
		Xadd_rm8_r8,
		Xadd_rm16_r16,
		Xadd_rm32_r32,
		Xadd_rm64_r64,
		Cmpps_xmm_xmmm128_imm8,
		VEX_Vcmpps_xmm_xmm_xmmm128_imm8,
		VEX_Vcmpps_ymm_ymm_ymmm256_imm8,
		EVEX_Vcmpps_kr_k1_xmm_xmmm128b32_imm8,
		EVEX_Vcmpps_kr_k1_ymm_ymmm256b32_imm8,
		EVEX_Vcmpps_kr_k1_zmm_zmmm512b32_imm8_sae,
		Cmppd_xmm_xmmm128_imm8,
		VEX_Vcmppd_xmm_xmm_xmmm128_imm8,
		VEX_Vcmppd_ymm_ymm_ymmm256_imm8,
		EVEX_Vcmppd_kr_k1_xmm_xmmm128b64_imm8,
		EVEX_Vcmppd_kr_k1_ymm_ymmm256b64_imm8,
		EVEX_Vcmppd_kr_k1_zmm_zmmm512b64_imm8_sae,
		Cmpss_xmm_xmmm32_imm8,
		VEX_Vcmpss_xmm_xmm_xmmm32_imm8,
		EVEX_Vcmpss_kr_k1_xmm_xmmm32_imm8_sae,
		Cmpsd_xmm_xmmm64_imm8,
		VEX_Vcmpsd_xmm_xmm_xmmm64_imm8,
		EVEX_Vcmpsd_kr_k1_xmm_xmmm64_imm8_sae,
		Movnti_m32_r32,
		Movnti_m64_r64,
		Pinsrw_mm_r32m16_imm8,
		Pinsrw_mm_r64m16_imm8,
		Pinsrw_xmm_r32m16_imm8,
		Pinsrw_xmm_r64m16_imm8,
		VEX_Vpinsrw_xmm_xmm_r32m16_imm8,
		VEX_Vpinsrw_xmm_xmm_r64m16_imm8,
		EVEX_Vpinsrw_xmm_xmm_r32m16_imm8,
		EVEX_Vpinsrw_xmm_xmm_r64m16_imm8,
		Pextrw_r32_mm_imm8,
		Pextrw_r64_mm_imm8,
		Pextrw_r32_xmm_imm8,
		Pextrw_r64_xmm_imm8,
		VEX_Vpextrw_r32_xmm_imm8,
		VEX_Vpextrw_r64_xmm_imm8,
		EVEX_Vpextrw_r32_xmm_imm8,
		EVEX_Vpextrw_r64_xmm_imm8,
		Shufps_xmm_xmmm128_imm8,
		VEX_Vshufps_xmm_xmm_xmmm128_imm8,
		VEX_Vshufps_ymm_ymm_ymmm256_imm8,
		EVEX_Vshufps_xmm_k1z_xmm_xmmm128b32_imm8,
		EVEX_Vshufps_ymm_k1z_ymm_ymmm256b32_imm8,
		EVEX_Vshufps_zmm_k1z_zmm_zmmm512b32_imm8,
		Shufpd_xmm_xmmm128_imm8,
		VEX_Vshufpd_xmm_xmm_xmmm128_imm8,
		VEX_Vshufpd_ymm_ymm_ymmm256_imm8,
		EVEX_Vshufpd_xmm_k1z_xmm_xmmm128b64_imm8,
		EVEX_Vshufpd_ymm_k1z_ymm_ymmm256b64_imm8,
		EVEX_Vshufpd_zmm_k1z_zmm_zmmm512b64_imm8,
		Cmpxchg8b_m64,
		Cmpxchg16b_m128,
		Xrstors_mem,
		Xrstors64_mem,
		Xsavec_mem,
		Xsavec64_mem,
		Xsaves_mem,
		Xsaves64_mem,
		Vmptrld_m64,
		Vmclear_m64,
		Vmxon_m64,
		Rdrand_r16,
		Rdrand_r32,
		Rdrand_r64,
		Vmptrst_m64,
		Rdseed_r16,
		Rdseed_r32,
		Rdseed_r64,
		Rdpid_r32,
		Rdpid_r64,
		Bswap_r16,
		Bswap_r32,
		Bswap_r64,
		Addsubpd_xmm_xmmm128,
		VEX_Vaddsubpd_xmm_xmm_xmmm128,
		VEX_Vaddsubpd_ymm_ymm_ymmm256,
		Addsubps_xmm_xmmm128,
		VEX_Vaddsubps_xmm_xmm_xmmm128,
		VEX_Vaddsubps_ymm_ymm_ymmm256,
		Psrlw_mm_mmm64,
		Psrlw_xmm_xmmm128,
		VEX_Vpsrlw_xmm_xmm_xmmm128,
		VEX_Vpsrlw_ymm_ymm_xmmm128,
		EVEX_Vpsrlw_xmm_k1z_xmm_xmmm128,
		EVEX_Vpsrlw_ymm_k1z_ymm_xmmm128,
		EVEX_Vpsrlw_zmm_k1z_zmm_xmmm128,
		Psrld_mm_mmm64,
		Psrld_xmm_xmmm128,
		VEX_Vpsrld_xmm_xmm_xmmm128,
		VEX_Vpsrld_ymm_ymm_xmmm128,
		EVEX_Vpsrld_xmm_k1z_xmm_xmmm128,
		EVEX_Vpsrld_ymm_k1z_ymm_xmmm128,
		EVEX_Vpsrld_zmm_k1z_zmm_xmmm128,
		Psrlq_mm_mmm64,
		Psrlq_xmm_xmmm128,
		VEX_Vpsrlq_xmm_xmm_xmmm128,
		VEX_Vpsrlq_ymm_ymm_xmmm128,
		EVEX_Vpsrlq_xmm_k1z_xmm_xmmm128,
		EVEX_Vpsrlq_ymm_k1z_ymm_xmmm128,
		EVEX_Vpsrlq_zmm_k1z_zmm_xmmm128,
		Paddq_mm_mmm64,
		Paddq_xmm_xmmm128,
		VEX_Vpaddq_xmm_xmm_xmmm128,
		VEX_Vpaddq_ymm_ymm_ymmm256,
		EVEX_Vpaddq_xmm_k1z_xmm_xmmm128b64,
		EVEX_Vpaddq_ymm_k1z_ymm_ymmm256b64,
		EVEX_Vpaddq_zmm_k1z_zmm_zmmm512b64,
		Pmullw_mm_mmm64,
		Pmullw_xmm_xmmm128,
		VEX_Vpmullw_xmm_xmm_xmmm128,
		VEX_Vpmullw_ymm_ymm_ymmm256,
		EVEX_Vpmullw_xmm_k1z_xmm_xmmm128,
		EVEX_Vpmullw_ymm_k1z_ymm_ymmm256,
		EVEX_Vpmullw_zmm_k1z_zmm_zmmm512,
		Movq_xmmm64_xmm,
		VEX_Vmovq_xmmm64_xmm,
		EVEX_Vmovq_xmmm64_xmm,
		Movq2dq_xmm_mm,
		Movdq2q_mm_xmm,
		Pmovmskb_r32_mm,
		Pmovmskb_r64_mm,
		Pmovmskb_r32_xmm,
		Pmovmskb_r64_xmm,
		VEX_Vpmovmskb_r32_xmm,
		VEX_Vpmovmskb_r64_xmm,
		VEX_Vpmovmskb_r32_ymm,
		VEX_Vpmovmskb_r64_ymm,
		Psubusb_mm_mmm64,
		Psubusb_xmm_xmmm128,
		VEX_Vpsubusb_xmm_xmm_xmmm128,
		VEX_Vpsubusb_ymm_ymm_ymmm256,
		EVEX_Vpsubusb_xmm_k1z_xmm_xmmm128,
		EVEX_Vpsubusb_ymm_k1z_ymm_ymmm256,
		EVEX_Vpsubusb_zmm_k1z_zmm_zmmm512,
		Psubusw_mm_mmm64,
		Psubusw_xmm_xmmm128,
		VEX_Vpsubusw_xmm_xmm_xmmm128,
		VEX_Vpsubusw_ymm_ymm_ymmm256,
		EVEX_Vpsubusw_xmm_k1z_xmm_xmmm128,
		EVEX_Vpsubusw_ymm_k1z_ymm_ymmm256,
		EVEX_Vpsubusw_zmm_k1z_zmm_zmmm512,
		Pminub_mm_mmm64,
		Pminub_xmm_xmmm128,
		VEX_Vpminub_xmm_xmm_xmmm128,
		VEX_Vpminub_ymm_ymm_ymmm256,
		EVEX_Vpminub_xmm_k1z_xmm_xmmm128,
		EVEX_Vpminub_ymm_k1z_ymm_ymmm256,
		EVEX_Vpminub_zmm_k1z_zmm_zmmm512,
		Pand_mm_mmm64,
		Pand_xmm_xmmm128,
		VEX_Vpand_xmm_xmm_xmmm128,
		VEX_Vpand_ymm_ymm_ymmm256,
		EVEX_Vpandd_xmm_k1z_xmm_xmmm128b32,
		EVEX_Vpandd_ymm_k1z_ymm_ymmm256b32,
		EVEX_Vpandd_zmm_k1z_zmm_zmmm512b32,
		EVEX_Vpandq_xmm_k1z_xmm_xmmm128b64,
		EVEX_Vpandq_ymm_k1z_ymm_ymmm256b64,
		EVEX_Vpandq_zmm_k1z_zmm_zmmm512b64,
		Paddusb_mm_mmm64,
		Paddusb_xmm_xmmm128,
		VEX_Vpaddusb_xmm_xmm_xmmm128,
		VEX_Vpaddusb_ymm_ymm_ymmm256,
		EVEX_Vpaddusb_xmm_k1z_xmm_xmmm128,
		EVEX_Vpaddusb_ymm_k1z_ymm_ymmm256,
		EVEX_Vpaddusb_zmm_k1z_zmm_zmmm512,
		Paddusw_mm_mmm64,
		Paddusw_xmm_xmmm128,
		VEX_Vpaddusw_xmm_xmm_xmmm128,
		VEX_Vpaddusw_ymm_ymm_ymmm256,
		EVEX_Vpaddusw_xmm_k1z_xmm_xmmm128,
		EVEX_Vpaddusw_ymm_k1z_ymm_ymmm256,
		EVEX_Vpaddusw_zmm_k1z_zmm_zmmm512,
		Pmaxub_mm_mmm64,
		Pmaxub_xmm_xmmm128,
		VEX_Vpmaxub_xmm_xmm_xmmm128,
		VEX_Vpmaxub_ymm_ymm_ymmm256,
		EVEX_Vpmaxub_xmm_k1z_xmm_xmmm128,
		EVEX_Vpmaxub_ymm_k1z_ymm_ymmm256,
		EVEX_Vpmaxub_zmm_k1z_zmm_zmmm512,
		Pandn_mm_mmm64,
		Pandn_xmm_xmmm128,
		VEX_Vpandn_xmm_xmm_xmmm128,
		VEX_Vpandn_ymm_ymm_ymmm256,
		EVEX_Vpandnd_xmm_k1z_xmm_xmmm128b32,
		EVEX_Vpandnd_ymm_k1z_ymm_ymmm256b32,
		EVEX_Vpandnd_zmm_k1z_zmm_zmmm512b32,
		EVEX_Vpandnq_xmm_k1z_xmm_xmmm128b64,
		EVEX_Vpandnq_ymm_k1z_ymm_ymmm256b64,
		EVEX_Vpandnq_zmm_k1z_zmm_zmmm512b64,
		Pavgb_mm_mmm64,
		Pavgb_xmm_xmmm128,
		VEX_Vpavgb_xmm_xmm_xmmm128,
		VEX_Vpavgb_ymm_ymm_ymmm256,
		EVEX_Vpavgb_xmm_k1z_xmm_xmmm128,
		EVEX_Vpavgb_ymm_k1z_ymm_ymmm256,
		EVEX_Vpavgb_zmm_k1z_zmm_zmmm512,
		Psraw_mm_mmm64,
		Psraw_xmm_xmmm128,
		VEX_Vpsraw_xmm_xmm_xmmm128,
		VEX_Vpsraw_ymm_ymm_xmmm128,
		EVEX_Vpsraw_xmm_k1z_xmm_xmmm128,
		EVEX_Vpsraw_ymm_k1z_ymm_xmmm128,
		EVEX_Vpsraw_zmm_k1z_zmm_xmmm128,
		Psrad_mm_mmm64,
		Psrad_xmm_xmmm128,
		VEX_Vpsrad_xmm_xmm_xmmm128,
		VEX_Vpsrad_ymm_ymm_xmmm128,
		EVEX_Vpsrad_xmm_k1z_xmm_xmmm128,
		EVEX_Vpsrad_ymm_k1z_ymm_xmmm128,
		EVEX_Vpsrad_zmm_k1z_zmm_xmmm128,
		EVEX_Vpsraq_xmm_k1z_xmm_xmmm128,
		EVEX_Vpsraq_ymm_k1z_ymm_xmmm128,
		EVEX_Vpsraq_zmm_k1z_zmm_xmmm128,
		Pavgw_mm_mmm64,
		Pavgw_xmm_xmmm128,
		VEX_Vpavgw_xmm_xmm_xmmm128,
		VEX_Vpavgw_ymm_ymm_ymmm256,
		EVEX_Vpavgw_xmm_k1z_xmm_xmmm128,
		EVEX_Vpavgw_ymm_k1z_ymm_ymmm256,
		EVEX_Vpavgw_zmm_k1z_zmm_zmmm512,
		Pmulhuw_mm_mmm64,
		Pmulhuw_xmm_xmmm128,
		VEX_Vpmulhuw_xmm_xmm_xmmm128,
		VEX_Vpmulhuw_ymm_ymm_ymmm256,
		EVEX_Vpmulhuw_xmm_k1z_xmm_xmmm128,
		EVEX_Vpmulhuw_ymm_k1z_ymm_ymmm256,
		EVEX_Vpmulhuw_zmm_k1z_zmm_zmmm512,
		Pmulhw_mm_mmm64,
		Pmulhw_xmm_xmmm128,
		VEX_Vpmulhw_xmm_xmm_xmmm128,
		VEX_Vpmulhw_ymm_ymm_ymmm256,
		EVEX_Vpmulhw_xmm_k1z_xmm_xmmm128,
		EVEX_Vpmulhw_ymm_k1z_ymm_ymmm256,
		EVEX_Vpmulhw_zmm_k1z_zmm_zmmm512,
		Cvttpd2dq_xmm_xmmm128,
		VEX_Vcvttpd2dq_xmm_xmmm128,
		VEX_Vcvttpd2dq_xmm_ymmm256,
		EVEX_Vcvttpd2dq_xmm_k1z_xmmm128b64,
		EVEX_Vcvttpd2dq_xmm_k1z_ymmm256b64,
		EVEX_Vcvttpd2dq_ymm_k1z_zmmm512b64_sae,
		Cvtdq2pd_xmm_xmmm64,
		VEX_Vcvtdq2pd_xmm_xmmm64,
		VEX_Vcvtdq2pd_ymm_xmmm128,
		EVEX_Vcvtdq2pd_xmm_k1z_xmmm64b32,
		EVEX_Vcvtdq2pd_ymm_k1z_xmmm128b32,
		EVEX_Vcvtdq2pd_zmm_k1z_ymmm256b32_er,
		EVEX_Vcvtqq2pd_xmm_k1z_xmmm128b64,
		EVEX_Vcvtqq2pd_ymm_k1z_ymmm256b64,
		EVEX_Vcvtqq2pd_zmm_k1z_zmmm512b64_er,
		Cvtpd2dq_xmm_xmmm128,
		VEX_Vcvtpd2dq_xmm_xmmm128,
		VEX_Vcvtpd2dq_xmm_ymmm256,
		EVEX_Vcvtpd2dq_xmm_k1z_xmmm128b64,
		EVEX_Vcvtpd2dq_xmm_k1z_ymmm256b64,
		EVEX_Vcvtpd2dq_ymm_k1z_zmmm512b64_er,
		Movntq_m64_mm,
		Movntdq_m128_xmm,
		VEX_Vmovntdq_m128_xmm,
		VEX_Vmovntdq_m256_ymm,
		EVEX_Vmovntdq_m128_xmm,
		EVEX_Vmovntdq_m256_ymm,
		EVEX_Vmovntdq_m512_zmm,
		Psubsb_mm_mmm64,
		Psubsb_xmm_xmmm128,
		VEX_Vpsubsb_xmm_xmm_xmmm128,
		VEX_Vpsubsb_ymm_ymm_ymmm256,
		EVEX_Vpsubsb_xmm_k1z_xmm_xmmm128,
		EVEX_Vpsubsb_ymm_k1z_ymm_ymmm256,
		EVEX_Vpsubsb_zmm_k1z_zmm_zmmm512,
		Psubsw_mm_mmm64,
		Psubsw_xmm_xmmm128,
		VEX_Vpsubsw_xmm_xmm_xmmm128,
		VEX_Vpsubsw_ymm_ymm_ymmm256,
		EVEX_Vpsubsw_xmm_k1z_xmm_xmmm128,
		EVEX_Vpsubsw_ymm_k1z_ymm_ymmm256,
		EVEX_Vpsubsw_zmm_k1z_zmm_zmmm512,
		Pminsw_mm_mmm64,
		Pminsw_xmm_xmmm128,
		VEX_Vpminsw_xmm_xmm_xmmm128,
		VEX_Vpminsw_ymm_ymm_ymmm256,
		EVEX_Vpminsw_xmm_k1z_xmm_xmmm128,
		EVEX_Vpminsw_ymm_k1z_ymm_ymmm256,
		EVEX_Vpminsw_zmm_k1z_zmm_zmmm512,
		Por_mm_mmm64,
		Por_xmm_xmmm128,
		VEX_Vpor_xmm_xmm_xmmm128,
		VEX_Vpor_ymm_ymm_ymmm256,
		EVEX_Vpord_xmm_k1z_xmm_xmmm128b32,
		EVEX_Vpord_ymm_k1z_ymm_ymmm256b32,
		EVEX_Vpord_zmm_k1z_zmm_zmmm512b32,
		EVEX_Vporq_xmm_k1z_xmm_xmmm128b64,
		EVEX_Vporq_ymm_k1z_ymm_ymmm256b64,
		EVEX_Vporq_zmm_k1z_zmm_zmmm512b64,
		Paddsb_mm_mmm64,
		Paddsb_xmm_xmmm128,
		VEX_Vpaddsb_xmm_xmm_xmmm128,
		VEX_Vpaddsb_ymm_ymm_ymmm256,
		EVEX_Vpaddsb_xmm_k1z_xmm_xmmm128,
		EVEX_Vpaddsb_ymm_k1z_ymm_ymmm256,
		EVEX_Vpaddsb_zmm_k1z_zmm_zmmm512,
		Paddsw_mm_mmm64,
		Paddsw_xmm_xmmm128,
		VEX_Vpaddsw_xmm_xmm_xmmm128,
		VEX_Vpaddsw_ymm_ymm_ymmm256,
		EVEX_Vpaddsw_xmm_k1z_xmm_xmmm128,
		EVEX_Vpaddsw_ymm_k1z_ymm_ymmm256,
		EVEX_Vpaddsw_zmm_k1z_zmm_zmmm512,
		Pmaxsw_mm_mmm64,
		Pmaxsw_xmm_xmmm128,
		VEX_Vpmaxsw_xmm_xmm_xmmm128,
		VEX_Vpmaxsw_ymm_ymm_ymmm256,
		EVEX_Vpmaxsw_xmm_k1z_xmm_xmmm128,
		EVEX_Vpmaxsw_ymm_k1z_ymm_ymmm256,
		EVEX_Vpmaxsw_zmm_k1z_zmm_zmmm512,
		Pxor_mm_mmm64,
		Pxor_xmm_xmmm128,
		VEX_Vpxor_xmm_xmm_xmmm128,
		VEX_Vpxor_ymm_ymm_ymmm256,
		EVEX_Vpxord_xmm_k1z_xmm_xmmm128b32,
		EVEX_Vpxord_ymm_k1z_ymm_ymmm256b32,
		EVEX_Vpxord_zmm_k1z_zmm_zmmm512b32,
		EVEX_Vpxorq_xmm_k1z_xmm_xmmm128b64,
		EVEX_Vpxorq_ymm_k1z_ymm_ymmm256b64,
		EVEX_Vpxorq_zmm_k1z_zmm_zmmm512b64,
		Lddqu_xmm_m128,
		VEX_Vlddqu_xmm_m128,
		VEX_Vlddqu_ymm_m256,
		Psllw_mm_mmm64,
		Psllw_xmm_xmmm128,
		VEX_Vpsllw_xmm_xmm_xmmm128,
		VEX_Vpsllw_ymm_ymm_xmmm128,
		EVEX_Vpsllw_xmm_k1z_xmm_xmmm128,
		EVEX_Vpsllw_ymm_k1z_ymm_xmmm128,
		EVEX_Vpsllw_zmm_k1z_zmm_xmmm128,
		Pslld_mm_mmm64,
		Pslld_xmm_xmmm128,
		VEX_Vpslld_xmm_xmm_xmmm128,
		VEX_Vpslld_ymm_ymm_xmmm128,
		EVEX_Vpslld_xmm_k1z_xmm_xmmm128,
		EVEX_Vpslld_ymm_k1z_ymm_xmmm128,
		EVEX_Vpslld_zmm_k1z_zmm_xmmm128,
		Psllq_mm_mmm64,
		Psllq_xmm_xmmm128,
		VEX_Vpsllq_xmm_xmm_xmmm128,
		VEX_Vpsllq_ymm_ymm_xmmm128,
		EVEX_Vpsllq_xmm_k1z_xmm_xmmm128,
		EVEX_Vpsllq_ymm_k1z_ymm_xmmm128,
		EVEX_Vpsllq_zmm_k1z_zmm_xmmm128,
		Pmuludq_mm_mmm64,
		Pmuludq_xmm_xmmm128,
		VEX_Vpmuludq_xmm_xmm_xmmm128,
		VEX_Vpmuludq_ymm_ymm_ymmm256,
		EVEX_Vpmuludq_xmm_k1z_xmm_xmmm128b64,
		EVEX_Vpmuludq_ymm_k1z_ymm_ymmm256b64,
		EVEX_Vpmuludq_zmm_k1z_zmm_zmmm512b64,
		Pmaddwd_mm_mmm64,
		Pmaddwd_xmm_xmmm128,
		VEX_Vpmaddwd_xmm_xmm_xmmm128,
		VEX_Vpmaddwd_ymm_ymm_ymmm256,
		EVEX_Vpmaddwd_xmm_k1z_xmm_xmmm128,
		EVEX_Vpmaddwd_ymm_k1z_ymm_ymmm256,
		EVEX_Vpmaddwd_zmm_k1z_zmm_zmmm512,
		Psadbw_mm_mmm64,
		Psadbw_xmm_xmmm128,
		VEX_Vpsadbw_xmm_xmm_xmmm128,
		VEX_Vpsadbw_ymm_ymm_ymmm256,
		EVEX_Vpsadbw_xmm_xmm_xmmm128,
		EVEX_Vpsadbw_ymm_ymm_ymmm256,
		EVEX_Vpsadbw_zmm_zmm_zmmm512,
		Maskmovq_rDI_mm_mm,
		Maskmovdqu_rDI_xmm_xmm,
		VEX_Vmaskmovdqu_rDI_xmm_xmm,
		Psubb_mm_mmm64,
		Psubb_xmm_xmmm128,
		VEX_Vpsubb_xmm_xmm_xmmm128,
		VEX_Vpsubb_ymm_ymm_ymmm256,
		EVEX_Vpsubb_xmm_k1z_xmm_xmmm128,
		EVEX_Vpsubb_ymm_k1z_ymm_ymmm256,
		EVEX_Vpsubb_zmm_k1z_zmm_zmmm512,
		Psubw_mm_mmm64,
		Psubw_xmm_xmmm128,
		VEX_Vpsubw_xmm_xmm_xmmm128,
		VEX_Vpsubw_ymm_ymm_ymmm256,
		EVEX_Vpsubw_xmm_k1z_xmm_xmmm128,
		EVEX_Vpsubw_ymm_k1z_ymm_ymmm256,
		EVEX_Vpsubw_zmm_k1z_zmm_zmmm512,
		Psubd_mm_mmm64,
		Psubd_xmm_xmmm128,
		VEX_Vpsubd_xmm_xmm_xmmm128,
		VEX_Vpsubd_ymm_ymm_ymmm256,
		EVEX_Vpsubd_xmm_k1z_xmm_xmmm128b32,
		EVEX_Vpsubd_ymm_k1z_ymm_ymmm256b32,
		EVEX_Vpsubd_zmm_k1z_zmm_zmmm512b32,
		Psubq_mm_mmm64,
		Psubq_xmm_xmmm128,
		VEX_Vpsubq_xmm_xmm_xmmm128,
		VEX_Vpsubq_ymm_ymm_ymmm256,
		EVEX_Vpsubq_xmm_k1z_xmm_xmmm128b64,
		EVEX_Vpsubq_ymm_k1z_ymm_ymmm256b64,
		EVEX_Vpsubq_zmm_k1z_zmm_zmmm512b64,
		Paddb_mm_mmm64,
		Paddb_xmm_xmmm128,
		VEX_Vpaddb_xmm_xmm_xmmm128,
		VEX_Vpaddb_ymm_ymm_ymmm256,
		EVEX_Vpaddb_xmm_k1z_xmm_xmmm128,
		EVEX_Vpaddb_ymm_k1z_ymm_ymmm256,
		EVEX_Vpaddb_zmm_k1z_zmm_zmmm512,
		Paddw_mm_mmm64,
		Paddw_xmm_xmmm128,
		VEX_Vpaddw_xmm_xmm_xmmm128,
		VEX_Vpaddw_ymm_ymm_ymmm256,
		EVEX_Vpaddw_xmm_k1z_xmm_xmmm128,
		EVEX_Vpaddw_ymm_k1z_ymm_ymmm256,
		EVEX_Vpaddw_zmm_k1z_zmm_zmmm512,
		Paddd_mm_mmm64,
		Paddd_xmm_xmmm128,
		VEX_Vpaddd_xmm_xmm_xmmm128,
		VEX_Vpaddd_ymm_ymm_ymmm256,
		EVEX_Vpaddd_xmm_k1z_xmm_xmmm128b32,
		EVEX_Vpaddd_ymm_k1z_ymm_ymmm256b32,
		EVEX_Vpaddd_zmm_k1z_zmm_zmmm512b32,
		Ud0_r16_rm16,
		Ud0_r32_rm32,
		Ud0_r64_rm64,
		Pshufb_mm_mmm64,
		Pshufb_xmm_xmmm128,
		VEX_Vpshufb_xmm_xmm_xmmm128,
		VEX_Vpshufb_ymm_ymm_ymmm256,
		EVEX_Vpshufb_xmm_k1z_xmm_xmmm128,
		EVEX_Vpshufb_ymm_k1z_ymm_ymmm256,
		EVEX_Vpshufb_zmm_k1z_zmm_zmmm512,
		Phaddw_mm_mmm64,
		Phaddw_xmm_xmmm128,
		VEX_Vphaddw_xmm_xmm_xmmm128,
		VEX_Vphaddw_ymm_ymm_ymmm256,
		Phaddd_mm_mmm64,
		Phaddd_xmm_xmmm128,
		VEX_Vphaddd_xmm_xmm_xmmm128,
		VEX_Vphaddd_ymm_ymm_ymmm256,
		Phaddsw_mm_mmm64,
		Phaddsw_xmm_xmmm128,
		VEX_Vphaddsw_xmm_xmm_xmmm128,
		VEX_Vphaddsw_ymm_ymm_ymmm256,
		Pmaddubsw_mm_mmm64,
		Pmaddubsw_xmm_xmmm128,
		VEX_Vpmaddubsw_xmm_xmm_xmmm128,
		VEX_Vpmaddubsw_ymm_ymm_ymmm256,
		EVEX_Vpmaddubsw_xmm_k1z_xmm_xmmm128,
		EVEX_Vpmaddubsw_ymm_k1z_ymm_ymmm256,
		EVEX_Vpmaddubsw_zmm_k1z_zmm_zmmm512,
		Phsubw_mm_mmm64,
		Phsubw_xmm_xmmm128,
		VEX_Vphsubw_xmm_xmm_xmmm128,
		VEX_Vphsubw_ymm_ymm_ymmm256,
		Phsubd_mm_mmm64,
		Phsubd_xmm_xmmm128,
		VEX_Vphsubd_xmm_xmm_xmmm128,
		VEX_Vphsubd_ymm_ymm_ymmm256,
		Phsubsw_mm_mmm64,
		Phsubsw_xmm_xmmm128,
		VEX_Vphsubsw_xmm_xmm_xmmm128,
		VEX_Vphsubsw_ymm_ymm_ymmm256,
		Psignb_mm_mmm64,
		Psignb_xmm_xmmm128,
		VEX_Vpsignb_xmm_xmm_xmmm128,
		VEX_Vpsignb_ymm_ymm_ymmm256,
		Psignw_mm_mmm64,
		Psignw_xmm_xmmm128,
		VEX_Vpsignw_xmm_xmm_xmmm128,
		VEX_Vpsignw_ymm_ymm_ymmm256,
		Psignd_mm_mmm64,
		Psignd_xmm_xmmm128,
		VEX_Vpsignd_xmm_xmm_xmmm128,
		VEX_Vpsignd_ymm_ymm_ymmm256,
		Pmulhrsw_mm_mmm64,
		Pmulhrsw_xmm_xmmm128,
		VEX_Vpmulhrsw_xmm_xmm_xmmm128,
		VEX_Vpmulhrsw_ymm_ymm_ymmm256,
		EVEX_Vpmulhrsw_xmm_k1z_xmm_xmmm128,
		EVEX_Vpmulhrsw_ymm_k1z_ymm_ymmm256,
		EVEX_Vpmulhrsw_zmm_k1z_zmm_zmmm512,
		VEX_Vpermilps_xmm_xmm_xmmm128,
		VEX_Vpermilps_ymm_ymm_ymmm256,
		EVEX_Vpermilps_xmm_k1z_xmm_xmmm128b32,
		EVEX_Vpermilps_ymm_k1z_ymm_ymmm256b32,
		EVEX_Vpermilps_zmm_k1z_zmm_zmmm512b32,
		VEX_Vpermilpd_xmm_xmm_xmmm128,
		VEX_Vpermilpd_ymm_ymm_ymmm256,
		EVEX_Vpermilpd_xmm_k1z_xmm_xmmm128b64,
		EVEX_Vpermilpd_ymm_k1z_ymm_ymmm256b64,
		EVEX_Vpermilpd_zmm_k1z_zmm_zmmm512b64,
		VEX_Vtestps_xmm_xmmm128,
		VEX_Vtestps_ymm_ymmm256,
		VEX_Vtestpd_xmm_xmmm128,
		VEX_Vtestpd_ymm_ymmm256,
		Pblendvb_xmm_xmmm128,
		EVEX_Vpsrlvw_xmm_k1z_xmm_xmmm128,
		EVEX_Vpsrlvw_ymm_k1z_ymm_ymmm256,
		EVEX_Vpsrlvw_zmm_k1z_zmm_zmmm512,
		EVEX_Vpmovuswb_xmmm64_k1z_xmm,
		EVEX_Vpmovuswb_xmmm128_k1z_ymm,
		EVEX_Vpmovuswb_ymmm256_k1z_zmm,
		EVEX_Vpsravw_xmm_k1z_xmm_xmmm128,
		EVEX_Vpsravw_ymm_k1z_ymm_ymmm256,
		EVEX_Vpsravw_zmm_k1z_zmm_zmmm512,
		EVEX_Vpmovusdb_xmmm32_k1z_xmm,
		EVEX_Vpmovusdb_xmmm64_k1z_ymm,
		EVEX_Vpmovusdb_xmmm128_k1z_zmm,
		EVEX_Vpsllvw_xmm_k1z_xmm_xmmm128,
		EVEX_Vpsllvw_ymm_k1z_ymm_ymmm256,
		EVEX_Vpsllvw_zmm_k1z_zmm_zmmm512,
		EVEX_Vpmovusqb_xmmm16_k1z_xmm,
		EVEX_Vpmovusqb_xmmm32_k1z_ymm,
		EVEX_Vpmovusqb_xmmm64_k1z_zmm,
		VEX_Vcvtph2ps_xmm_xmmm64,
		VEX_Vcvtph2ps_ymm_xmmm128,
		EVEX_Vcvtph2ps_xmm_k1z_xmmm64,
		EVEX_Vcvtph2ps_ymm_k1z_xmmm128,
		EVEX_Vcvtph2ps_zmm_k1z_ymmm256_sae,
		EVEX_Vpmovusdw_xmmm64_k1z_xmm,
		EVEX_Vpmovusdw_xmmm128_k1z_ymm,
		EVEX_Vpmovusdw_ymmm256_k1z_zmm,
		Blendvps_xmm_xmmm128,
		EVEX_Vprorvd_xmm_k1z_xmm_xmmm128b32,
		EVEX_Vprorvd_ymm_k1z_ymm_ymmm256b32,
		EVEX_Vprorvd_zmm_k1z_zmm_zmmm512b32,
		EVEX_Vprorvq_xmm_k1z_xmm_xmmm128b64,
		EVEX_Vprorvq_ymm_k1z_ymm_ymmm256b64,
		EVEX_Vprorvq_zmm_k1z_zmm_zmmm512b64,
		EVEX_Vpmovusqw_xmmm32_k1z_xmm,
		EVEX_Vpmovusqw_xmmm64_k1z_ymm,
		EVEX_Vpmovusqw_xmmm128_k1z_zmm,
		Blendvpd_xmm_xmmm128,
		EVEX_Vprolvd_xmm_k1z_xmm_xmmm128b32,
		EVEX_Vprolvd_ymm_k1z_ymm_ymmm256b32,
		EVEX_Vprolvd_zmm_k1z_zmm_zmmm512b32,
		EVEX_Vprolvq_xmm_k1z_xmm_xmmm128b64,
		EVEX_Vprolvq_ymm_k1z_ymm_ymmm256b64,
		EVEX_Vprolvq_zmm_k1z_zmm_zmmm512b64,
		EVEX_Vpmovusqd_xmmm64_k1z_xmm,
		EVEX_Vpmovusqd_xmmm128_k1z_ymm,
		EVEX_Vpmovusqd_ymmm256_k1z_zmm,
		VEX_Vpermps_ymm_ymm_ymmm256,
		EVEX_Vpermps_ymm_k1z_ymm_ymmm256b32,
		EVEX_Vpermps_zmm_k1z_zmm_zmmm512b32,
		EVEX_Vpermpd_ymm_k1z_ymm_ymmm256b64,
		EVEX_Vpermpd_zmm_k1z_zmm_zmmm512b64,
		Ptest_xmm_xmmm128,
		VEX_Vptest_xmm_xmmm128,
		VEX_Vptest_ymm_ymmm256,
		VEX_Vbroadcastss_xmm_m32,
		VEX_Vbroadcastss_ymm_m32,
		EVEX_Vbroadcastss_xmm_k1z_xmmm32,
		EVEX_Vbroadcastss_ymm_k1z_xmmm32,
		EVEX_Vbroadcastss_zmm_k1z_xmmm32,
		VEX_Vbroadcastsd_ymm_m64,
		EVEX_Vbroadcastf32x2_ymm_k1z_xmmm64,
		EVEX_Vbroadcastf32x2_zmm_k1z_xmmm64,
		EVEX_Vbroadcastsd_ymm_k1z_xmmm64,
		EVEX_Vbroadcastsd_zmm_k1z_xmmm64,
		VEX_Vbroadcastf128_ymm_m128,
		EVEX_Vbroadcastf32x4_ymm_k1z_m128,
		EVEX_Vbroadcastf32x4_zmm_k1z_m128,
		EVEX_Vbroadcastf64x2_ymm_k1z_m128,
		EVEX_Vbroadcastf64x2_zmm_k1z_m128,
		EVEX_Vbroadcastf32x8_zmm_k1z_m256,
		EVEX_Vbroadcastf64x4_zmm_k1z_m256,
		Pabsb_mm_mmm64,
		Pabsb_xmm_xmmm128,
		VEX_Vpabsb_xmm_xmmm128,
		VEX_Vpabsb_ymm_ymmm256,
		EVEX_Vpabsb_xmm_k1z_xmmm128,
		EVEX_Vpabsb_ymm_k1z_ymmm256,
		EVEX_Vpabsb_zmm_k1z_zmmm512,
		Pabsw_mm_mmm64,
		Pabsw_xmm_xmmm128,
		VEX_Vpabsw_xmm_xmmm128,
		VEX_Vpabsw_ymm_ymmm256,
		EVEX_Vpabsw_xmm_k1z_xmmm128,
		EVEX_Vpabsw_ymm_k1z_ymmm256,
		EVEX_Vpabsw_zmm_k1z_zmmm512,
		Pabsd_mm_mmm64,
		Pabsd_xmm_xmmm128,
		VEX_Vpabsd_xmm_xmmm128,
		VEX_Vpabsd_ymm_ymmm256,
		EVEX_Vpabsd_xmm_k1z_xmmm128b32,
		EVEX_Vpabsd_ymm_k1z_ymmm256b32,
		EVEX_Vpabsd_zmm_k1z_zmmm512b32,
		EVEX_Vpabsq_xmm_k1z_xmmm128b64,
		EVEX_Vpabsq_ymm_k1z_ymmm256b64,
		EVEX_Vpabsq_zmm_k1z_zmmm512b64,
		Pmovsxbw_xmm_xmmm64,
		VEX_Vpmovsxbw_xmm_xmmm64,
		VEX_Vpmovsxbw_ymm_xmmm128,
		EVEX_Vpmovsxbw_xmm_k1z_xmmm64,
		EVEX_Vpmovsxbw_ymm_k1z_xmmm128,
		EVEX_Vpmovsxbw_zmm_k1z_ymmm256,
		EVEX_Vpmovswb_xmmm64_k1z_xmm,
		EVEX_Vpmovswb_xmmm128_k1z_ymm,
		EVEX_Vpmovswb_ymmm256_k1z_zmm,
		Pmovsxbd_xmm_xmmm32,
		VEX_Vpmovsxbd_xmm_xmmm32,
		VEX_Vpmovsxbd_ymm_xmmm64,
		EVEX_Vpmovsxbd_xmm_k1z_xmmm32,
		EVEX_Vpmovsxbd_ymm_k1z_xmmm64,
		EVEX_Vpmovsxbd_zmm_k1z_xmmm128,
		EVEX_Vpmovsdb_xmmm32_k1z_xmm,
		EVEX_Vpmovsdb_xmmm64_k1z_ymm,
		EVEX_Vpmovsdb_xmmm128_k1z_zmm,
		Pmovsxbq_xmm_xmmm16,
		VEX_Vpmovsxbq_xmm_xmmm16,
		VEX_Vpmovsxbq_ymm_xmmm32,
		EVEX_Vpmovsxbq_xmm_k1z_xmmm16,
		EVEX_Vpmovsxbq_ymm_k1z_xmmm32,
		EVEX_Vpmovsxbq_zmm_k1z_xmmm64,
		EVEX_Vpmovsqb_xmmm16_k1z_xmm,
		EVEX_Vpmovsqb_xmmm32_k1z_ymm,
		EVEX_Vpmovsqb_xmmm64_k1z_zmm,
		Pmovsxwd_xmm_xmmm64,
		VEX_Vpmovsxwd_xmm_xmmm64,
		VEX_Vpmovsxwd_ymm_xmmm128,
		EVEX_Vpmovsxwd_xmm_k1z_xmmm64,
		EVEX_Vpmovsxwd_ymm_k1z_xmmm128,
		EVEX_Vpmovsxwd_zmm_k1z_ymmm256,
		EVEX_Vpmovsdw_xmmm64_k1z_xmm,
		EVEX_Vpmovsdw_xmmm128_k1z_ymm,
		EVEX_Vpmovsdw_ymmm256_k1z_zmm,
		Pmovsxwq_xmm_xmmm32,
		VEX_Vpmovsxwq_xmm_xmmm32,
		VEX_Vpmovsxwq_ymm_xmmm64,
		EVEX_Vpmovsxwq_xmm_k1z_xmmm32,
		EVEX_Vpmovsxwq_ymm_k1z_xmmm64,
		EVEX_Vpmovsxwq_zmm_k1z_xmmm128,
		EVEX_Vpmovsqw_xmmm32_k1z_xmm,
		EVEX_Vpmovsqw_xmmm64_k1z_ymm,
		EVEX_Vpmovsqw_xmmm128_k1z_zmm,
		Pmovsxdq_xmm_xmmm64,
		VEX_Vpmovsxdq_xmm_xmmm64,
		VEX_Vpmovsxdq_ymm_xmmm128,
		EVEX_Vpmovsxdq_xmm_k1z_xmmm64,
		EVEX_Vpmovsxdq_ymm_k1z_xmmm128,
		EVEX_Vpmovsxdq_zmm_k1z_ymmm256,
		EVEX_Vpmovsqd_xmmm64_k1z_xmm,
		EVEX_Vpmovsqd_xmmm128_k1z_ymm,
		EVEX_Vpmovsqd_ymmm256_k1z_zmm,
		EVEX_Vptestmb_kr_k1_xmm_xmmm128,
		EVEX_Vptestmb_kr_k1_ymm_ymmm256,
		EVEX_Vptestmb_kr_k1_zmm_zmmm512,
		EVEX_Vptestmw_kr_k1_xmm_xmmm128,
		EVEX_Vptestmw_kr_k1_ymm_ymmm256,
		EVEX_Vptestmw_kr_k1_zmm_zmmm512,
		EVEX_Vptestnmb_kr_k1_xmm_xmmm128,
		EVEX_Vptestnmb_kr_k1_ymm_ymmm256,
		EVEX_Vptestnmb_kr_k1_zmm_zmmm512,
		EVEX_Vptestnmw_kr_k1_xmm_xmmm128,
		EVEX_Vptestnmw_kr_k1_ymm_ymmm256,
		EVEX_Vptestnmw_kr_k1_zmm_zmmm512,
		EVEX_Vptestmd_kr_k1_xmm_xmmm128b32,
		EVEX_Vptestmd_kr_k1_ymm_ymmm256b32,
		EVEX_Vptestmd_kr_k1_zmm_zmmm512b32,
		EVEX_Vptestmq_kr_k1_xmm_xmmm128b64,
		EVEX_Vptestmq_kr_k1_ymm_ymmm256b64,
		EVEX_Vptestmq_kr_k1_zmm_zmmm512b64,
		EVEX_Vptestnmd_kr_k1_xmm_xmmm128b32,
		EVEX_Vptestnmd_kr_k1_ymm_ymmm256b32,
		EVEX_Vptestnmd_kr_k1_zmm_zmmm512b32,
		EVEX_Vptestnmq_kr_k1_xmm_xmmm128b64,
		EVEX_Vptestnmq_kr_k1_ymm_ymmm256b64,
		EVEX_Vptestnmq_kr_k1_zmm_zmmm512b64,
		Pmuldq_xmm_xmmm128,
		VEX_Vpmuldq_xmm_xmm_xmmm128,
		VEX_Vpmuldq_ymm_ymm_ymmm256,
		EVEX_Vpmuldq_xmm_k1z_xmm_xmmm128b64,
		EVEX_Vpmuldq_ymm_k1z_ymm_ymmm256b64,
		EVEX_Vpmuldq_zmm_k1z_zmm_zmmm512b64,
		EVEX_Vpmovm2b_xmm_kr,
		EVEX_Vpmovm2b_ymm_kr,
		EVEX_Vpmovm2b_zmm_kr,
		EVEX_Vpmovm2w_xmm_kr,
		EVEX_Vpmovm2w_ymm_kr,
		EVEX_Vpmovm2w_zmm_kr,
		Pcmpeqq_xmm_xmmm128,
		VEX_Vpcmpeqq_xmm_xmm_xmmm128,
		VEX_Vpcmpeqq_ymm_ymm_ymmm256,
		EVEX_Vpcmpeqq_kr_k1_xmm_xmmm128b64,
		EVEX_Vpcmpeqq_kr_k1_ymm_ymmm256b64,
		EVEX_Vpcmpeqq_kr_k1_zmm_zmmm512b64,
		EVEX_Vpmovb2m_kr_xmm,
		EVEX_Vpmovb2m_kr_ymm,
		EVEX_Vpmovb2m_kr_zmm,
		EVEX_Vpmovw2m_kr_xmm,
		EVEX_Vpmovw2m_kr_ymm,
		EVEX_Vpmovw2m_kr_zmm,
		Movntdqa_xmm_m128,
		VEX_Vmovntdqa_xmm_m128,
		VEX_Vmovntdqa_ymm_m256,
		EVEX_Vmovntdqa_xmm_m128,
		EVEX_Vmovntdqa_ymm_m256,
		EVEX_Vmovntdqa_zmm_m512,
		EVEX_Vpbroadcastmb2q_xmm_kr,
		EVEX_Vpbroadcastmb2q_ymm_kr,
		EVEX_Vpbroadcastmb2q_zmm_kr,
		Packusdw_xmm_xmmm128,
		VEX_Vpackusdw_xmm_xmm_xmmm128,
		VEX_Vpackusdw_ymm_ymm_ymmm256,
		EVEX_Vpackusdw_xmm_k1z_xmm_xmmm128b32,
		EVEX_Vpackusdw_ymm_k1z_ymm_ymmm256b32,
		EVEX_Vpackusdw_zmm_k1z_zmm_zmmm512b32,
		VEX_Vmaskmovps_xmm_xmm_m128,
		VEX_Vmaskmovps_ymm_ymm_m256,
		EVEX_Vscalefps_xmm_k1z_xmm_xmmm128b32,
		EVEX_Vscalefps_ymm_k1z_ymm_ymmm256b32,
		EVEX_Vscalefps_zmm_k1z_zmm_zmmm512b32_er,
		EVEX_Vscalefpd_xmm_k1z_xmm_xmmm128b64,
		EVEX_Vscalefpd_ymm_k1z_ymm_ymmm256b64,
		EVEX_Vscalefpd_zmm_k1z_zmm_zmmm512b64_er,
		VEX_Vmaskmovpd_xmm_xmm_m128,
		VEX_Vmaskmovpd_ymm_ymm_m256,
		EVEX_Vscalefss_xmm_k1z_xmm_xmmm32_er,
		EVEX_Vscalefsd_xmm_k1z_xmm_xmmm64_er,
		VEX_Vmaskmovps_m128_xmm_xmm,
		VEX_Vmaskmovps_m256_ymm_ymm,
		VEX_Vmaskmovpd_m128_xmm_xmm,
		VEX_Vmaskmovpd_m256_ymm_ymm,
		Pmovzxbw_xmm_xmmm64,
		VEX_Vpmovzxbw_xmm_xmmm64,
		VEX_Vpmovzxbw_ymm_xmmm128,
		EVEX_Vpmovzxbw_xmm_k1z_xmmm64,
		EVEX_Vpmovzxbw_ymm_k1z_xmmm128,
		EVEX_Vpmovzxbw_zmm_k1z_ymmm256,
		EVEX_Vpmovwb_xmmm64_k1z_xmm,
		EVEX_Vpmovwb_xmmm128_k1z_ymm,
		EVEX_Vpmovwb_ymmm256_k1z_zmm,
		Pmovzxbd_xmm_xmmm32,
		VEX_Vpmovzxbd_xmm_xmmm32,
		VEX_Vpmovzxbd_ymm_xmmm64,
		EVEX_Vpmovzxbd_xmm_k1z_xmmm32,
		EVEX_Vpmovzxbd_ymm_k1z_xmmm64,
		EVEX_Vpmovzxbd_zmm_k1z_xmmm128,
		EVEX_Vpmovdb_xmmm32_k1z_xmm,
		EVEX_Vpmovdb_xmmm64_k1z_ymm,
		EVEX_Vpmovdb_xmmm128_k1z_zmm,
		Pmovzxbq_xmm_xmmm16,
		VEX_Vpmovzxbq_xmm_xmmm16,
		VEX_Vpmovzxbq_ymm_xmmm32,
		EVEX_Vpmovzxbq_xmm_k1z_xmmm16,
		EVEX_Vpmovzxbq_ymm_k1z_xmmm32,
		EVEX_Vpmovzxbq_zmm_k1z_xmmm64,
		EVEX_Vpmovqb_xmmm16_k1z_xmm,
		EVEX_Vpmovqb_xmmm32_k1z_ymm,
		EVEX_Vpmovqb_xmmm64_k1z_zmm,
		Pmovzxwd_xmm_xmmm64,
		VEX_Vpmovzxwd_xmm_xmmm64,
		VEX_Vpmovzxwd_ymm_xmmm128,
		EVEX_Vpmovzxwd_xmm_k1z_xmmm64,
		EVEX_Vpmovzxwd_ymm_k1z_xmmm128,
		EVEX_Vpmovzxwd_zmm_k1z_ymmm256,
		EVEX_Vpmovdw_xmmm64_k1z_xmm,
		EVEX_Vpmovdw_xmmm128_k1z_ymm,
		EVEX_Vpmovdw_ymmm256_k1z_zmm,
		Pmovzxwq_xmm_xmmm32,
		VEX_Vpmovzxwq_xmm_xmmm32,
		VEX_Vpmovzxwq_ymm_xmmm64,
		EVEX_Vpmovzxwq_xmm_k1z_xmmm32,
		EVEX_Vpmovzxwq_ymm_k1z_xmmm64,
		EVEX_Vpmovzxwq_zmm_k1z_xmmm128,
		EVEX_Vpmovqw_xmmm32_k1z_xmm,
		EVEX_Vpmovqw_xmmm64_k1z_ymm,
		EVEX_Vpmovqw_xmmm128_k1z_zmm,
		Pmovzxdq_xmm_xmmm64,
		VEX_Vpmovzxdq_xmm_xmmm64,
		VEX_Vpmovzxdq_ymm_xmmm128,
		EVEX_Vpmovzxdq_xmm_k1z_xmmm64,
		EVEX_Vpmovzxdq_ymm_k1z_xmmm128,
		EVEX_Vpmovzxdq_zmm_k1z_ymmm256,
		EVEX_Vpmovqd_xmmm64_k1z_xmm,
		EVEX_Vpmovqd_xmmm128_k1z_ymm,
		EVEX_Vpmovqd_ymmm256_k1z_zmm,
		VEX_Vpermd_ymm_ymm_ymmm256,
		EVEX_Vpermd_ymm_k1z_ymm_ymmm256b32,
		EVEX_Vpermd_zmm_k1z_zmm_zmmm512b32,
		EVEX_Vpermq_ymm_k1z_ymm_ymmm256b64,
		EVEX_Vpermq_zmm_k1z_zmm_zmmm512b64,
		Pcmpgtq_xmm_xmmm128,
		VEX_Vpcmpgtq_xmm_xmm_xmmm128,
		VEX_Vpcmpgtq_ymm_ymm_ymmm256,
		EVEX_Vpcmpgtq_kr_k1_xmm_xmmm128b64,
		EVEX_Vpcmpgtq_kr_k1_ymm_ymmm256b64,
		EVEX_Vpcmpgtq_kr_k1_zmm_zmmm512b64,
		Pminsb_xmm_xmmm128,
		VEX_Vpminsb_xmm_xmm_xmmm128,
		VEX_Vpminsb_ymm_ymm_ymmm256,
		EVEX_Vpminsb_xmm_k1z_xmm_xmmm128,
		EVEX_Vpminsb_ymm_k1z_ymm_ymmm256,
		EVEX_Vpminsb_zmm_k1z_zmm_zmmm512,
		EVEX_Vpmovm2d_xmm_kr,
		EVEX_Vpmovm2d_ymm_kr,
		EVEX_Vpmovm2d_zmm_kr,
		EVEX_Vpmovm2q_xmm_kr,
		EVEX_Vpmovm2q_ymm_kr,
		EVEX_Vpmovm2q_zmm_kr,
		Pminsd_xmm_xmmm128,
		VEX_Vpminsd_xmm_xmm_xmmm128,
		VEX_Vpminsd_ymm_ymm_ymmm256,
		EVEX_Vpminsd_xmm_k1z_xmm_xmmm128b32,
		EVEX_Vpminsd_ymm_k1z_ymm_ymmm256b32,
		EVEX_Vpminsd_zmm_k1z_zmm_zmmm512b32,
		EVEX_Vpminsq_xmm_k1z_xmm_xmmm128b64,
		EVEX_Vpminsq_ymm_k1z_ymm_ymmm256b64,
		EVEX_Vpminsq_zmm_k1z_zmm_zmmm512b64,
		EVEX_Vpmovd2m_kr_xmm,
		EVEX_Vpmovd2m_kr_ymm,
		EVEX_Vpmovd2m_kr_zmm,
		EVEX_Vpmovq2m_kr_xmm,
		EVEX_Vpmovq2m_kr_ymm,
		EVEX_Vpmovq2m_kr_zmm,
		Pminuw_xmm_xmmm128,
		VEX_Vpminuw_xmm_xmm_xmmm128,
		VEX_Vpminuw_ymm_ymm_ymmm256,
		EVEX_Vpminuw_xmm_k1z_xmm_xmmm128,
		EVEX_Vpminuw_ymm_k1z_ymm_ymmm256,
		EVEX_Vpminuw_zmm_k1z_zmm_zmmm512,
		EVEX_Vpbroadcastmw2d_xmm_kr,
		EVEX_Vpbroadcastmw2d_ymm_kr,
		EVEX_Vpbroadcastmw2d_zmm_kr,
		Pminud_xmm_xmmm128,
		VEX_Vpminud_xmm_xmm_xmmm128,
		VEX_Vpminud_ymm_ymm_ymmm256,
		EVEX_Vpminud_xmm_k1z_xmm_xmmm128b32,
		EVEX_Vpminud_ymm_k1z_ymm_ymmm256b32,
		EVEX_Vpminud_zmm_k1z_zmm_zmmm512b32,
		EVEX_Vpminuq_xmm_k1z_xmm_xmmm128b64,
		EVEX_Vpminuq_ymm_k1z_ymm_ymmm256b64,
		EVEX_Vpminuq_zmm_k1z_zmm_zmmm512b64,
		Pmaxsb_xmm_xmmm128,
		VEX_Vpmaxsb_xmm_xmm_xmmm128,
		VEX_Vpmaxsb_ymm_ymm_ymmm256,
		EVEX_Vpmaxsb_xmm_k1z_xmm_xmmm128,
		EVEX_Vpmaxsb_ymm_k1z_ymm_ymmm256,
		EVEX_Vpmaxsb_zmm_k1z_zmm_zmmm512,
		Pmaxsd_xmm_xmmm128,
		VEX_Vpmaxsd_xmm_xmm_xmmm128,
		VEX_Vpmaxsd_ymm_ymm_ymmm256,
		EVEX_Vpmaxsd_xmm_k1z_xmm_xmmm128b32,
		EVEX_Vpmaxsd_ymm_k1z_ymm_ymmm256b32,
		EVEX_Vpmaxsd_zmm_k1z_zmm_zmmm512b32,
		EVEX_Vpmaxsq_xmm_k1z_xmm_xmmm128b64,
		EVEX_Vpmaxsq_ymm_k1z_ymm_ymmm256b64,
		EVEX_Vpmaxsq_zmm_k1z_zmm_zmmm512b64,
		Pmaxuw_xmm_xmmm128,
		VEX_Vpmaxuw_xmm_xmm_xmmm128,
		VEX_Vpmaxuw_ymm_ymm_ymmm256,
		EVEX_Vpmaxuw_xmm_k1z_xmm_xmmm128,
		EVEX_Vpmaxuw_ymm_k1z_ymm_ymmm256,
		EVEX_Vpmaxuw_zmm_k1z_zmm_zmmm512,
		Pmaxud_xmm_xmmm128,
		VEX_Vpmaxud_xmm_xmm_xmmm128,
		VEX_Vpmaxud_ymm_ymm_ymmm256,
		EVEX_Vpmaxud_xmm_k1z_xmm_xmmm128b32,
		EVEX_Vpmaxud_ymm_k1z_ymm_ymmm256b32,
		EVEX_Vpmaxud_zmm_k1z_zmm_zmmm512b32,
		EVEX_Vpmaxuq_xmm_k1z_xmm_xmmm128b64,
		EVEX_Vpmaxuq_ymm_k1z_ymm_ymmm256b64,
		EVEX_Vpmaxuq_zmm_k1z_zmm_zmmm512b64,
		Pmulld_xmm_xmmm128,
		VEX_Vpmulld_xmm_xmm_xmmm128,
		VEX_Vpmulld_ymm_ymm_ymmm256,
		EVEX_Vpmulld_xmm_k1z_xmm_xmmm128b32,
		EVEX_Vpmulld_ymm_k1z_ymm_ymmm256b32,
		EVEX_Vpmulld_zmm_k1z_zmm_zmmm512b32,
		EVEX_Vpmullq_xmm_k1z_xmm_xmmm128b64,
		EVEX_Vpmullq_ymm_k1z_ymm_ymmm256b64,
		EVEX_Vpmullq_zmm_k1z_zmm_zmmm512b64,
		Phminposuw_xmm_xmmm128,
		VEX_Vphminposuw_xmm_xmmm128,
		EVEX_Vgetexpps_xmm_k1z_xmmm128b32,
		EVEX_Vgetexpps_ymm_k1z_ymmm256b32,
		EVEX_Vgetexpps_zmm_k1z_zmmm512b32_sae,
		EVEX_Vgetexppd_xmm_k1z_xmmm128b64,
		EVEX_Vgetexppd_ymm_k1z_ymmm256b64,
		EVEX_Vgetexppd_zmm_k1z_zmmm512b64_sae,
		EVEX_Vgetexpss_xmm_k1z_xmm_xmmm32_sae,
		EVEX_Vgetexpsd_xmm_k1z_xmm_xmmm64_sae,
		EVEX_Vplzcntd_xmm_k1z_xmmm128b32,
		EVEX_Vplzcntd_ymm_k1z_ymmm256b32,
		EVEX_Vplzcntd_zmm_k1z_zmmm512b32,
		EVEX_Vplzcntq_xmm_k1z_xmmm128b64,
		EVEX_Vplzcntq_ymm_k1z_ymmm256b64,
		EVEX_Vplzcntq_zmm_k1z_zmmm512b64,
		VEX_Vpsrlvd_xmm_xmm_xmmm128,
		VEX_Vpsrlvd_ymm_ymm_ymmm256,
		VEX_Vpsrlvq_xmm_xmm_xmmm128,
		VEX_Vpsrlvq_ymm_ymm_ymmm256,
		EVEX_Vpsrlvd_xmm_k1z_xmm_xmmm128b32,
		EVEX_Vpsrlvd_ymm_k1z_ymm_ymmm256b32,
		EVEX_Vpsrlvd_zmm_k1z_zmm_zmmm512b32,
		EVEX_Vpsrlvq_xmm_k1z_xmm_xmmm128b64,
		EVEX_Vpsrlvq_ymm_k1z_ymm_ymmm256b64,
		EVEX_Vpsrlvq_zmm_k1z_zmm_zmmm512b64,
		VEX_Vpsravd_xmm_xmm_xmmm128,
		VEX_Vpsravd_ymm_ymm_ymmm256,
		EVEX_Vpsravd_xmm_k1z_xmm_xmmm128b32,
		EVEX_Vpsravd_ymm_k1z_ymm_ymmm256b32,
		EVEX_Vpsravd_zmm_k1z_zmm_zmmm512b32,
		EVEX_Vpsravq_xmm_k1z_xmm_xmmm128b64,
		EVEX_Vpsravq_ymm_k1z_ymm_ymmm256b64,
		EVEX_Vpsravq_zmm_k1z_zmm_zmmm512b64,
		VEX_Vpsllvd_xmm_xmm_xmmm128,
		VEX_Vpsllvd_ymm_ymm_ymmm256,
		VEX_Vpsllvq_xmm_xmm_xmmm128,
		VEX_Vpsllvq_ymm_ymm_ymmm256,
		EVEX_Vpsllvd_xmm_k1z_xmm_xmmm128b32,
		EVEX_Vpsllvd_ymm_k1z_ymm_ymmm256b32,
		EVEX_Vpsllvd_zmm_k1z_zmm_zmmm512b32,
		EVEX_Vpsllvq_xmm_k1z_xmm_xmmm128b64,
		EVEX_Vpsllvq_ymm_k1z_ymm_ymmm256b64,
		EVEX_Vpsllvq_zmm_k1z_zmm_zmmm512b64,
		EVEX_Vrcp14ps_xmm_k1z_xmmm128b32,
		EVEX_Vrcp14ps_ymm_k1z_ymmm256b32,
		EVEX_Vrcp14ps_zmm_k1z_zmmm512b32,
		EVEX_Vrcp14pd_xmm_k1z_xmmm128b64,
		EVEX_Vrcp14pd_ymm_k1z_ymmm256b64,
		EVEX_Vrcp14pd_zmm_k1z_zmmm512b64,
		EVEX_Vrcp14ss_xmm_k1z_xmm_xmmm32,
		EVEX_Vrcp14sd_xmm_k1z_xmm_xmmm64,
		EVEX_Vrsqrt14ps_xmm_k1z_xmmm128b32,
		EVEX_Vrsqrt14ps_ymm_k1z_ymmm256b32,
		EVEX_Vrsqrt14ps_zmm_k1z_zmmm512b32,
		EVEX_Vrsqrt14pd_xmm_k1z_xmmm128b64,
		EVEX_Vrsqrt14pd_ymm_k1z_ymmm256b64,
		EVEX_Vrsqrt14pd_zmm_k1z_zmmm512b64,
		EVEX_Vrsqrt14ss_xmm_k1z_xmm_xmmm32,
		EVEX_Vrsqrt14sd_xmm_k1z_xmm_xmmm64,
		EVEX_Vpdpbusd_xmm_k1z_xmm_xmmm128b32,
		EVEX_Vpdpbusd_ymm_k1z_ymm_ymmm256b32,
		EVEX_Vpdpbusd_zmm_k1z_zmm_zmmm512b32,
		EVEX_Vpdpbusds_xmm_k1z_xmm_xmmm128b32,
		EVEX_Vpdpbusds_ymm_k1z_ymm_ymmm256b32,
		EVEX_Vpdpbusds_zmm_k1z_zmm_zmmm512b32,
		EVEX_Vpdpwssd_xmm_k1z_xmm_xmmm128b32,
		EVEX_Vpdpwssd_ymm_k1z_ymm_ymmm256b32,
		EVEX_Vpdpwssd_zmm_k1z_zmm_zmmm512b32,
		EVEX_Vdpbf16ps_xmm_k1z_xmm_xmmm128b32,
		EVEX_Vdpbf16ps_ymm_k1z_ymm_ymmm256b32,
		EVEX_Vdpbf16ps_zmm_k1z_zmm_zmmm512b32,
		EVEX_Vp4dpwssd_zmm_k1z_zmmp3_m128,
		EVEX_Vpdpwssds_xmm_k1z_xmm_xmmm128b32,
		EVEX_Vpdpwssds_ymm_k1z_ymm_ymmm256b32,
		EVEX_Vpdpwssds_zmm_k1z_zmm_zmmm512b32,
		EVEX_Vp4dpwssds_zmm_k1z_zmmp3_m128,
		EVEX_Vpopcntb_xmm_k1z_xmmm128,
		EVEX_Vpopcntb_ymm_k1z_ymmm256,
		EVEX_Vpopcntb_zmm_k1z_zmmm512,
		EVEX_Vpopcntw_xmm_k1z_xmmm128,
		EVEX_Vpopcntw_ymm_k1z_ymmm256,
		EVEX_Vpopcntw_zmm_k1z_zmmm512,
		EVEX_Vpopcntd_xmm_k1z_xmmm128b32,
		EVEX_Vpopcntd_ymm_k1z_ymmm256b32,
		EVEX_Vpopcntd_zmm_k1z_zmmm512b32,
		EVEX_Vpopcntq_xmm_k1z_xmmm128b64,
		EVEX_Vpopcntq_ymm_k1z_ymmm256b64,
		EVEX_Vpopcntq_zmm_k1z_zmmm512b64,
		VEX_Vpbroadcastd_xmm_xmmm32,
		VEX_Vpbroadcastd_ymm_xmmm32,
		EVEX_Vpbroadcastd_xmm_k1z_xmmm32,
		EVEX_Vpbroadcastd_ymm_k1z_xmmm32,
		EVEX_Vpbroadcastd_zmm_k1z_xmmm32,
		VEX_Vpbroadcastq_xmm_xmmm64,
		VEX_Vpbroadcastq_ymm_xmmm64,
		EVEX_Vbroadcasti32x2_xmm_k1z_xmmm64,
		EVEX_Vbroadcasti32x2_ymm_k1z_xmmm64,
		EVEX_Vbroadcasti32x2_zmm_k1z_xmmm64,
		EVEX_Vpbroadcastq_xmm_k1z_xmmm64,
		EVEX_Vpbroadcastq_ymm_k1z_xmmm64,
		EVEX_Vpbroadcastq_zmm_k1z_xmmm64,
		VEX_Vbroadcasti128_ymm_m128,
		EVEX_Vbroadcasti32x4_ymm_k1z_m128,
		EVEX_Vbroadcasti32x4_zmm_k1z_m128,
		EVEX_Vbroadcasti64x2_ymm_k1z_m128,
		EVEX_Vbroadcasti64x2_zmm_k1z_m128,
		EVEX_Vbroadcasti32x8_zmm_k1z_m256,
		EVEX_Vbroadcasti64x4_zmm_k1z_m256,
		EVEX_Vpexpandb_xmm_k1z_xmmm128,
		EVEX_Vpexpandb_ymm_k1z_ymmm256,
		EVEX_Vpexpandb_zmm_k1z_zmmm512,
		EVEX_Vpexpandw_xmm_k1z_xmmm128,
		EVEX_Vpexpandw_ymm_k1z_ymmm256,
		EVEX_Vpexpandw_zmm_k1z_zmmm512,
		EVEX_Vpcompressb_xmmm128_k1z_xmm,
		EVEX_Vpcompressb_ymmm256_k1z_ymm,
		EVEX_Vpcompressb_zmmm512_k1z_zmm,
		EVEX_Vpcompressw_xmmm128_k1z_xmm,
		EVEX_Vpcompressw_ymmm256_k1z_ymm,
		EVEX_Vpcompressw_zmmm512_k1z_zmm,
		EVEX_Vpblendmd_xmm_k1z_xmm_xmmm128b32,
		EVEX_Vpblendmd_ymm_k1z_ymm_ymmm256b32,
		EVEX_Vpblendmd_zmm_k1z_zmm_zmmm512b32,
		EVEX_Vpblendmq_xmm_k1z_xmm_xmmm128b64,
		EVEX_Vpblendmq_ymm_k1z_ymm_ymmm256b64,
		EVEX_Vpblendmq_zmm_k1z_zmm_zmmm512b64,
		EVEX_Vblendmps_xmm_k1z_xmm_xmmm128b32,
		EVEX_Vblendmps_ymm_k1z_ymm_ymmm256b32,
		EVEX_Vblendmps_zmm_k1z_zmm_zmmm512b32,
		EVEX_Vblendmpd_xmm_k1z_xmm_xmmm128b64,
		EVEX_Vblendmpd_ymm_k1z_ymm_ymmm256b64,
		EVEX_Vblendmpd_zmm_k1z_zmm_zmmm512b64,
		EVEX_Vpblendmb_xmm_k1z_xmm_xmmm128,
		EVEX_Vpblendmb_ymm_k1z_ymm_ymmm256,
		EVEX_Vpblendmb_zmm_k1z_zmm_zmmm512,
		EVEX_Vpblendmw_xmm_k1z_xmm_xmmm128,
		EVEX_Vpblendmw_ymm_k1z_ymm_ymmm256,
		EVEX_Vpblendmw_zmm_k1z_zmm_zmmm512,
		EVEX_Vp2intersectd_kp1_xmm_xmmm128b32,
		EVEX_Vp2intersectd_kp1_ymm_ymmm256b32,
		EVEX_Vp2intersectd_kp1_zmm_zmmm512b32,
		EVEX_Vp2intersectq_kp1_xmm_xmmm128b64,
		EVEX_Vp2intersectq_kp1_ymm_ymmm256b64,
		EVEX_Vp2intersectq_kp1_zmm_zmmm512b64,
		EVEX_Vpshldvw_xmm_k1z_xmm_xmmm128,
		EVEX_Vpshldvw_ymm_k1z_ymm_ymmm256,
		EVEX_Vpshldvw_zmm_k1z_zmm_zmmm512,
		EVEX_Vpshldvd_xmm_k1z_xmm_xmmm128b32,
		EVEX_Vpshldvd_ymm_k1z_ymm_ymmm256b32,
		EVEX_Vpshldvd_zmm_k1z_zmm_zmmm512b32,
		EVEX_Vpshldvq_xmm_k1z_xmm_xmmm128b64,
		EVEX_Vpshldvq_ymm_k1z_ymm_ymmm256b64,
		EVEX_Vpshldvq_zmm_k1z_zmm_zmmm512b64,
		EVEX_Vpshrdvw_xmm_k1z_xmm_xmmm128,
		EVEX_Vpshrdvw_ymm_k1z_ymm_ymmm256,
		EVEX_Vpshrdvw_zmm_k1z_zmm_zmmm512,
		EVEX_Vcvtneps2bf16_xmm_k1z_xmmm128b32,
		EVEX_Vcvtneps2bf16_xmm_k1z_ymmm256b32,
		EVEX_Vcvtneps2bf16_ymm_k1z_zmmm512b32,
		EVEX_Vcvtne2ps2bf16_xmm_k1z_xmm_xmmm128b32,
		EVEX_Vcvtne2ps2bf16_ymm_k1z_ymm_ymmm256b32,
		EVEX_Vcvtne2ps2bf16_zmm_k1z_zmm_zmmm512b32,
		EVEX_Vpshrdvd_xmm_k1z_xmm_xmmm128b32,
		EVEX_Vpshrdvd_ymm_k1z_ymm_ymmm256b32,
		EVEX_Vpshrdvd_zmm_k1z_zmm_zmmm512b32,
		EVEX_Vpshrdvq_xmm_k1z_xmm_xmmm128b64,
		EVEX_Vpshrdvq_ymm_k1z_ymm_ymmm256b64,
		EVEX_Vpshrdvq_zmm_k1z_zmm_zmmm512b64,
		EVEX_Vpermi2b_xmm_k1z_xmm_xmmm128,
		EVEX_Vpermi2b_ymm_k1z_ymm_ymmm256,
		EVEX_Vpermi2b_zmm_k1z_zmm_zmmm512,
		EVEX_Vpermi2w_xmm_k1z_xmm_xmmm128,
		EVEX_Vpermi2w_ymm_k1z_ymm_ymmm256,
		EVEX_Vpermi2w_zmm_k1z_zmm_zmmm512,
		EVEX_Vpermi2d_xmm_k1z_xmm_xmmm128b32,
		EVEX_Vpermi2d_ymm_k1z_ymm_ymmm256b32,
		EVEX_Vpermi2d_zmm_k1z_zmm_zmmm512b32,
		EVEX_Vpermi2q_xmm_k1z_xmm_xmmm128b64,
		EVEX_Vpermi2q_ymm_k1z_ymm_ymmm256b64,
		EVEX_Vpermi2q_zmm_k1z_zmm_zmmm512b64,
		EVEX_Vpermi2ps_xmm_k1z_xmm_xmmm128b32,
		EVEX_Vpermi2ps_ymm_k1z_ymm_ymmm256b32,
		EVEX_Vpermi2ps_zmm_k1z_zmm_zmmm512b32,
		EVEX_Vpermi2pd_xmm_k1z_xmm_xmmm128b64,
		EVEX_Vpermi2pd_ymm_k1z_ymm_ymmm256b64,
		EVEX_Vpermi2pd_zmm_k1z_zmm_zmmm512b64,
		VEX_Vpbroadcastb_xmm_xmmm8,
		VEX_Vpbroadcastb_ymm_xmmm8,
		EVEX_Vpbroadcastb_xmm_k1z_xmmm8,
		EVEX_Vpbroadcastb_ymm_k1z_xmmm8,
		EVEX_Vpbroadcastb_zmm_k1z_xmmm8,
		VEX_Vpbroadcastw_xmm_xmmm16,
		VEX_Vpbroadcastw_ymm_xmmm16,
		EVEX_Vpbroadcastw_xmm_k1z_xmmm16,
		EVEX_Vpbroadcastw_ymm_k1z_xmmm16,
		EVEX_Vpbroadcastw_zmm_k1z_xmmm16,
		EVEX_Vpbroadcastb_xmm_k1z_r32,
		EVEX_Vpbroadcastb_ymm_k1z_r32,
		EVEX_Vpbroadcastb_zmm_k1z_r32,
		EVEX_Vpbroadcastw_xmm_k1z_r32,
		EVEX_Vpbroadcastw_ymm_k1z_r32,
		EVEX_Vpbroadcastw_zmm_k1z_r32,
		EVEX_Vpbroadcastd_xmm_k1z_r32,
		EVEX_Vpbroadcastd_ymm_k1z_r32,
		EVEX_Vpbroadcastd_zmm_k1z_r32,
		EVEX_Vpbroadcastq_xmm_k1z_r64,
		EVEX_Vpbroadcastq_ymm_k1z_r64,
		EVEX_Vpbroadcastq_zmm_k1z_r64,
		EVEX_Vpermt2b_xmm_k1z_xmm_xmmm128,
		EVEX_Vpermt2b_ymm_k1z_ymm_ymmm256,
		EVEX_Vpermt2b_zmm_k1z_zmm_zmmm512,
		EVEX_Vpermt2w_xmm_k1z_xmm_xmmm128,
		EVEX_Vpermt2w_ymm_k1z_ymm_ymmm256,
		EVEX_Vpermt2w_zmm_k1z_zmm_zmmm512,
		EVEX_Vpermt2d_xmm_k1z_xmm_xmmm128b32,
		EVEX_Vpermt2d_ymm_k1z_ymm_ymmm256b32,
		EVEX_Vpermt2d_zmm_k1z_zmm_zmmm512b32,
		EVEX_Vpermt2q_xmm_k1z_xmm_xmmm128b64,
		EVEX_Vpermt2q_ymm_k1z_ymm_ymmm256b64,
		EVEX_Vpermt2q_zmm_k1z_zmm_zmmm512b64,
		EVEX_Vpermt2ps_xmm_k1z_xmm_xmmm128b32,
		EVEX_Vpermt2ps_ymm_k1z_ymm_ymmm256b32,
		EVEX_Vpermt2ps_zmm_k1z_zmm_zmmm512b32,
		EVEX_Vpermt2pd_xmm_k1z_xmm_xmmm128b64,
		EVEX_Vpermt2pd_ymm_k1z_ymm_ymmm256b64,
		EVEX_Vpermt2pd_zmm_k1z_zmm_zmmm512b64,
		Invept_r32_m128,
		Invept_r64_m128,
		Invvpid_r32_m128,
		Invvpid_r64_m128,
		Invpcid_r32_m128,
		Invpcid_r64_m128,
		EVEX_Vpmultishiftqb_xmm_k1z_xmm_xmmm128b64,
		EVEX_Vpmultishiftqb_ymm_k1z_ymm_ymmm256b64,
		EVEX_Vpmultishiftqb_zmm_k1z_zmm_zmmm512b64,
		EVEX_Vexpandps_xmm_k1z_xmmm128,
		EVEX_Vexpandps_ymm_k1z_ymmm256,
		EVEX_Vexpandps_zmm_k1z_zmmm512,
		EVEX_Vexpandpd_xmm_k1z_xmmm128,
		EVEX_Vexpandpd_ymm_k1z_ymmm256,
		EVEX_Vexpandpd_zmm_k1z_zmmm512,
		EVEX_Vpexpandd_xmm_k1z_xmmm128,
		EVEX_Vpexpandd_ymm_k1z_ymmm256,
		EVEX_Vpexpandd_zmm_k1z_zmmm512,
		EVEX_Vpexpandq_xmm_k1z_xmmm128,
		EVEX_Vpexpandq_ymm_k1z_ymmm256,
		EVEX_Vpexpandq_zmm_k1z_zmmm512,
		EVEX_Vcompressps_xmmm128_k1z_xmm,
		EVEX_Vcompressps_ymmm256_k1z_ymm,
		EVEX_Vcompressps_zmmm512_k1z_zmm,
		EVEX_Vcompresspd_xmmm128_k1z_xmm,
		EVEX_Vcompresspd_ymmm256_k1z_ymm,
		EVEX_Vcompresspd_zmmm512_k1z_zmm,
		EVEX_Vpcompressd_xmmm128_k1z_xmm,
		EVEX_Vpcompressd_ymmm256_k1z_ymm,
		EVEX_Vpcompressd_zmmm512_k1z_zmm,
		EVEX_Vpcompressq_xmmm128_k1z_xmm,
		EVEX_Vpcompressq_ymmm256_k1z_ymm,
		EVEX_Vpcompressq_zmmm512_k1z_zmm,
		VEX_Vpmaskmovd_xmm_xmm_m128,
		VEX_Vpmaskmovd_ymm_ymm_m256,
		VEX_Vpmaskmovq_xmm_xmm_m128,
		VEX_Vpmaskmovq_ymm_ymm_m256,
		EVEX_Vpermb_xmm_k1z_xmm_xmmm128,
		EVEX_Vpermb_ymm_k1z_ymm_ymmm256,
		EVEX_Vpermb_zmm_k1z_zmm_zmmm512,
		EVEX_Vpermw_xmm_k1z_xmm_xmmm128,
		EVEX_Vpermw_ymm_k1z_ymm_ymmm256,
		EVEX_Vpermw_zmm_k1z_zmm_zmmm512,
		VEX_Vpmaskmovd_m128_xmm_xmm,
		VEX_Vpmaskmovd_m256_ymm_ymm,
		VEX_Vpmaskmovq_m128_xmm_xmm,
		VEX_Vpmaskmovq_m256_ymm_ymm,
		EVEX_Vpshufbitqmb_kr_k1_xmm_xmmm128,
		EVEX_Vpshufbitqmb_kr_k1_ymm_ymmm256,
		EVEX_Vpshufbitqmb_kr_k1_zmm_zmmm512,
		VEX_Vpgatherdd_xmm_vm32x_xmm,
		VEX_Vpgatherdd_ymm_vm32y_ymm,
		VEX_Vpgatherdq_xmm_vm32x_xmm,
		VEX_Vpgatherdq_ymm_vm32x_ymm,
		EVEX_Vpgatherdd_xmm_k1_vm32x,
		EVEX_Vpgatherdd_ymm_k1_vm32y,
		EVEX_Vpgatherdd_zmm_k1_vm32z,
		EVEX_Vpgatherdq_xmm_k1_vm32x,
		EVEX_Vpgatherdq_ymm_k1_vm32x,
		EVEX_Vpgatherdq_zmm_k1_vm32y,
		VEX_Vpgatherqd_xmm_vm64x_xmm,
		VEX_Vpgatherqd_xmm_vm64y_xmm,
		VEX_Vpgatherqq_xmm_vm64x_xmm,
		VEX_Vpgatherqq_ymm_vm64y_ymm,
		EVEX_Vpgatherqd_xmm_k1_vm64x,
		EVEX_Vpgatherqd_xmm_k1_vm64y,
		EVEX_Vpgatherqd_ymm_k1_vm64z,
		EVEX_Vpgatherqq_xmm_k1_vm64x,
		EVEX_Vpgatherqq_ymm_k1_vm64y,
		EVEX_Vpgatherqq_zmm_k1_vm64z,
		VEX_Vgatherdps_xmm_vm32x_xmm,
		VEX_Vgatherdps_ymm_vm32y_ymm,
		VEX_Vgatherdpd_xmm_vm32x_xmm,
		VEX_Vgatherdpd_ymm_vm32x_ymm,
		EVEX_Vgatherdps_xmm_k1_vm32x,
		EVEX_Vgatherdps_ymm_k1_vm32y,
		EVEX_Vgatherdps_zmm_k1_vm32z,
		EVEX_Vgatherdpd_xmm_k1_vm32x,
		EVEX_Vgatherdpd_ymm_k1_vm32x,
		EVEX_Vgatherdpd_zmm_k1_vm32y,
		VEX_Vgatherqps_xmm_vm64x_xmm,
		VEX_Vgatherqps_xmm_vm64y_xmm,
		VEX_Vgatherqpd_xmm_vm64x_xmm,
		VEX_Vgatherqpd_ymm_vm64y_ymm,
		EVEX_Vgatherqps_xmm_k1_vm64x,
		EVEX_Vgatherqps_xmm_k1_vm64y,
		EVEX_Vgatherqps_ymm_k1_vm64z,
		EVEX_Vgatherqpd_xmm_k1_vm64x,
		EVEX_Vgatherqpd_ymm_k1_vm64y,
		EVEX_Vgatherqpd_zmm_k1_vm64z,
		VEX_Vfmaddsub132ps_xmm_xmm_xmmm128,
		VEX_Vfmaddsub132ps_ymm_ymm_ymmm256,
		VEX_Vfmaddsub132pd_xmm_xmm_xmmm128,
		VEX_Vfmaddsub132pd_ymm_ymm_ymmm256,
		EVEX_Vfmaddsub132ps_xmm_k1z_xmm_xmmm128b32,
		EVEX_Vfmaddsub132ps_ymm_k1z_ymm_ymmm256b32,
		EVEX_Vfmaddsub132ps_zmm_k1z_zmm_zmmm512b32_er,
		EVEX_Vfmaddsub132pd_xmm_k1z_xmm_xmmm128b64,
		EVEX_Vfmaddsub132pd_ymm_k1z_ymm_ymmm256b64,
		EVEX_Vfmaddsub132pd_zmm_k1z_zmm_zmmm512b64_er,
		VEX_Vfmsubadd132ps_xmm_xmm_xmmm128,
		VEX_Vfmsubadd132ps_ymm_ymm_ymmm256,
		VEX_Vfmsubadd132pd_xmm_xmm_xmmm128,
		VEX_Vfmsubadd132pd_ymm_ymm_ymmm256,
		EVEX_Vfmsubadd132ps_xmm_k1z_xmm_xmmm128b32,
		EVEX_Vfmsubadd132ps_ymm_k1z_ymm_ymmm256b32,
		EVEX_Vfmsubadd132ps_zmm_k1z_zmm_zmmm512b32_er,
		EVEX_Vfmsubadd132pd_xmm_k1z_xmm_xmmm128b64,
		EVEX_Vfmsubadd132pd_ymm_k1z_ymm_ymmm256b64,
		EVEX_Vfmsubadd132pd_zmm_k1z_zmm_zmmm512b64_er,
		VEX_Vfmadd132ps_xmm_xmm_xmmm128,
		VEX_Vfmadd132ps_ymm_ymm_ymmm256,
		VEX_Vfmadd132pd_xmm_xmm_xmmm128,
		VEX_Vfmadd132pd_ymm_ymm_ymmm256,
		EVEX_Vfmadd132ps_xmm_k1z_xmm_xmmm128b32,
		EVEX_Vfmadd132ps_ymm_k1z_ymm_ymmm256b32,
		EVEX_Vfmadd132ps_zmm_k1z_zmm_zmmm512b32_er,
		EVEX_Vfmadd132pd_xmm_k1z_xmm_xmmm128b64,
		EVEX_Vfmadd132pd_ymm_k1z_ymm_ymmm256b64,
		EVEX_Vfmadd132pd_zmm_k1z_zmm_zmmm512b64_er,
		VEX_Vfmadd132ss_xmm_xmm_xmmm32,
		VEX_Vfmadd132sd_xmm_xmm_xmmm64,
		EVEX_Vfmadd132ss_xmm_k1z_xmm_xmmm32_er,
		EVEX_Vfmadd132sd_xmm_k1z_xmm_xmmm64_er,
		VEX_Vfmsub132ps_xmm_xmm_xmmm128,
		VEX_Vfmsub132ps_ymm_ymm_ymmm256,
		VEX_Vfmsub132pd_xmm_xmm_xmmm128,
		VEX_Vfmsub132pd_ymm_ymm_ymmm256,
		EVEX_Vfmsub132ps_xmm_k1z_xmm_xmmm128b32,
		EVEX_Vfmsub132ps_ymm_k1z_ymm_ymmm256b32,
		EVEX_Vfmsub132ps_zmm_k1z_zmm_zmmm512b32_er,
		EVEX_Vfmsub132pd_xmm_k1z_xmm_xmmm128b64,
		EVEX_Vfmsub132pd_ymm_k1z_ymm_ymmm256b64,
		EVEX_Vfmsub132pd_zmm_k1z_zmm_zmmm512b64_er,
		EVEX_V4fmaddps_zmm_k1z_zmmp3_m128,
		VEX_Vfmsub132ss_xmm_xmm_xmmm32,
		VEX_Vfmsub132sd_xmm_xmm_xmmm64,
		EVEX_Vfmsub132ss_xmm_k1z_xmm_xmmm32_er,
		EVEX_Vfmsub132sd_xmm_k1z_xmm_xmmm64_er,
		EVEX_V4fmaddss_xmm_k1z_xmmp3_m128,
		VEX_Vfnmadd132ps_xmm_xmm_xmmm128,
		VEX_Vfnmadd132ps_ymm_ymm_ymmm256,
		VEX_Vfnmadd132pd_xmm_xmm_xmmm128,
		VEX_Vfnmadd132pd_ymm_ymm_ymmm256,
		EVEX_Vfnmadd132ps_xmm_k1z_xmm_xmmm128b32,
		EVEX_Vfnmadd132ps_ymm_k1z_ymm_ymmm256b32,
		EVEX_Vfnmadd132ps_zmm_k1z_zmm_zmmm512b32_er,
		EVEX_Vfnmadd132pd_xmm_k1z_xmm_xmmm128b64,
		EVEX_Vfnmadd132pd_ymm_k1z_ymm_ymmm256b64,
		EVEX_Vfnmadd132pd_zmm_k1z_zmm_zmmm512b64_er,
		VEX_Vfnmadd132ss_xmm_xmm_xmmm32,
		VEX_Vfnmadd132sd_xmm_xmm_xmmm64,
		EVEX_Vfnmadd132ss_xmm_k1z_xmm_xmmm32_er,
		EVEX_Vfnmadd132sd_xmm_k1z_xmm_xmmm64_er,
		VEX_Vfnmsub132ps_xmm_xmm_xmmm128,
		VEX_Vfnmsub132ps_ymm_ymm_ymmm256,
		VEX_Vfnmsub132pd_xmm_xmm_xmmm128,
		VEX_Vfnmsub132pd_ymm_ymm_ymmm256,
		EVEX_Vfnmsub132ps_xmm_k1z_xmm_xmmm128b32,
		EVEX_Vfnmsub132ps_ymm_k1z_ymm_ymmm256b32,
		EVEX_Vfnmsub132ps_zmm_k1z_zmm_zmmm512b32_er,
		EVEX_Vfnmsub132pd_xmm_k1z_xmm_xmmm128b64,
		EVEX_Vfnmsub132pd_ymm_k1z_ymm_ymmm256b64,
		EVEX_Vfnmsub132pd_zmm_k1z_zmm_zmmm512b64_er,
		VEX_Vfnmsub132ss_xmm_xmm_xmmm32,
		VEX_Vfnmsub132sd_xmm_xmm_xmmm64,
		EVEX_Vfnmsub132ss_xmm_k1z_xmm_xmmm32_er,
		EVEX_Vfnmsub132sd_xmm_k1z_xmm_xmmm64_er,
		EVEX_Vpscatterdd_vm32x_k1_xmm,
		EVEX_Vpscatterdd_vm32y_k1_ymm,
		EVEX_Vpscatterdd_vm32z_k1_zmm,
		EVEX_Vpscatterdq_vm32x_k1_xmm,
		EVEX_Vpscatterdq_vm32x_k1_ymm,
		EVEX_Vpscatterdq_vm32y_k1_zmm,
		EVEX_Vpscatterqd_vm64x_k1_xmm,
		EVEX_Vpscatterqd_vm64y_k1_xmm,
		EVEX_Vpscatterqd_vm64z_k1_ymm,
		EVEX_Vpscatterqq_vm64x_k1_xmm,
		EVEX_Vpscatterqq_vm64y_k1_ymm,
		EVEX_Vpscatterqq_vm64z_k1_zmm,
		EVEX_Vscatterdps_vm32x_k1_xmm,
		EVEX_Vscatterdps_vm32y_k1_ymm,
		EVEX_Vscatterdps_vm32z_k1_zmm,
		EVEX_Vscatterdpd_vm32x_k1_xmm,
		EVEX_Vscatterdpd_vm32x_k1_ymm,
		EVEX_Vscatterdpd_vm32y_k1_zmm,
		EVEX_Vscatterqps_vm64x_k1_xmm,
		EVEX_Vscatterqps_vm64y_k1_xmm,
		EVEX_Vscatterqps_vm64z_k1_ymm,
		EVEX_Vscatterqpd_vm64x_k1_xmm,
		EVEX_Vscatterqpd_vm64y_k1_ymm,
		EVEX_Vscatterqpd_vm64z_k1_zmm,
		VEX_Vfmaddsub213ps_xmm_xmm_xmmm128,
		VEX_Vfmaddsub213ps_ymm_ymm_ymmm256,
		VEX_Vfmaddsub213pd_xmm_xmm_xmmm128,
		VEX_Vfmaddsub213pd_ymm_ymm_ymmm256,
		EVEX_Vfmaddsub213ps_xmm_k1z_xmm_xmmm128b32,
		EVEX_Vfmaddsub213ps_ymm_k1z_ymm_ymmm256b32,
		EVEX_Vfmaddsub213ps_zmm_k1z_zmm_zmmm512b32_er,
		EVEX_Vfmaddsub213pd_xmm_k1z_xmm_xmmm128b64,
		EVEX_Vfmaddsub213pd_ymm_k1z_ymm_ymmm256b64,
		EVEX_Vfmaddsub213pd_zmm_k1z_zmm_zmmm512b64_er,
		VEX_Vfmsubadd213ps_xmm_xmm_xmmm128,
		VEX_Vfmsubadd213ps_ymm_ymm_ymmm256,
		VEX_Vfmsubadd213pd_xmm_xmm_xmmm128,
		VEX_Vfmsubadd213pd_ymm_ymm_ymmm256,
		EVEX_Vfmsubadd213ps_xmm_k1z_xmm_xmmm128b32,
		EVEX_Vfmsubadd213ps_ymm_k1z_ymm_ymmm256b32,
		EVEX_Vfmsubadd213ps_zmm_k1z_zmm_zmmm512b32_er,
		EVEX_Vfmsubadd213pd_xmm_k1z_xmm_xmmm128b64,
		EVEX_Vfmsubadd213pd_ymm_k1z_ymm_ymmm256b64,
		EVEX_Vfmsubadd213pd_zmm_k1z_zmm_zmmm512b64_er,
		VEX_Vfmadd213ps_xmm_xmm_xmmm128,
		VEX_Vfmadd213ps_ymm_ymm_ymmm256,
		VEX_Vfmadd213pd_xmm_xmm_xmmm128,
		VEX_Vfmadd213pd_ymm_ymm_ymmm256,
		EVEX_Vfmadd213ps_xmm_k1z_xmm_xmmm128b32,
		EVEX_Vfmadd213ps_ymm_k1z_ymm_ymmm256b32,
		EVEX_Vfmadd213ps_zmm_k1z_zmm_zmmm512b32_er,
		EVEX_Vfmadd213pd_xmm_k1z_xmm_xmmm128b64,
		EVEX_Vfmadd213pd_ymm_k1z_ymm_ymmm256b64,
		EVEX_Vfmadd213pd_zmm_k1z_zmm_zmmm512b64_er,
		VEX_Vfmadd213ss_xmm_xmm_xmmm32,
		VEX_Vfmadd213sd_xmm_xmm_xmmm64,
		EVEX_Vfmadd213ss_xmm_k1z_xmm_xmmm32_er,
		EVEX_Vfmadd213sd_xmm_k1z_xmm_xmmm64_er,
		VEX_Vfmsub213ps_xmm_xmm_xmmm128,
		VEX_Vfmsub213ps_ymm_ymm_ymmm256,
		VEX_Vfmsub213pd_xmm_xmm_xmmm128,
		VEX_Vfmsub213pd_ymm_ymm_ymmm256,
		EVEX_Vfmsub213ps_xmm_k1z_xmm_xmmm128b32,
		EVEX_Vfmsub213ps_ymm_k1z_ymm_ymmm256b32,
		EVEX_Vfmsub213ps_zmm_k1z_zmm_zmmm512b32_er,
		EVEX_Vfmsub213pd_xmm_k1z_xmm_xmmm128b64,
		EVEX_Vfmsub213pd_ymm_k1z_ymm_ymmm256b64,
		EVEX_Vfmsub213pd_zmm_k1z_zmm_zmmm512b64_er,
		EVEX_V4fnmaddps_zmm_k1z_zmmp3_m128,
		VEX_Vfmsub213ss_xmm_xmm_xmmm32,
		VEX_Vfmsub213sd_xmm_xmm_xmmm64,
		EVEX_Vfmsub213ss_xmm_k1z_xmm_xmmm32_er,
		EVEX_Vfmsub213sd_xmm_k1z_xmm_xmmm64_er,
		EVEX_V4fnmaddss_xmm_k1z_xmmp3_m128,
		VEX_Vfnmadd213ps_xmm_xmm_xmmm128,
		VEX_Vfnmadd213ps_ymm_ymm_ymmm256,
		VEX_Vfnmadd213pd_xmm_xmm_xmmm128,
		VEX_Vfnmadd213pd_ymm_ymm_ymmm256,
		EVEX_Vfnmadd213ps_xmm_k1z_xmm_xmmm128b32,
		EVEX_Vfnmadd213ps_ymm_k1z_ymm_ymmm256b32,
		EVEX_Vfnmadd213ps_zmm_k1z_zmm_zmmm512b32_er,
		EVEX_Vfnmadd213pd_xmm_k1z_xmm_xmmm128b64,
		EVEX_Vfnmadd213pd_ymm_k1z_ymm_ymmm256b64,
		EVEX_Vfnmadd213pd_zmm_k1z_zmm_zmmm512b64_er,
		VEX_Vfnmadd213ss_xmm_xmm_xmmm32,
		VEX_Vfnmadd213sd_xmm_xmm_xmmm64,
		EVEX_Vfnmadd213ss_xmm_k1z_xmm_xmmm32_er,
		EVEX_Vfnmadd213sd_xmm_k1z_xmm_xmmm64_er,
		VEX_Vfnmsub213ps_xmm_xmm_xmmm128,
		VEX_Vfnmsub213ps_ymm_ymm_ymmm256,
		VEX_Vfnmsub213pd_xmm_xmm_xmmm128,
		VEX_Vfnmsub213pd_ymm_ymm_ymmm256,
		EVEX_Vfnmsub213ps_xmm_k1z_xmm_xmmm128b32,
		EVEX_Vfnmsub213ps_ymm_k1z_ymm_ymmm256b32,
		EVEX_Vfnmsub213ps_zmm_k1z_zmm_zmmm512b32_er,
		EVEX_Vfnmsub213pd_xmm_k1z_xmm_xmmm128b64,
		EVEX_Vfnmsub213pd_ymm_k1z_ymm_ymmm256b64,
		EVEX_Vfnmsub213pd_zmm_k1z_zmm_zmmm512b64_er,
		VEX_Vfnmsub213ss_xmm_xmm_xmmm32,
		VEX_Vfnmsub213sd_xmm_xmm_xmmm64,
		EVEX_Vfnmsub213ss_xmm_k1z_xmm_xmmm32_er,
		EVEX_Vfnmsub213sd_xmm_k1z_xmm_xmmm64_er,
		EVEX_Vpmadd52luq_xmm_k1z_xmm_xmmm128b64,
		EVEX_Vpmadd52luq_ymm_k1z_ymm_ymmm256b64,
		EVEX_Vpmadd52luq_zmm_k1z_zmm_zmmm512b64,
		EVEX_Vpmadd52huq_xmm_k1z_xmm_xmmm128b64,
		EVEX_Vpmadd52huq_ymm_k1z_ymm_ymmm256b64,
		EVEX_Vpmadd52huq_zmm_k1z_zmm_zmmm512b64,
		VEX_Vfmaddsub231ps_xmm_xmm_xmmm128,
		VEX_Vfmaddsub231ps_ymm_ymm_ymmm256,
		VEX_Vfmaddsub231pd_xmm_xmm_xmmm128,
		VEX_Vfmaddsub231pd_ymm_ymm_ymmm256,
		EVEX_Vfmaddsub231ps_xmm_k1z_xmm_xmmm128b32,
		EVEX_Vfmaddsub231ps_ymm_k1z_ymm_ymmm256b32,
		EVEX_Vfmaddsub231ps_zmm_k1z_zmm_zmmm512b32_er,
		EVEX_Vfmaddsub231pd_xmm_k1z_xmm_xmmm128b64,
		EVEX_Vfmaddsub231pd_ymm_k1z_ymm_ymmm256b64,
		EVEX_Vfmaddsub231pd_zmm_k1z_zmm_zmmm512b64_er,
		VEX_Vfmsubadd231ps_xmm_xmm_xmmm128,
		VEX_Vfmsubadd231ps_ymm_ymm_ymmm256,
		VEX_Vfmsubadd231pd_xmm_xmm_xmmm128,
		VEX_Vfmsubadd231pd_ymm_ymm_ymmm256,
		EVEX_Vfmsubadd231ps_xmm_k1z_xmm_xmmm128b32,
		EVEX_Vfmsubadd231ps_ymm_k1z_ymm_ymmm256b32,
		EVEX_Vfmsubadd231ps_zmm_k1z_zmm_zmmm512b32_er,
		EVEX_Vfmsubadd231pd_xmm_k1z_xmm_xmmm128b64,
		EVEX_Vfmsubadd231pd_ymm_k1z_ymm_ymmm256b64,
		EVEX_Vfmsubadd231pd_zmm_k1z_zmm_zmmm512b64_er,
		VEX_Vfmadd231ps_xmm_xmm_xmmm128,
		VEX_Vfmadd231ps_ymm_ymm_ymmm256,
		VEX_Vfmadd231pd_xmm_xmm_xmmm128,
		VEX_Vfmadd231pd_ymm_ymm_ymmm256,
		EVEX_Vfmadd231ps_xmm_k1z_xmm_xmmm128b32,
		EVEX_Vfmadd231ps_ymm_k1z_ymm_ymmm256b32,
		EVEX_Vfmadd231ps_zmm_k1z_zmm_zmmm512b32_er,
		EVEX_Vfmadd231pd_xmm_k1z_xmm_xmmm128b64,
		EVEX_Vfmadd231pd_ymm_k1z_ymm_ymmm256b64,
		EVEX_Vfmadd231pd_zmm_k1z_zmm_zmmm512b64_er,
		VEX_Vfmadd231ss_xmm_xmm_xmmm32,
		VEX_Vfmadd231sd_xmm_xmm_xmmm64,
		EVEX_Vfmadd231ss_xmm_k1z_xmm_xmmm32_er,
		EVEX_Vfmadd231sd_xmm_k1z_xmm_xmmm64_er,
		VEX_Vfmsub231ps_xmm_xmm_xmmm128,
		VEX_Vfmsub231ps_ymm_ymm_ymmm256,
		VEX_Vfmsub231pd_xmm_xmm_xmmm128,
		VEX_Vfmsub231pd_ymm_ymm_ymmm256,
		EVEX_Vfmsub231ps_xmm_k1z_xmm_xmmm128b32,
		EVEX_Vfmsub231ps_ymm_k1z_ymm_ymmm256b32,
		EVEX_Vfmsub231ps_zmm_k1z_zmm_zmmm512b32_er,
		EVEX_Vfmsub231pd_xmm_k1z_xmm_xmmm128b64,
		EVEX_Vfmsub231pd_ymm_k1z_ymm_ymmm256b64,
		EVEX_Vfmsub231pd_zmm_k1z_zmm_zmmm512b64_er,
		VEX_Vfmsub231ss_xmm_xmm_xmmm32,
		VEX_Vfmsub231sd_xmm_xmm_xmmm64,
		EVEX_Vfmsub231ss_xmm_k1z_xmm_xmmm32_er,
		EVEX_Vfmsub231sd_xmm_k1z_xmm_xmmm64_er,
		VEX_Vfnmadd231ps_xmm_xmm_xmmm128,
		VEX_Vfnmadd231ps_ymm_ymm_ymmm256,
		VEX_Vfnmadd231pd_xmm_xmm_xmmm128,
		VEX_Vfnmadd231pd_ymm_ymm_ymmm256,
		EVEX_Vfnmadd231ps_xmm_k1z_xmm_xmmm128b32,
		EVEX_Vfnmadd231ps_ymm_k1z_ymm_ymmm256b32,
		EVEX_Vfnmadd231ps_zmm_k1z_zmm_zmmm512b32_er,
		EVEX_Vfnmadd231pd_xmm_k1z_xmm_xmmm128b64,
		EVEX_Vfnmadd231pd_ymm_k1z_ymm_ymmm256b64,
		EVEX_Vfnmadd231pd_zmm_k1z_zmm_zmmm512b64_er,
		VEX_Vfnmadd231ss_xmm_xmm_xmmm32,
		VEX_Vfnmadd231sd_xmm_xmm_xmmm64,
		EVEX_Vfnmadd231ss_xmm_k1z_xmm_xmmm32_er,
		EVEX_Vfnmadd231sd_xmm_k1z_xmm_xmmm64_er,
		VEX_Vfnmsub231ps_xmm_xmm_xmmm128,
		VEX_Vfnmsub231ps_ymm_ymm_ymmm256,
		VEX_Vfnmsub231pd_xmm_xmm_xmmm128,
		VEX_Vfnmsub231pd_ymm_ymm_ymmm256,
		EVEX_Vfnmsub231ps_xmm_k1z_xmm_xmmm128b32,
		EVEX_Vfnmsub231ps_ymm_k1z_ymm_ymmm256b32,
		EVEX_Vfnmsub231ps_zmm_k1z_zmm_zmmm512b32_er,
		EVEX_Vfnmsub231pd_xmm_k1z_xmm_xmmm128b64,
		EVEX_Vfnmsub231pd_ymm_k1z_ymm_ymmm256b64,
		EVEX_Vfnmsub231pd_zmm_k1z_zmm_zmmm512b64_er,
		VEX_Vfnmsub231ss_xmm_xmm_xmmm32,
		VEX_Vfnmsub231sd_xmm_xmm_xmmm64,
		EVEX_Vfnmsub231ss_xmm_k1z_xmm_xmmm32_er,
		EVEX_Vfnmsub231sd_xmm_k1z_xmm_xmmm64_er,
		EVEX_Vpconflictd_xmm_k1z_xmmm128b32,
		EVEX_Vpconflictd_ymm_k1z_ymmm256b32,
		EVEX_Vpconflictd_zmm_k1z_zmmm512b32,
		EVEX_Vpconflictq_xmm_k1z_xmmm128b64,
		EVEX_Vpconflictq_ymm_k1z_ymmm256b64,
		EVEX_Vpconflictq_zmm_k1z_zmmm512b64,
		EVEX_Vgatherpf0dps_vm32z_k1,
		EVEX_Vgatherpf0dpd_vm32y_k1,
		EVEX_Vgatherpf1dps_vm32z_k1,
		EVEX_Vgatherpf1dpd_vm32y_k1,
		EVEX_Vscatterpf0dps_vm32z_k1,
		EVEX_Vscatterpf0dpd_vm32y_k1,
		EVEX_Vscatterpf1dps_vm32z_k1,
		EVEX_Vscatterpf1dpd_vm32y_k1,
		EVEX_Vgatherpf0qps_vm64z_k1,
		EVEX_Vgatherpf0qpd_vm64z_k1,
		EVEX_Vgatherpf1qps_vm64z_k1,
		EVEX_Vgatherpf1qpd_vm64z_k1,
		EVEX_Vscatterpf0qps_vm64z_k1,
		EVEX_Vscatterpf0qpd_vm64z_k1,
		EVEX_Vscatterpf1qps_vm64z_k1,
		EVEX_Vscatterpf1qpd_vm64z_k1,
		Sha1nexte_xmm_xmmm128,
		EVEX_Vexp2ps_zmm_k1z_zmmm512b32_sae,
		EVEX_Vexp2pd_zmm_k1z_zmmm512b64_sae,
		Sha1msg1_xmm_xmmm128,
		Sha1msg2_xmm_xmmm128,
		EVEX_Vrcp28ps_zmm_k1z_zmmm512b32_sae,
		EVEX_Vrcp28pd_zmm_k1z_zmmm512b64_sae,
		Sha256rnds2_xmm_xmmm128,
		EVEX_Vrcp28ss_xmm_k1z_xmm_xmmm32_sae,
		EVEX_Vrcp28sd_xmm_k1z_xmm_xmmm64_sae,
		Sha256msg1_xmm_xmmm128,
		EVEX_Vrsqrt28ps_zmm_k1z_zmmm512b32_sae,
		EVEX_Vrsqrt28pd_zmm_k1z_zmmm512b64_sae,
		Sha256msg2_xmm_xmmm128,
		EVEX_Vrsqrt28ss_xmm_k1z_xmm_xmmm32_sae,
		EVEX_Vrsqrt28sd_xmm_k1z_xmm_xmmm64_sae,
		Gf2p8mulb_xmm_xmmm128,
		VEX_Vgf2p8mulb_xmm_xmm_xmmm128,
		VEX_Vgf2p8mulb_ymm_ymm_ymmm256,
		EVEX_Vgf2p8mulb_xmm_k1z_xmm_xmmm128,
		EVEX_Vgf2p8mulb_ymm_k1z_ymm_ymmm256,
		EVEX_Vgf2p8mulb_zmm_k1z_zmm_zmmm512,
		Aesimc_xmm_xmmm128,
		VEX_Vaesimc_xmm_xmmm128,
		Aesenc_xmm_xmmm128,
		VEX_Vaesenc_xmm_xmm_xmmm128,
		VEX_Vaesenc_ymm_ymm_ymmm256,
		EVEX_Vaesenc_xmm_xmm_xmmm128,
		EVEX_Vaesenc_ymm_ymm_ymmm256,
		EVEX_Vaesenc_zmm_zmm_zmmm512,
		Aesenclast_xmm_xmmm128,
		VEX_Vaesenclast_xmm_xmm_xmmm128,
		VEX_Vaesenclast_ymm_ymm_ymmm256,
		EVEX_Vaesenclast_xmm_xmm_xmmm128,
		EVEX_Vaesenclast_ymm_ymm_ymmm256,
		EVEX_Vaesenclast_zmm_zmm_zmmm512,
		Aesdec_xmm_xmmm128,
		VEX_Vaesdec_xmm_xmm_xmmm128,
		VEX_Vaesdec_ymm_ymm_ymmm256,
		EVEX_Vaesdec_xmm_xmm_xmmm128,
		EVEX_Vaesdec_ymm_ymm_ymmm256,
		EVEX_Vaesdec_zmm_zmm_zmmm512,
		Aesdeclast_xmm_xmmm128,
		VEX_Vaesdeclast_xmm_xmm_xmmm128,
		VEX_Vaesdeclast_ymm_ymm_ymmm256,
		EVEX_Vaesdeclast_xmm_xmm_xmmm128,
		EVEX_Vaesdeclast_ymm_ymm_ymmm256,
		EVEX_Vaesdeclast_zmm_zmm_zmmm512,
		Movbe_r16_m16,
		Movbe_r32_m32,
		Movbe_r64_m64,
		Crc32_r32_rm8,
		Crc32_r64_rm8,
		Movbe_m16_r16,
		Movbe_m32_r32,
		Movbe_m64_r64,
		Crc32_r32_rm16,
		Crc32_r32_rm32,
		Crc32_r64_rm64,
		VEX_Andn_r32_r32_rm32,
		VEX_Andn_r64_r64_rm64,
		VEX_Blsr_r32_rm32,
		VEX_Blsr_r64_rm64,
		VEX_Blsmsk_r32_rm32,
		VEX_Blsmsk_r64_rm64,
		VEX_Blsi_r32_rm32,
		VEX_Blsi_r64_rm64,
		VEX_Bzhi_r32_rm32_r32,
		VEX_Bzhi_r64_rm64_r64,
		Wrussd_m32_r32,
		Wrussq_m64_r64,
		VEX_Pext_r32_r32_rm32,
		VEX_Pext_r64_r64_rm64,
		VEX_Pdep_r32_r32_rm32,
		VEX_Pdep_r64_r64_rm64,
		Wrssd_m32_r32,
		Wrssq_m64_r64,
		Adcx_r32_rm32,
		Adcx_r64_rm64,
		Adox_r32_rm32,
		Adox_r64_rm64,
		VEX_Mulx_r32_r32_rm32,
		VEX_Mulx_r64_r64_rm64,
		VEX_Bextr_r32_rm32_r32,
		VEX_Bextr_r64_rm64_r64,
		VEX_Shlx_r32_rm32_r32,
		VEX_Shlx_r64_rm64_r64,
		VEX_Sarx_r32_rm32_r32,
		VEX_Sarx_r64_rm64_r64,
		VEX_Shrx_r32_rm32_r32,
		VEX_Shrx_r64_rm64_r64,
		Movdir64b_r16_m512,
		Movdir64b_r32_m512,
		Movdir64b_r64_m512,
		Enqcmds_r16_m512,
		Enqcmds_r32_m512,
		Enqcmds_r64_m512,
		Enqcmd_r16_m512,
		Enqcmd_r32_m512,
		Enqcmd_r64_m512,
		Movdiri_m32_r32,
		Movdiri_m64_r64,
		VEX_Vpermq_ymm_ymmm256_imm8,
		EVEX_Vpermq_ymm_k1z_ymmm256b64_imm8,
		EVEX_Vpermq_zmm_k1z_zmmm512b64_imm8,
		VEX_Vpermpd_ymm_ymmm256_imm8,
		EVEX_Vpermpd_ymm_k1z_ymmm256b64_imm8,
		EVEX_Vpermpd_zmm_k1z_zmmm512b64_imm8,
		VEX_Vpblendd_xmm_xmm_xmmm128_imm8,
		VEX_Vpblendd_ymm_ymm_ymmm256_imm8,
		EVEX_Valignd_xmm_k1z_xmm_xmmm128b32_imm8,
		EVEX_Valignd_ymm_k1z_ymm_ymmm256b32_imm8,
		EVEX_Valignd_zmm_k1z_zmm_zmmm512b32_imm8,
		EVEX_Valignq_xmm_k1z_xmm_xmmm128b64_imm8,
		EVEX_Valignq_ymm_k1z_ymm_ymmm256b64_imm8,
		EVEX_Valignq_zmm_k1z_zmm_zmmm512b64_imm8,
		VEX_Vpermilps_xmm_xmmm128_imm8,
		VEX_Vpermilps_ymm_ymmm256_imm8,
		EVEX_Vpermilps_xmm_k1z_xmmm128b32_imm8,
		EVEX_Vpermilps_ymm_k1z_ymmm256b32_imm8,
		EVEX_Vpermilps_zmm_k1z_zmmm512b32_imm8,
		VEX_Vpermilpd_xmm_xmmm128_imm8,
		VEX_Vpermilpd_ymm_ymmm256_imm8,
		EVEX_Vpermilpd_xmm_k1z_xmmm128b64_imm8,
		EVEX_Vpermilpd_ymm_k1z_ymmm256b64_imm8,
		EVEX_Vpermilpd_zmm_k1z_zmmm512b64_imm8,
		VEX_Vperm2f128_ymm_ymm_ymmm256_imm8,
		Roundps_xmm_xmmm128_imm8,
		VEX_Vroundps_xmm_xmmm128_imm8,
		VEX_Vroundps_ymm_ymmm256_imm8,
		EVEX_Vrndscaleps_xmm_k1z_xmmm128b32_imm8,
		EVEX_Vrndscaleps_ymm_k1z_ymmm256b32_imm8,
		EVEX_Vrndscaleps_zmm_k1z_zmmm512b32_imm8_sae,
		Roundpd_xmm_xmmm128_imm8,
		VEX_Vroundpd_xmm_xmmm128_imm8,
		VEX_Vroundpd_ymm_ymmm256_imm8,
		EVEX_Vrndscalepd_xmm_k1z_xmmm128b64_imm8,
		EVEX_Vrndscalepd_ymm_k1z_ymmm256b64_imm8,
		EVEX_Vrndscalepd_zmm_k1z_zmmm512b64_imm8_sae,
		Roundss_xmm_xmmm32_imm8,
		VEX_Vroundss_xmm_xmm_xmmm32_imm8,
		EVEX_Vrndscaless_xmm_k1z_xmm_xmmm32_imm8_sae,
		Roundsd_xmm_xmmm64_imm8,
		VEX_Vroundsd_xmm_xmm_xmmm64_imm8,
		EVEX_Vrndscalesd_xmm_k1z_xmm_xmmm64_imm8_sae,
		Blendps_xmm_xmmm128_imm8,
		VEX_Vblendps_xmm_xmm_xmmm128_imm8,
		VEX_Vblendps_ymm_ymm_ymmm256_imm8,
		Blendpd_xmm_xmmm128_imm8,
		VEX_Vblendpd_xmm_xmm_xmmm128_imm8,
		VEX_Vblendpd_ymm_ymm_ymmm256_imm8,
		Pblendw_xmm_xmmm128_imm8,
		VEX_Vpblendw_xmm_xmm_xmmm128_imm8,
		VEX_Vpblendw_ymm_ymm_ymmm256_imm8,
		Palignr_mm_mmm64_imm8,
		Palignr_xmm_xmmm128_imm8,
		VEX_Vpalignr_xmm_xmm_xmmm128_imm8,
		VEX_Vpalignr_ymm_ymm_ymmm256_imm8,
		EVEX_Vpalignr_xmm_k1z_xmm_xmmm128_imm8,
		EVEX_Vpalignr_ymm_k1z_ymm_ymmm256_imm8,
		EVEX_Vpalignr_zmm_k1z_zmm_zmmm512_imm8,
		Pextrb_r32m8_xmm_imm8,
		Pextrb_r64m8_xmm_imm8,
		VEX_Vpextrb_r32m8_xmm_imm8,
		VEX_Vpextrb_r64m8_xmm_imm8,
		EVEX_Vpextrb_r32m8_xmm_imm8,
		EVEX_Vpextrb_r64m8_xmm_imm8,
		Pextrw_r32m16_xmm_imm8,
		Pextrw_r64m16_xmm_imm8,
		VEX_Vpextrw_r32m16_xmm_imm8,
		VEX_Vpextrw_r64m16_xmm_imm8,
		EVEX_Vpextrw_r32m16_xmm_imm8,
		EVEX_Vpextrw_r64m16_xmm_imm8,
		Pextrd_rm32_xmm_imm8,
		Pextrq_rm64_xmm_imm8,
		VEX_Vpextrd_rm32_xmm_imm8,
		VEX_Vpextrq_rm64_xmm_imm8,
		EVEX_Vpextrd_rm32_xmm_imm8,
		EVEX_Vpextrq_rm64_xmm_imm8,
		Extractps_rm32_xmm_imm8,
		Extractps_r64m32_xmm_imm8,
		VEX_Vextractps_rm32_xmm_imm8,
		VEX_Vextractps_r64m32_xmm_imm8,
		EVEX_Vextractps_rm32_xmm_imm8,
		EVEX_Vextractps_r64m32_xmm_imm8,
		VEX_Vinsertf128_ymm_ymm_xmmm128_imm8,
		EVEX_Vinsertf32x4_ymm_k1z_ymm_xmmm128_imm8,
		EVEX_Vinsertf32x4_zmm_k1z_zmm_xmmm128_imm8,
		EVEX_Vinsertf64x2_ymm_k1z_ymm_xmmm128_imm8,
		EVEX_Vinsertf64x2_zmm_k1z_zmm_xmmm128_imm8,
		VEX_Vextractf128_xmmm128_ymm_imm8,
		EVEX_Vextractf32x4_xmmm128_k1z_ymm_imm8,
		EVEX_Vextractf32x4_xmmm128_k1z_zmm_imm8,
		EVEX_Vextractf64x2_xmmm128_k1z_ymm_imm8,
		EVEX_Vextractf64x2_xmmm128_k1z_zmm_imm8,
		EVEX_Vinsertf32x8_zmm_k1z_zmm_ymmm256_imm8,
		EVEX_Vinsertf64x4_zmm_k1z_zmm_ymmm256_imm8,
		EVEX_Vextractf32x8_ymmm256_k1z_zmm_imm8,
		EVEX_Vextractf64x4_ymmm256_k1z_zmm_imm8,
		VEX_Vcvtps2ph_xmmm64_xmm_imm8,
		VEX_Vcvtps2ph_xmmm128_ymm_imm8,
		EVEX_Vcvtps2ph_xmmm64_k1z_xmm_imm8,
		EVEX_Vcvtps2ph_xmmm128_k1z_ymm_imm8,
		EVEX_Vcvtps2ph_ymmm256_k1z_zmm_imm8_sae,
		EVEX_Vpcmpud_kr_k1_xmm_xmmm128b32_imm8,
		EVEX_Vpcmpud_kr_k1_ymm_ymmm256b32_imm8,
		EVEX_Vpcmpud_kr_k1_zmm_zmmm512b32_imm8,
		EVEX_Vpcmpuq_kr_k1_xmm_xmmm128b64_imm8,
		EVEX_Vpcmpuq_kr_k1_ymm_ymmm256b64_imm8,
		EVEX_Vpcmpuq_kr_k1_zmm_zmmm512b64_imm8,
		EVEX_Vpcmpd_kr_k1_xmm_xmmm128b32_imm8,
		EVEX_Vpcmpd_kr_k1_ymm_ymmm256b32_imm8,
		EVEX_Vpcmpd_kr_k1_zmm_zmmm512b32_imm8,
		EVEX_Vpcmpq_kr_k1_xmm_xmmm128b64_imm8,
		EVEX_Vpcmpq_kr_k1_ymm_ymmm256b64_imm8,
		EVEX_Vpcmpq_kr_k1_zmm_zmmm512b64_imm8,
		Pinsrb_xmm_r32m8_imm8,
		Pinsrb_xmm_r64m8_imm8,
		VEX_Vpinsrb_xmm_xmm_r32m8_imm8,
		VEX_Vpinsrb_xmm_xmm_r64m8_imm8,
		EVEX_Vpinsrb_xmm_xmm_r32m8_imm8,
		EVEX_Vpinsrb_xmm_xmm_r64m8_imm8,
		Insertps_xmm_xmmm32_imm8,
		VEX_Vinsertps_xmm_xmm_xmmm32_imm8,
		EVEX_Vinsertps_xmm_xmm_xmmm32_imm8,
		Pinsrd_xmm_rm32_imm8,
		Pinsrq_xmm_rm64_imm8,
		VEX_Vpinsrd_xmm_xmm_rm32_imm8,
		VEX_Vpinsrq_xmm_xmm_rm64_imm8,
		EVEX_Vpinsrd_xmm_xmm_rm32_imm8,
		EVEX_Vpinsrq_xmm_xmm_rm64_imm8,
		EVEX_Vshuff32x4_ymm_k1z_ymm_ymmm256b32_imm8,
		EVEX_Vshuff32x4_zmm_k1z_zmm_zmmm512b32_imm8,
		EVEX_Vshuff64x2_ymm_k1z_ymm_ymmm256b64_imm8,
		EVEX_Vshuff64x2_zmm_k1z_zmm_zmmm512b64_imm8,
		EVEX_Vpternlogd_xmm_k1z_xmm_xmmm128b32_imm8,
		EVEX_Vpternlogd_ymm_k1z_ymm_ymmm256b32_imm8,
		EVEX_Vpternlogd_zmm_k1z_zmm_zmmm512b32_imm8,
		EVEX_Vpternlogq_xmm_k1z_xmm_xmmm128b64_imm8,
		EVEX_Vpternlogq_ymm_k1z_ymm_ymmm256b64_imm8,
		EVEX_Vpternlogq_zmm_k1z_zmm_zmmm512b64_imm8,
		EVEX_Vgetmantps_xmm_k1z_xmmm128b32_imm8,
		EVEX_Vgetmantps_ymm_k1z_ymmm256b32_imm8,
		EVEX_Vgetmantps_zmm_k1z_zmmm512b32_imm8_sae,
		EVEX_Vgetmantpd_xmm_k1z_xmmm128b64_imm8,
		EVEX_Vgetmantpd_ymm_k1z_ymmm256b64_imm8,
		EVEX_Vgetmantpd_zmm_k1z_zmmm512b64_imm8_sae,
		EVEX_Vgetmantss_xmm_k1z_xmm_xmmm32_imm8_sae,
		EVEX_Vgetmantsd_xmm_k1z_xmm_xmmm64_imm8_sae,
		VEX_Kshiftrb_kr_kr_imm8,
		VEX_Kshiftrw_kr_kr_imm8,
		VEX_Kshiftrd_kr_kr_imm8,
		VEX_Kshiftrq_kr_kr_imm8,
		VEX_Kshiftlb_kr_kr_imm8,
		VEX_Kshiftlw_kr_kr_imm8,
		VEX_Kshiftld_kr_kr_imm8,
		VEX_Kshiftlq_kr_kr_imm8,
		VEX_Vinserti128_ymm_ymm_xmmm128_imm8,
		EVEX_Vinserti32x4_ymm_k1z_ymm_xmmm128_imm8,
		EVEX_Vinserti32x4_zmm_k1z_zmm_xmmm128_imm8,
		EVEX_Vinserti64x2_ymm_k1z_ymm_xmmm128_imm8,
		EVEX_Vinserti64x2_zmm_k1z_zmm_xmmm128_imm8,
		VEX_Vextracti128_xmmm128_ymm_imm8,
		EVEX_Vextracti32x4_xmmm128_k1z_ymm_imm8,
		EVEX_Vextracti32x4_xmmm128_k1z_zmm_imm8,
		EVEX_Vextracti64x2_xmmm128_k1z_ymm_imm8,
		EVEX_Vextracti64x2_xmmm128_k1z_zmm_imm8,
		EVEX_Vinserti32x8_zmm_k1z_zmm_ymmm256_imm8,
		EVEX_Vinserti64x4_zmm_k1z_zmm_ymmm256_imm8,
		EVEX_Vextracti32x8_ymmm256_k1z_zmm_imm8,
		EVEX_Vextracti64x4_ymmm256_k1z_zmm_imm8,
		EVEX_Vpcmpub_kr_k1_xmm_xmmm128_imm8,
		EVEX_Vpcmpub_kr_k1_ymm_ymmm256_imm8,
		EVEX_Vpcmpub_kr_k1_zmm_zmmm512_imm8,
		EVEX_Vpcmpuw_kr_k1_xmm_xmmm128_imm8,
		EVEX_Vpcmpuw_kr_k1_ymm_ymmm256_imm8,
		EVEX_Vpcmpuw_kr_k1_zmm_zmmm512_imm8,
		EVEX_Vpcmpb_kr_k1_xmm_xmmm128_imm8,
		EVEX_Vpcmpb_kr_k1_ymm_ymmm256_imm8,
		EVEX_Vpcmpb_kr_k1_zmm_zmmm512_imm8,
		EVEX_Vpcmpw_kr_k1_xmm_xmmm128_imm8,
		EVEX_Vpcmpw_kr_k1_ymm_ymmm256_imm8,
		EVEX_Vpcmpw_kr_k1_zmm_zmmm512_imm8,
		Dpps_xmm_xmmm128_imm8,
		VEX_Vdpps_xmm_xmm_xmmm128_imm8,
		VEX_Vdpps_ymm_ymm_ymmm256_imm8,
		Dppd_xmm_xmmm128_imm8,
		VEX_Vdppd_xmm_xmm_xmmm128_imm8,
		Mpsadbw_xmm_xmmm128_imm8,
		VEX_Vmpsadbw_xmm_xmm_xmmm128_imm8,
		VEX_Vmpsadbw_ymm_ymm_ymmm256_imm8,
		EVEX_Vdbpsadbw_xmm_k1z_xmm_xmmm128_imm8,
		EVEX_Vdbpsadbw_ymm_k1z_ymm_ymmm256_imm8,
		EVEX_Vdbpsadbw_zmm_k1z_zmm_zmmm512_imm8,
		EVEX_Vshufi32x4_ymm_k1z_ymm_ymmm256b32_imm8,
		EVEX_Vshufi32x4_zmm_k1z_zmm_zmmm512b32_imm8,
		EVEX_Vshufi64x2_ymm_k1z_ymm_ymmm256b64_imm8,
		EVEX_Vshufi64x2_zmm_k1z_zmm_zmmm512b64_imm8,
		Pclmulqdq_xmm_xmmm128_imm8,
		VEX_Vpclmulqdq_xmm_xmm_xmmm128_imm8,
		VEX_Vpclmulqdq_ymm_ymm_ymmm256_imm8,
		EVEX_Vpclmulqdq_xmm_xmm_xmmm128_imm8,
		EVEX_Vpclmulqdq_ymm_ymm_ymmm256_imm8,
		EVEX_Vpclmulqdq_zmm_zmm_zmmm512_imm8,
		VEX_Vperm2i128_ymm_ymm_ymmm256_imm8,
		VEX_Vpermil2ps_xmm_xmm_xmmm128_xmm_imm4,
		VEX_Vpermil2ps_ymm_ymm_ymmm256_ymm_imm4,
		VEX_Vpermil2ps_xmm_xmm_xmm_xmmm128_imm4,
		VEX_Vpermil2ps_ymm_ymm_ymm_ymmm256_imm4,
		VEX_Vpermil2pd_xmm_xmm_xmmm128_xmm_imm4,
		VEX_Vpermil2pd_ymm_ymm_ymmm256_ymm_imm4,
		VEX_Vpermil2pd_xmm_xmm_xmm_xmmm128_imm4,
		VEX_Vpermil2pd_ymm_ymm_ymm_ymmm256_imm4,
		VEX_Vblendvps_xmm_xmm_xmmm128_xmm,
		VEX_Vblendvps_ymm_ymm_ymmm256_ymm,
		VEX_Vblendvpd_xmm_xmm_xmmm128_xmm,
		VEX_Vblendvpd_ymm_ymm_ymmm256_ymm,
		VEX_Vpblendvb_xmm_xmm_xmmm128_xmm,
		VEX_Vpblendvb_ymm_ymm_ymmm256_ymm,
		EVEX_Vrangeps_xmm_k1z_xmm_xmmm128b32_imm8,
		EVEX_Vrangeps_ymm_k1z_ymm_ymmm256b32_imm8,
		EVEX_Vrangeps_zmm_k1z_zmm_zmmm512b32_imm8_sae,
		EVEX_Vrangepd_xmm_k1z_xmm_xmmm128b64_imm8,
		EVEX_Vrangepd_ymm_k1z_ymm_ymmm256b64_imm8,
		EVEX_Vrangepd_zmm_k1z_zmm_zmmm512b64_imm8_sae,
		EVEX_Vrangess_xmm_k1z_xmm_xmmm32_imm8_sae,
		EVEX_Vrangesd_xmm_k1z_xmm_xmmm64_imm8_sae,
		EVEX_Vfixupimmps_xmm_k1z_xmm_xmmm128b32_imm8,
		EVEX_Vfixupimmps_ymm_k1z_ymm_ymmm256b32_imm8,
		EVEX_Vfixupimmps_zmm_k1z_zmm_zmmm512b32_imm8_sae,
		EVEX_Vfixupimmpd_xmm_k1z_xmm_xmmm128b64_imm8,
		EVEX_Vfixupimmpd_ymm_k1z_ymm_ymmm256b64_imm8,
		EVEX_Vfixupimmpd_zmm_k1z_zmm_zmmm512b64_imm8_sae,
		EVEX_Vfixupimmss_xmm_k1z_xmm_xmmm32_imm8_sae,
		EVEX_Vfixupimmsd_xmm_k1z_xmm_xmmm64_imm8_sae,
		EVEX_Vreduceps_xmm_k1z_xmmm128b32_imm8,
		EVEX_Vreduceps_ymm_k1z_ymmm256b32_imm8,
		EVEX_Vreduceps_zmm_k1z_zmmm512b32_imm8_sae,
		EVEX_Vreducepd_xmm_k1z_xmmm128b64_imm8,
		EVEX_Vreducepd_ymm_k1z_ymmm256b64_imm8,
		EVEX_Vreducepd_zmm_k1z_zmmm512b64_imm8_sae,
		EVEX_Vreducess_xmm_k1z_xmm_xmmm32_imm8_sae,
		EVEX_Vreducesd_xmm_k1z_xmm_xmmm64_imm8_sae,
		VEX_Vfmaddsubps_xmm_xmm_xmmm128_xmm,
		VEX_Vfmaddsubps_ymm_ymm_ymmm256_ymm,
		VEX_Vfmaddsubps_xmm_xmm_xmm_xmmm128,
		VEX_Vfmaddsubps_ymm_ymm_ymm_ymmm256,
		VEX_Vfmaddsubpd_xmm_xmm_xmmm128_xmm,
		VEX_Vfmaddsubpd_ymm_ymm_ymmm256_ymm,
		VEX_Vfmaddsubpd_xmm_xmm_xmm_xmmm128,
		VEX_Vfmaddsubpd_ymm_ymm_ymm_ymmm256,
		VEX_Vfmsubaddps_xmm_xmm_xmmm128_xmm,
		VEX_Vfmsubaddps_ymm_ymm_ymmm256_ymm,
		VEX_Vfmsubaddps_xmm_xmm_xmm_xmmm128,
		VEX_Vfmsubaddps_ymm_ymm_ymm_ymmm256,
		VEX_Vfmsubaddpd_xmm_xmm_xmmm128_xmm,
		VEX_Vfmsubaddpd_ymm_ymm_ymmm256_ymm,
		VEX_Vfmsubaddpd_xmm_xmm_xmm_xmmm128,
		VEX_Vfmsubaddpd_ymm_ymm_ymm_ymmm256,
		Pcmpestrm_xmm_xmmm128_imm8,
		Pcmpestrm64_xmm_xmmm128_imm8,
		VEX_Vpcmpestrm_xmm_xmmm128_imm8,
		VEX_Vpcmpestrm64_xmm_xmmm128_imm8,
		Pcmpestri_xmm_xmmm128_imm8,
		Pcmpestri64_xmm_xmmm128_imm8,
		VEX_Vpcmpestri_xmm_xmmm128_imm8,
		VEX_Vpcmpestri64_xmm_xmmm128_imm8,
		Pcmpistrm_xmm_xmmm128_imm8,
		VEX_Vpcmpistrm_xmm_xmmm128_imm8,
		Pcmpistri_xmm_xmmm128_imm8,
		VEX_Vpcmpistri_xmm_xmmm128_imm8,
		EVEX_Vfpclassps_kr_k1_xmmm128b32_imm8,
		EVEX_Vfpclassps_kr_k1_ymmm256b32_imm8,
		EVEX_Vfpclassps_kr_k1_zmmm512b32_imm8,
		EVEX_Vfpclasspd_kr_k1_xmmm128b64_imm8,
		EVEX_Vfpclasspd_kr_k1_ymmm256b64_imm8,
		EVEX_Vfpclasspd_kr_k1_zmmm512b64_imm8,
		EVEX_Vfpclassss_kr_k1_xmmm32_imm8,
		EVEX_Vfpclasssd_kr_k1_xmmm64_imm8,
		VEX_Vfmaddps_xmm_xmm_xmmm128_xmm,
		VEX_Vfmaddps_ymm_ymm_ymmm256_ymm,
		VEX_Vfmaddps_xmm_xmm_xmm_xmmm128,
		VEX_Vfmaddps_ymm_ymm_ymm_ymmm256,
		VEX_Vfmaddpd_xmm_xmm_xmmm128_xmm,
		VEX_Vfmaddpd_ymm_ymm_ymmm256_ymm,
		VEX_Vfmaddpd_xmm_xmm_xmm_xmmm128,
		VEX_Vfmaddpd_ymm_ymm_ymm_ymmm256,
		VEX_Vfmaddss_xmm_xmm_xmmm32_xmm,
		VEX_Vfmaddss_xmm_xmm_xmm_xmmm32,
		VEX_Vfmaddsd_xmm_xmm_xmmm64_xmm,
		VEX_Vfmaddsd_xmm_xmm_xmm_xmmm64,
		VEX_Vfmsubps_xmm_xmm_xmmm128_xmm,
		VEX_Vfmsubps_ymm_ymm_ymmm256_ymm,
		VEX_Vfmsubps_xmm_xmm_xmm_xmmm128,
		VEX_Vfmsubps_ymm_ymm_ymm_ymmm256,
		VEX_Vfmsubpd_xmm_xmm_xmmm128_xmm,
		VEX_Vfmsubpd_ymm_ymm_ymmm256_ymm,
		VEX_Vfmsubpd_xmm_xmm_xmm_xmmm128,
		VEX_Vfmsubpd_ymm_ymm_ymm_ymmm256,
		VEX_Vfmsubss_xmm_xmm_xmmm32_xmm,
		VEX_Vfmsubss_xmm_xmm_xmm_xmmm32,
		VEX_Vfmsubsd_xmm_xmm_xmmm64_xmm,
		VEX_Vfmsubsd_xmm_xmm_xmm_xmmm64,
		EVEX_Vpshldw_xmm_k1z_xmm_xmmm128_imm8,
		EVEX_Vpshldw_ymm_k1z_ymm_ymmm256_imm8,
		EVEX_Vpshldw_zmm_k1z_zmm_zmmm512_imm8,
		EVEX_Vpshldd_xmm_k1z_xmm_xmmm128b32_imm8,
		EVEX_Vpshldd_ymm_k1z_ymm_ymmm256b32_imm8,
		EVEX_Vpshldd_zmm_k1z_zmm_zmmm512b32_imm8,
		EVEX_Vpshldq_xmm_k1z_xmm_xmmm128b64_imm8,
		EVEX_Vpshldq_ymm_k1z_ymm_ymmm256b64_imm8,
		EVEX_Vpshldq_zmm_k1z_zmm_zmmm512b64_imm8,
		EVEX_Vpshrdw_xmm_k1z_xmm_xmmm128_imm8,
		EVEX_Vpshrdw_ymm_k1z_ymm_ymmm256_imm8,
		EVEX_Vpshrdw_zmm_k1z_zmm_zmmm512_imm8,
		EVEX_Vpshrdd_xmm_k1z_xmm_xmmm128b32_imm8,
		EVEX_Vpshrdd_ymm_k1z_ymm_ymmm256b32_imm8,
		EVEX_Vpshrdd_zmm_k1z_zmm_zmmm512b32_imm8,
		EVEX_Vpshrdq_xmm_k1z_xmm_xmmm128b64_imm8,
		EVEX_Vpshrdq_ymm_k1z_ymm_ymmm256b64_imm8,
		EVEX_Vpshrdq_zmm_k1z_zmm_zmmm512b64_imm8,
		VEX_Vfnmaddps_xmm_xmm_xmmm128_xmm,
		VEX_Vfnmaddps_ymm_ymm_ymmm256_ymm,
		VEX_Vfnmaddps_xmm_xmm_xmm_xmmm128,
		VEX_Vfnmaddps_ymm_ymm_ymm_ymmm256,
		VEX_Vfnmaddpd_xmm_xmm_xmmm128_xmm,
		VEX_Vfnmaddpd_ymm_ymm_ymmm256_ymm,
		VEX_Vfnmaddpd_xmm_xmm_xmm_xmmm128,
		VEX_Vfnmaddpd_ymm_ymm_ymm_ymmm256,
		VEX_Vfnmaddss_xmm_xmm_xmmm32_xmm,
		VEX_Vfnmaddss_xmm_xmm_xmm_xmmm32,
		VEX_Vfnmaddsd_xmm_xmm_xmmm64_xmm,
		VEX_Vfnmaddsd_xmm_xmm_xmm_xmmm64,
		VEX_Vfnmsubps_xmm_xmm_xmmm128_xmm,
		VEX_Vfnmsubps_ymm_ymm_ymmm256_ymm,
		VEX_Vfnmsubps_xmm_xmm_xmm_xmmm128,
		VEX_Vfnmsubps_ymm_ymm_ymm_ymmm256,
		VEX_Vfnmsubpd_xmm_xmm_xmmm128_xmm,
		VEX_Vfnmsubpd_ymm_ymm_ymmm256_ymm,
		VEX_Vfnmsubpd_xmm_xmm_xmm_xmmm128,
		VEX_Vfnmsubpd_ymm_ymm_ymm_ymmm256,
		VEX_Vfnmsubss_xmm_xmm_xmmm32_xmm,
		VEX_Vfnmsubss_xmm_xmm_xmm_xmmm32,
		VEX_Vfnmsubsd_xmm_xmm_xmmm64_xmm,
		VEX_Vfnmsubsd_xmm_xmm_xmm_xmmm64,
		Sha1rnds4_xmm_xmmm128_imm8,
		Gf2p8affineqb_xmm_xmmm128_imm8,
		VEX_Vgf2p8affineqb_xmm_xmm_xmmm128_imm8,
		VEX_Vgf2p8affineqb_ymm_ymm_ymmm256_imm8,
		EVEX_Vgf2p8affineqb_xmm_k1z_xmm_xmmm128b64_imm8,
		EVEX_Vgf2p8affineqb_ymm_k1z_ymm_ymmm256b64_imm8,
		EVEX_Vgf2p8affineqb_zmm_k1z_zmm_zmmm512b64_imm8,
		Gf2p8affineinvqb_xmm_xmmm128_imm8,
		VEX_Vgf2p8affineinvqb_xmm_xmm_xmmm128_imm8,
		VEX_Vgf2p8affineinvqb_ymm_ymm_ymmm256_imm8,
		EVEX_Vgf2p8affineinvqb_xmm_k1z_xmm_xmmm128b64_imm8,
		EVEX_Vgf2p8affineinvqb_ymm_k1z_ymm_ymmm256b64_imm8,
		EVEX_Vgf2p8affineinvqb_zmm_k1z_zmm_zmmm512b64_imm8,
		Aeskeygenassist_xmm_xmmm128_imm8,
		VEX_Vaeskeygenassist_xmm_xmmm128_imm8,
		VEX_Rorx_r32_rm32_imm8,
		VEX_Rorx_r64_rm64_imm8,
		XOP_Vpmacssww_xmm_xmm_xmmm128_xmm,
		XOP_Vpmacsswd_xmm_xmm_xmmm128_xmm,
		XOP_Vpmacssdql_xmm_xmm_xmmm128_xmm,
		XOP_Vpmacssdd_xmm_xmm_xmmm128_xmm,
		XOP_Vpmacssdqh_xmm_xmm_xmmm128_xmm,
		XOP_Vpmacsww_xmm_xmm_xmmm128_xmm,
		XOP_Vpmacswd_xmm_xmm_xmmm128_xmm,
		XOP_Vpmacsdql_xmm_xmm_xmmm128_xmm,
		XOP_Vpmacsdd_xmm_xmm_xmmm128_xmm,
		XOP_Vpmacsdqh_xmm_xmm_xmmm128_xmm,
		XOP_Vpcmov_xmm_xmm_xmmm128_xmm,
		XOP_Vpcmov_ymm_ymm_ymmm256_ymm,
		XOP_Vpcmov_xmm_xmm_xmm_xmmm128,
		XOP_Vpcmov_ymm_ymm_ymm_ymmm256,
		XOP_Vpperm_xmm_xmm_xmmm128_xmm,
		XOP_Vpperm_xmm_xmm_xmm_xmmm128,
		XOP_Vpmadcsswd_xmm_xmm_xmmm128_xmm,
		XOP_Vpmadcswd_xmm_xmm_xmmm128_xmm,
		XOP_Vprotb_xmm_xmmm128_imm8,
		XOP_Vprotw_xmm_xmmm128_imm8,
		XOP_Vprotd_xmm_xmmm128_imm8,
		XOP_Vprotq_xmm_xmmm128_imm8,
		XOP_Vpcomb_xmm_xmm_xmmm128_imm8,
		XOP_Vpcomw_xmm_xmm_xmmm128_imm8,
		XOP_Vpcomd_xmm_xmm_xmmm128_imm8,
		XOP_Vpcomq_xmm_xmm_xmmm128_imm8,
		XOP_Vpcomub_xmm_xmm_xmmm128_imm8,
		XOP_Vpcomuw_xmm_xmm_xmmm128_imm8,
		XOP_Vpcomud_xmm_xmm_xmmm128_imm8,
		XOP_Vpcomuq_xmm_xmm_xmmm128_imm8,
		XOP_Blcfill_r32_rm32,
		XOP_Blcfill_r64_rm64,
		XOP_Blsfill_r32_rm32,
		XOP_Blsfill_r64_rm64,
		XOP_Blcs_r32_rm32,
		XOP_Blcs_r64_rm64,
		XOP_Tzmsk_r32_rm32,
		XOP_Tzmsk_r64_rm64,
		XOP_Blcic_r32_rm32,
		XOP_Blcic_r64_rm64,
		XOP_Blsic_r32_rm32,
		XOP_Blsic_r64_rm64,
		XOP_T1mskc_r32_rm32,
		XOP_T1mskc_r64_rm64,
		XOP_Blcmsk_r32_rm32,
		XOP_Blcmsk_r64_rm64,
		XOP_Blci_r32_rm32,
		XOP_Blci_r64_rm64,
		XOP_Llwpcb_r32,
		XOP_Llwpcb_r64,
		XOP_Slwpcb_r32,
		XOP_Slwpcb_r64,
		XOP_Vfrczps_xmm_xmmm128,
		XOP_Vfrczps_ymm_ymmm256,
		XOP_Vfrczpd_xmm_xmmm128,
		XOP_Vfrczpd_ymm_ymmm256,
		XOP_Vfrczss_xmm_xmmm32,
		XOP_Vfrczsd_xmm_xmmm64,
		XOP_Vprotb_xmm_xmmm128_xmm,
		XOP_Vprotb_xmm_xmm_xmmm128,
		XOP_Vprotw_xmm_xmmm128_xmm,
		XOP_Vprotw_xmm_xmm_xmmm128,
		XOP_Vprotd_xmm_xmmm128_xmm,
		XOP_Vprotd_xmm_xmm_xmmm128,
		XOP_Vprotq_xmm_xmmm128_xmm,
		XOP_Vprotq_xmm_xmm_xmmm128,
		XOP_Vpshlb_xmm_xmmm128_xmm,
		XOP_Vpshlb_xmm_xmm_xmmm128,
		XOP_Vpshlw_xmm_xmmm128_xmm,
		XOP_Vpshlw_xmm_xmm_xmmm128,
		XOP_Vpshld_xmm_xmmm128_xmm,
		XOP_Vpshld_xmm_xmm_xmmm128,
		XOP_Vpshlq_xmm_xmmm128_xmm,
		XOP_Vpshlq_xmm_xmm_xmmm128,
		XOP_Vpshab_xmm_xmmm128_xmm,
		XOP_Vpshab_xmm_xmm_xmmm128,
		XOP_Vpshaw_xmm_xmmm128_xmm,
		XOP_Vpshaw_xmm_xmm_xmmm128,
		XOP_Vpshad_xmm_xmmm128_xmm,
		XOP_Vpshad_xmm_xmm_xmmm128,
		XOP_Vpshaq_xmm_xmmm128_xmm,
		XOP_Vpshaq_xmm_xmm_xmmm128,
		XOP_Vphaddbw_xmm_xmmm128,
		XOP_Vphaddbd_xmm_xmmm128,
		XOP_Vphaddbq_xmm_xmmm128,
		XOP_Vphaddwd_xmm_xmmm128,
		XOP_Vphaddwq_xmm_xmmm128,
		XOP_Vphadddq_xmm_xmmm128,
		XOP_Vphaddubw_xmm_xmmm128,
		XOP_Vphaddubd_xmm_xmmm128,
		XOP_Vphaddubq_xmm_xmmm128,
		XOP_Vphadduwd_xmm_xmmm128,
		XOP_Vphadduwq_xmm_xmmm128,
		XOP_Vphaddudq_xmm_xmmm128,
		XOP_Vphsubbw_xmm_xmmm128,
		XOP_Vphsubwd_xmm_xmmm128,
		XOP_Vphsubdq_xmm_xmmm128,
		XOP_Bextr_r32_rm32_imm32,
		XOP_Bextr_r64_rm64_imm32,
		XOP_Lwpins_r32_rm32_imm32,
		XOP_Lwpins_r64_rm32_imm32,
		XOP_Lwpval_r32_rm32_imm32,
		XOP_Lwpval_r64_rm32_imm32,
		D3NOW_Pi2fw_mm_mmm64,
		D3NOW_Pi2fd_mm_mmm64,
		D3NOW_Pf2iw_mm_mmm64,
		D3NOW_Pf2id_mm_mmm64,
		D3NOW_Pfrcpv_mm_mmm64,
		D3NOW_Pfrsqrtv_mm_mmm64,
		D3NOW_Pfnacc_mm_mmm64,
		D3NOW_Pfpnacc_mm_mmm64,
		D3NOW_Pfcmpge_mm_mmm64,
		D3NOW_Pfmin_mm_mmm64,
		D3NOW_Pfrcp_mm_mmm64,
		D3NOW_Pfrsqrt_mm_mmm64,
		D3NOW_Pfsub_mm_mmm64,
		D3NOW_Pfadd_mm_mmm64,
		D3NOW_Pfcmpgt_mm_mmm64,
		D3NOW_Pfmax_mm_mmm64,
		D3NOW_Pfrcpit1_mm_mmm64,
		D3NOW_Pfrsqit1_mm_mmm64,
		D3NOW_Pfsubr_mm_mmm64,
		D3NOW_Pfacc_mm_mmm64,
		D3NOW_Pfcmpeq_mm_mmm64,
		D3NOW_Pfmul_mm_mmm64,
		D3NOW_Pfrcpit2_mm_mmm64,
		D3NOW_Pmulhrw_mm_mmm64,
		D3NOW_Pswapd_mm_mmm64,
		D3NOW_Pavgusb_mm_mmm64,
		Rmpadjust,
		Rmpupdate,
		Psmash,
		Pvalidatew,
		Pvalidated,
		Pvalidateq,
		Serialize,
		Xsusldtrk,
		Xresldtrk,
		Invlpgbw,
		Invlpgbd,
		Invlpgbq,
		Tlbsync,
		Prefetchreserved3_m8,
		Prefetchreserved4_m8,
		Prefetchreserved5_m8,
		Prefetchreserved6_m8,
		Prefetchreserved7_m8,
		Ud0,
		Vmgexit,
		Getsecq,
		VEX_Ldtilecfg_m512,
		VEX_Tilerelease,
		VEX_Sttilecfg_m512,
		VEX_Tilezero_tmm,
		VEX_Tileloaddt1_tmm_sibmem,
		VEX_Tilestored_sibmem_tmm,
		VEX_Tileloadd_tmm_sibmem,
		VEX_Tdpbf16ps_tmm_tmm_tmm,
		VEX_Tdpbuud_tmm_tmm_tmm,
		VEX_Tdpbusd_tmm_tmm_tmm,
		VEX_Tdpbsud_tmm_tmm_tmm,
		VEX_Tdpbssd_tmm_tmm_tmm,
		Fnstdw_AX,
		Fnstsg_AX,
		Rdshr_rm32,
		Wrshr_rm32,
		Smint,
		Dmint,
		Rdm,
		Svdc_m80_Sreg,
		Rsdc_Sreg_m80,
		Svldt_m80,
		Rsldt_m80,
		Svts_m80,
		Rsts_m80,
		Smint_0F7E,
		Bb0_reset,
		Bb1_reset,
		Cpu_write,
		Cpu_read,
		Altinst,
		Paveb_mm_mmm64,
		Paddsiw_mm_mmm64,
		Pmagw_mm_mmm64,
		Pdistib_mm_m64,
		Psubsiw_mm_mmm64,
		Pmvzb_mm_m64,
		Pmulhrw_mm_mmm64,
		Pmvnzb_mm_m64,
		Pmvlzb_mm_m64,
		Pmvgezb_mm_m64,
		Pmulhriw_mm_mmm64,
		Pmachriw_mm_m64,
		Cyrix_D9D7,
		Cyrix_D9E2,
		Ftstp,
		Cyrix_D9E7,
		Frint2,
		Frichop,
		Cyrix_DED8,
		Cyrix_DEDA,
		Cyrix_DEDC,
		Cyrix_DEDD,
		Cyrix_DEDE,
		Frinear,
		Tdcall,
		Seamret,
		Seamops,
		Seamcall,
		Aesencwide128kl_m384,
		Aesdecwide128kl_m384,
		Aesencwide256kl_m512,
		Aesdecwide256kl_m512,
		Loadiwkey_xmm_xmm,
		Aesenc128kl_xmm_m384,
		Aesdec128kl_xmm_m384,
		Aesenc256kl_xmm_m512,
		Aesdec256kl_xmm_m512,
		Encodekey128_r32_r32,
		Encodekey256_r32_r32,
		VEX_Vbroadcastss_xmm_xmm,
		VEX_Vbroadcastss_ymm_xmm,
		VEX_Vbroadcastsd_ymm_xmm,
		Vmgexit_F2,
		Uiret,
		Testui,
		Clui,
		Stui,
		Senduipi_r64,
		Hreset_imm8,
		VEX_Vpdpbusd_xmm_xmm_xmmm128,
		VEX_Vpdpbusd_ymm_ymm_ymmm256,
		VEX_Vpdpbusds_xmm_xmm_xmmm128,
		VEX_Vpdpbusds_ymm_ymm_ymmm256,
		VEX_Vpdpwssd_xmm_xmm_xmmm128,
		VEX_Vpdpwssd_ymm_ymm_ymmm256,
		VEX_Vpdpwssds_xmm_xmm_xmmm128,
		VEX_Vpdpwssds_ymm_ymm_ymmm256,
		Ccs_hash_16,
		Ccs_hash_32,
		Ccs_hash_64,
		Ccs_encrypt_16,
		Ccs_encrypt_32,
		Ccs_encrypt_64,
		Lkgs_rm16,
		Lkgs_r32m16,
		Lkgs_r64m16,
		Eretu,
		Erets,
		EVEX_Vaddph_xmm_k1z_xmm_xmmm128b16,
		EVEX_Vaddph_ymm_k1z_ymm_ymmm256b16,
		EVEX_Vaddph_zmm_k1z_zmm_zmmm512b16_er,
		EVEX_Vaddsh_xmm_k1z_xmm_xmmm16_er,
		EVEX_Vcmpph_kr_k1_xmm_xmmm128b16_imm8,
		EVEX_Vcmpph_kr_k1_ymm_ymmm256b16_imm8,
		EVEX_Vcmpph_kr_k1_zmm_zmmm512b16_imm8_sae,
		EVEX_Vcmpsh_kr_k1_xmm_xmmm16_imm8_sae,
		EVEX_Vcomish_xmm_xmmm16_sae,
		EVEX_Vcvtdq2ph_xmm_k1z_xmmm128b32,
		EVEX_Vcvtdq2ph_xmm_k1z_ymmm256b32,
		EVEX_Vcvtdq2ph_ymm_k1z_zmmm512b32_er,
		EVEX_Vcvtpd2ph_xmm_k1z_xmmm128b64,
		EVEX_Vcvtpd2ph_xmm_k1z_ymmm256b64,
		EVEX_Vcvtpd2ph_xmm_k1z_zmmm512b64_er,
		EVEX_Vcvtph2dq_xmm_k1z_xmmm64b16,
		EVEX_Vcvtph2dq_ymm_k1z_xmmm128b16,
		EVEX_Vcvtph2dq_zmm_k1z_ymmm256b16_er,
		EVEX_Vcvtph2pd_xmm_k1z_xmmm32b16,
		EVEX_Vcvtph2pd_ymm_k1z_xmmm64b16,
		EVEX_Vcvtph2pd_zmm_k1z_xmmm128b16_sae,
		EVEX_Vcvtph2psx_xmm_k1z_xmmm64b16,
		EVEX_Vcvtph2psx_ymm_k1z_xmmm128b16,
		EVEX_Vcvtph2psx_zmm_k1z_ymmm256b16_sae,
		EVEX_Vcvtph2qq_xmm_k1z_xmmm32b16,
		EVEX_Vcvtph2qq_ymm_k1z_xmmm64b16,
		EVEX_Vcvtph2qq_zmm_k1z_xmmm128b16_er,
		EVEX_Vcvtph2udq_xmm_k1z_xmmm64b16,
		EVEX_Vcvtph2udq_ymm_k1z_xmmm128b16,
		EVEX_Vcvtph2udq_zmm_k1z_ymmm256b16_er,
		EVEX_Vcvtph2uqq_xmm_k1z_xmmm32b16,
		EVEX_Vcvtph2uqq_ymm_k1z_xmmm64b16,
		EVEX_Vcvtph2uqq_zmm_k1z_xmmm128b16_er,
		EVEX_Vcvtph2uw_xmm_k1z_xmmm128b16,
		EVEX_Vcvtph2uw_ymm_k1z_ymmm256b16,
		EVEX_Vcvtph2uw_zmm_k1z_zmmm512b16_er,
		EVEX_Vcvtph2w_xmm_k1z_xmmm128b16,
		EVEX_Vcvtph2w_ymm_k1z_ymmm256b16,
		EVEX_Vcvtph2w_zmm_k1z_zmmm512b16_er,
		EVEX_Vcvtps2phx_xmm_k1z_xmmm128b32,
		EVEX_Vcvtps2phx_xmm_k1z_ymmm256b32,
		EVEX_Vcvtps2phx_ymm_k1z_zmmm512b32_er,
		EVEX_Vcvtqq2ph_xmm_k1z_xmmm128b64,
		EVEX_Vcvtqq2ph_xmm_k1z_ymmm256b64,
		EVEX_Vcvtqq2ph_xmm_k1z_zmmm512b64_er,
		EVEX_Vcvtsd2sh_xmm_k1z_xmm_xmmm64_er,
		EVEX_Vcvtsh2sd_xmm_k1z_xmm_xmmm16_sae,
		EVEX_Vcvtsh2si_r32_xmmm16_er,
		EVEX_Vcvtsh2si_r64_xmmm16_er,
		EVEX_Vcvtsh2ss_xmm_k1z_xmm_xmmm16_sae,
		EVEX_Vcvtsh2usi_r32_xmmm16_er,
		EVEX_Vcvtsh2usi_r64_xmmm16_er,
		EVEX_Vcvtsi2sh_xmm_xmm_rm32_er,
		EVEX_Vcvtsi2sh_xmm_xmm_rm64_er,
		EVEX_Vcvtss2sh_xmm_k1z_xmm_xmmm32_er,
		EVEX_Vcvttph2dq_xmm_k1z_xmmm64b16,
		EVEX_Vcvttph2dq_ymm_k1z_xmmm128b16,
		EVEX_Vcvttph2dq_zmm_k1z_ymmm256b16_sae,
		EVEX_Vcvttph2qq_xmm_k1z_xmmm32b16,
		EVEX_Vcvttph2qq_ymm_k1z_xmmm64b16,
		EVEX_Vcvttph2qq_zmm_k1z_xmmm128b16_sae,
		EVEX_Vcvttph2udq_xmm_k1z_xmmm64b16,
		EVEX_Vcvttph2udq_ymm_k1z_xmmm128b16,
		EVEX_Vcvttph2udq_zmm_k1z_ymmm256b16_sae,
		EVEX_Vcvttph2uqq_xmm_k1z_xmmm32b16,
		EVEX_Vcvttph2uqq_ymm_k1z_xmmm64b16,
		EVEX_Vcvttph2uqq_zmm_k1z_xmmm128b16_sae,
		EVEX_Vcvttph2uw_xmm_k1z_xmmm128b16,
		EVEX_Vcvttph2uw_ymm_k1z_ymmm256b16,
		EVEX_Vcvttph2uw_zmm_k1z_zmmm512b16_sae,
		EVEX_Vcvttph2w_xmm_k1z_xmmm128b16,
		EVEX_Vcvttph2w_ymm_k1z_ymmm256b16,
		EVEX_Vcvttph2w_zmm_k1z_zmmm512b16_sae,
		EVEX_Vcvttsh2si_r32_xmmm16_sae,
		EVEX_Vcvttsh2si_r64_xmmm16_sae,
		EVEX_Vcvttsh2usi_r32_xmmm16_sae,
		EVEX_Vcvttsh2usi_r64_xmmm16_sae,
		EVEX_Vcvtudq2ph_xmm_k1z_xmmm128b32,
		EVEX_Vcvtudq2ph_xmm_k1z_ymmm256b32,
		EVEX_Vcvtudq2ph_ymm_k1z_zmmm512b32_er,
		EVEX_Vcvtuqq2ph_xmm_k1z_xmmm128b64,
		EVEX_Vcvtuqq2ph_xmm_k1z_ymmm256b64,
		EVEX_Vcvtuqq2ph_xmm_k1z_zmmm512b64_er,
		EVEX_Vcvtusi2sh_xmm_xmm_rm32_er,
		EVEX_Vcvtusi2sh_xmm_xmm_rm64_er,
		EVEX_Vcvtuw2ph_xmm_k1z_xmmm128b16,
		EVEX_Vcvtuw2ph_ymm_k1z_ymmm256b16,
		EVEX_Vcvtuw2ph_zmm_k1z_zmmm512b16_er,
		EVEX_Vcvtw2ph_xmm_k1z_xmmm128b16,
		EVEX_Vcvtw2ph_ymm_k1z_ymmm256b16,
		EVEX_Vcvtw2ph_zmm_k1z_zmmm512b16_er,
		EVEX_Vdivph_xmm_k1z_xmm_xmmm128b16,
		EVEX_Vdivph_ymm_k1z_ymm_ymmm256b16,
		EVEX_Vdivph_zmm_k1z_zmm_zmmm512b16_er,
		EVEX_Vdivsh_xmm_k1z_xmm_xmmm16_er,
		EVEX_Vfcmaddcph_xmm_k1z_xmm_xmmm128b32,
		EVEX_Vfcmaddcph_ymm_k1z_ymm_ymmm256b32,
		EVEX_Vfcmaddcph_zmm_k1z_zmm_zmmm512b32_er,
		EVEX_Vfmaddcph_xmm_k1z_xmm_xmmm128b32,
		EVEX_Vfmaddcph_ymm_k1z_ymm_ymmm256b32,
		EVEX_Vfmaddcph_zmm_k1z_zmm_zmmm512b32_er,
		EVEX_Vfcmaddcsh_xmm_k1z_xmm_xmmm32_er,
		EVEX_Vfmaddcsh_xmm_k1z_xmm_xmmm32_er,
		EVEX_Vfcmulcph_xmm_k1z_xmm_xmmm128b32,
		EVEX_Vfcmulcph_ymm_k1z_ymm_ymmm256b32,
		EVEX_Vfcmulcph_zmm_k1z_zmm_zmmm512b32_er,
		EVEX_Vfmulcph_xmm_k1z_xmm_xmmm128b32,
		EVEX_Vfmulcph_ymm_k1z_ymm_ymmm256b32,
		EVEX_Vfmulcph_zmm_k1z_zmm_zmmm512b32_er,
		EVEX_Vfcmulcsh_xmm_k1z_xmm_xmmm32_er,
		EVEX_Vfmulcsh_xmm_k1z_xmm_xmmm32_er,
		EVEX_Vfmaddsub132ph_xmm_k1z_xmm_xmmm128b16,
		EVEX_Vfmaddsub132ph_ymm_k1z_ymm_ymmm256b16,
		EVEX_Vfmaddsub132ph_zmm_k1z_zmm_zmmm512b16_er,
		EVEX_Vfmaddsub213ph_xmm_k1z_xmm_xmmm128b16,
		EVEX_Vfmaddsub213ph_ymm_k1z_ymm_ymmm256b16,
		EVEX_Vfmaddsub213ph_zmm_k1z_zmm_zmmm512b16_er,
		EVEX_Vfmaddsub231ph_xmm_k1z_xmm_xmmm128b16,
		EVEX_Vfmaddsub231ph_ymm_k1z_ymm_ymmm256b16,
		EVEX_Vfmaddsub231ph_zmm_k1z_zmm_zmmm512b16_er,
		EVEX_Vfmsubadd132ph_xmm_k1z_xmm_xmmm128b16,
		EVEX_Vfmsubadd132ph_ymm_k1z_ymm_ymmm256b16,
		EVEX_Vfmsubadd132ph_zmm_k1z_zmm_zmmm512b16_er,
		EVEX_Vfmsubadd213ph_xmm_k1z_xmm_xmmm128b16,
		EVEX_Vfmsubadd213ph_ymm_k1z_ymm_ymmm256b16,
		EVEX_Vfmsubadd213ph_zmm_k1z_zmm_zmmm512b16_er,
		EVEX_Vfmsubadd231ph_xmm_k1z_xmm_xmmm128b16,
		EVEX_Vfmsubadd231ph_ymm_k1z_ymm_ymmm256b16,
		EVEX_Vfmsubadd231ph_zmm_k1z_zmm_zmmm512b16_er,
		EVEX_Vfmadd132ph_xmm_k1z_xmm_xmmm128b16,
		EVEX_Vfmadd132ph_ymm_k1z_ymm_ymmm256b16,
		EVEX_Vfmadd132ph_zmm_k1z_zmm_zmmm512b16_er,
		EVEX_Vfmadd213ph_xmm_k1z_xmm_xmmm128b16,
		EVEX_Vfmadd213ph_ymm_k1z_ymm_ymmm256b16,
		EVEX_Vfmadd213ph_zmm_k1z_zmm_zmmm512b16_er,
		EVEX_Vfmadd231ph_xmm_k1z_xmm_xmmm128b16,
		EVEX_Vfmadd231ph_ymm_k1z_ymm_ymmm256b16,
		EVEX_Vfmadd231ph_zmm_k1z_zmm_zmmm512b16_er,
		EVEX_Vfnmadd132ph_xmm_k1z_xmm_xmmm128b16,
		EVEX_Vfnmadd132ph_ymm_k1z_ymm_ymmm256b16,
		EVEX_Vfnmadd132ph_zmm_k1z_zmm_zmmm512b16_er,
		EVEX_Vfnmadd213ph_xmm_k1z_xmm_xmmm128b16,
		EVEX_Vfnmadd213ph_ymm_k1z_ymm_ymmm256b16,
		EVEX_Vfnmadd213ph_zmm_k1z_zmm_zmmm512b16_er,
		EVEX_Vfnmadd231ph_xmm_k1z_xmm_xmmm128b16,
		EVEX_Vfnmadd231ph_ymm_k1z_ymm_ymmm256b16,
		EVEX_Vfnmadd231ph_zmm_k1z_zmm_zmmm512b16_er,
		EVEX_Vfmadd132sh_xmm_k1z_xmm_xmmm16_er,
		EVEX_Vfmadd213sh_xmm_k1z_xmm_xmmm16_er,
		EVEX_Vfmadd231sh_xmm_k1z_xmm_xmmm16_er,
		EVEX_Vfnmadd132sh_xmm_k1z_xmm_xmmm16_er,
		EVEX_Vfnmadd213sh_xmm_k1z_xmm_xmmm16_er,
		EVEX_Vfnmadd231sh_xmm_k1z_xmm_xmmm16_er,
		EVEX_Vfmsub132ph_xmm_k1z_xmm_xmmm128b16,
		EVEX_Vfmsub132ph_ymm_k1z_ymm_ymmm256b16,
		EVEX_Vfmsub132ph_zmm_k1z_zmm_zmmm512b16_er,
		EVEX_Vfmsub213ph_xmm_k1z_xmm_xmmm128b16,
		EVEX_Vfmsub213ph_ymm_k1z_ymm_ymmm256b16,
		EVEX_Vfmsub213ph_zmm_k1z_zmm_zmmm512b16_er,
		EVEX_Vfmsub231ph_xmm_k1z_xmm_xmmm128b16,
		EVEX_Vfmsub231ph_ymm_k1z_ymm_ymmm256b16,
		EVEX_Vfmsub231ph_zmm_k1z_zmm_zmmm512b16_er,
		EVEX_Vfnmsub132ph_xmm_k1z_xmm_xmmm128b16,
		EVEX_Vfnmsub132ph_ymm_k1z_ymm_ymmm256b16,
		EVEX_Vfnmsub132ph_zmm_k1z_zmm_zmmm512b16_er,
		EVEX_Vfnmsub213ph_xmm_k1z_xmm_xmmm128b16,
		EVEX_Vfnmsub213ph_ymm_k1z_ymm_ymmm256b16,
		EVEX_Vfnmsub213ph_zmm_k1z_zmm_zmmm512b16_er,
		EVEX_Vfnmsub231ph_xmm_k1z_xmm_xmmm128b16,
		EVEX_Vfnmsub231ph_ymm_k1z_ymm_ymmm256b16,
		EVEX_Vfnmsub231ph_zmm_k1z_zmm_zmmm512b16_er,
		EVEX_Vfmsub132sh_xmm_k1z_xmm_xmmm16_er,
		EVEX_Vfmsub213sh_xmm_k1z_xmm_xmmm16_er,
		EVEX_Vfmsub231sh_xmm_k1z_xmm_xmmm16_er,
		EVEX_Vfnmsub132sh_xmm_k1z_xmm_xmmm16_er,
		EVEX_Vfnmsub213sh_xmm_k1z_xmm_xmmm16_er,
		EVEX_Vfnmsub231sh_xmm_k1z_xmm_xmmm16_er,
		EVEX_Vfpclassph_kr_k1_xmmm128b16_imm8,
		EVEX_Vfpclassph_kr_k1_ymmm256b16_imm8,
		EVEX_Vfpclassph_kr_k1_zmmm512b16_imm8,
		EVEX_Vfpclasssh_kr_k1_xmmm16_imm8,
		EVEX_Vgetexpph_xmm_k1z_xmmm128b16,
		EVEX_Vgetexpph_ymm_k1z_ymmm256b16,
		EVEX_Vgetexpph_zmm_k1z_zmmm512b16_sae,
		EVEX_Vgetexpsh_xmm_k1z_xmm_xmmm16_sae,
		EVEX_Vgetmantph_xmm_k1z_xmmm128b16_imm8,
		EVEX_Vgetmantph_ymm_k1z_ymmm256b16_imm8,
		EVEX_Vgetmantph_zmm_k1z_zmmm512b16_imm8_sae,
		EVEX_Vgetmantsh_xmm_k1z_xmm_xmmm16_imm8_sae,
		EVEX_Vmaxph_xmm_k1z_xmm_xmmm128b16,
		EVEX_Vmaxph_ymm_k1z_ymm_ymmm256b16,
		EVEX_Vmaxph_zmm_k1z_zmm_zmmm512b16_sae,
		EVEX_Vmaxsh_xmm_k1z_xmm_xmmm16_sae,
		EVEX_Vminph_xmm_k1z_xmm_xmmm128b16,
		EVEX_Vminph_ymm_k1z_ymm_ymmm256b16,
		EVEX_Vminph_zmm_k1z_zmm_zmmm512b16_sae,
		EVEX_Vminsh_xmm_k1z_xmm_xmmm16_sae,
		EVEX_Vmovsh_xmm_k1z_m16,
		EVEX_Vmovsh_m16_k1_xmm,
		EVEX_Vmovsh_xmm_k1z_xmm_xmm,
		EVEX_Vmovsh_xmm_k1z_xmm_xmm_MAP5_11,
		EVEX_Vmovw_xmm_r32m16,
		EVEX_Vmovw_xmm_r64m16,
		EVEX_Vmovw_r32m16_xmm,
		EVEX_Vmovw_r64m16_xmm,
		EVEX_Vmulph_xmm_k1z_xmm_xmmm128b16,
		EVEX_Vmulph_ymm_k1z_ymm_ymmm256b16,
		EVEX_Vmulph_zmm_k1z_zmm_zmmm512b16_er,
		EVEX_Vmulsh_xmm_k1z_xmm_xmmm16_er,
		EVEX_Vrcpph_xmm_k1z_xmmm128b16,
		EVEX_Vrcpph_ymm_k1z_ymmm256b16,
		EVEX_Vrcpph_zmm_k1z_zmmm512b16,
		EVEX_Vrcpsh_xmm_k1z_xmm_xmmm16,
		EVEX_Vreduceph_xmm_k1z_xmmm128b16_imm8,
		EVEX_Vreduceph_ymm_k1z_ymmm256b16_imm8,
		EVEX_Vreduceph_zmm_k1z_zmmm512b16_imm8_sae,
		EVEX_Vreducesh_xmm_k1z_xmm_xmmm16_imm8_sae,
		EVEX_Vrndscaleph_xmm_k1z_xmmm128b16_imm8,
		EVEX_Vrndscaleph_ymm_k1z_ymmm256b16_imm8,
		EVEX_Vrndscaleph_zmm_k1z_zmmm512b16_imm8_sae,
		EVEX_Vrndscalesh_xmm_k1z_xmm_xmmm16_imm8_sae,
		EVEX_Vrsqrtph_xmm_k1z_xmmm128b16,
		EVEX_Vrsqrtph_ymm_k1z_ymmm256b16,
		EVEX_Vrsqrtph_zmm_k1z_zmmm512b16,
		EVEX_Vrsqrtsh_xmm_k1z_xmm_xmmm16,
		EVEX_Vscalefph_xmm_k1z_xmm_xmmm128b16,
		EVEX_Vscalefph_ymm_k1z_ymm_ymmm256b16,
		EVEX_Vscalefph_zmm_k1z_zmm_zmmm512b16_er,
		EVEX_Vscalefsh_xmm_k1z_xmm_xmmm16_er,
		EVEX_Vsqrtph_xmm_k1z_xmmm128b16,
		EVEX_Vsqrtph_ymm_k1z_ymmm256b16,
		EVEX_Vsqrtph_zmm_k1z_zmmm512b16_er,
		EVEX_Vsqrtsh_xmm_k1z_xmm_xmmm16_er,
		EVEX_Vsubph_xmm_k1z_xmm_xmmm128b16,
		EVEX_Vsubph_ymm_k1z_ymm_ymmm256b16,
		EVEX_Vsubph_zmm_k1z_zmm_zmmm512b16_er,
		EVEX_Vsubsh_xmm_k1z_xmm_xmmm16_er,
		EVEX_Vucomish_xmm_xmmm16_sae,
		Rdudbg,
		Wrudbg,
		VEX_KNC_Jkzd_kr_rel8_64,
		VEX_KNC_Jknzd_kr_rel8_64,
		VEX_KNC_Vprefetchnta_m8,
		VEX_KNC_Vprefetch0_m8,
		VEX_KNC_Vprefetch1_m8,
		VEX_KNC_Vprefetch2_m8,
		VEX_KNC_Vprefetchenta_m8,
		VEX_KNC_Vprefetche0_m8,
		VEX_KNC_Vprefetche1_m8,
		VEX_KNC_Vprefetche2_m8,
		VEX_KNC_Kand_kr_kr,
		VEX_KNC_Kandn_kr_kr,
		VEX_KNC_Kandnr_kr_kr,
		VEX_KNC_Knot_kr_kr,
		VEX_KNC_Kor_kr_kr,
		VEX_KNC_Kxnor_kr_kr,
		VEX_KNC_Kxor_kr_kr,
		VEX_KNC_Kmerge2l1h_kr_kr,
		VEX_KNC_Kmerge2l1l_kr_kr,
		VEX_KNC_Jkzd_kr_rel32_64,
		VEX_KNC_Jknzd_kr_rel32_64,
		VEX_KNC_Kmov_kr_kr,
		VEX_KNC_Kmov_kr_r32,
		VEX_KNC_Kmov_r32_kr,
		VEX_KNC_Kconcath_r64_kr_kr,
		VEX_KNC_Kconcatl_r64_kr_kr,
		VEX_KNC_Kortest_kr_kr,
		VEX_KNC_Delay_r32,
		VEX_KNC_Delay_r64,
		VEX_KNC_Spflt_r32,
		VEX_KNC_Spflt_r64,
		VEX_KNC_Clevict1_m8,
		VEX_KNC_Clevict0_m8,
		VEX_KNC_Popcnt_r32_r32,
		VEX_KNC_Popcnt_r64_r64,
		VEX_KNC_Tzcnt_r32_r32,
		VEX_KNC_Tzcnt_r64_r64,
		VEX_KNC_Tzcnti_r32_r32,
		VEX_KNC_Tzcnti_r64_r64,
		VEX_KNC_Lzcnt_r32_r32,
		VEX_KNC_Lzcnt_r64_r64,
		VEX_KNC_Undoc_r32_rm32_128_F3_0F38_W0_F0,
		VEX_KNC_Undoc_r64_rm64_128_F3_0F38_W1_F0,
		VEX_KNC_Undoc_r32_rm32_128_F2_0F38_W0_F0,
		VEX_KNC_Undoc_r64_rm64_128_F2_0F38_W1_F0,
		VEX_KNC_Undoc_r32_rm32_128_F2_0F38_W0_F1,
		VEX_KNC_Undoc_r64_rm64_128_F2_0F38_W1_F1,
		VEX_KNC_Kextract_kr_r64_imm8,
		MVEX_Vprefetchnta_m,
		MVEX_Vprefetch0_m,
		MVEX_Vprefetch1_m,
		MVEX_Vprefetch2_m,
		MVEX_Vprefetchenta_m,
		MVEX_Vprefetche0_m,
		MVEX_Vprefetche1_m,
		MVEX_Vprefetche2_m,
		MVEX_Vmovaps_zmm_k1_zmmmt,
		MVEX_Vmovapd_zmm_k1_zmmmt,
		MVEX_Vmovaps_mt_k1_zmm,
		MVEX_Vmovapd_mt_k1_zmm,
		MVEX_Vmovnrapd_m_k1_zmm,
		MVEX_Vmovnrngoapd_m_k1_zmm,
		MVEX_Vmovnraps_m_k1_zmm,
		MVEX_Vmovnrngoaps_m_k1_zmm,
		MVEX_Vaddps_zmm_k1_zmm_zmmmt,
		MVEX_Vaddpd_zmm_k1_zmm_zmmmt,
		MVEX_Vmulps_zmm_k1_zmm_zmmmt,
		MVEX_Vmulpd_zmm_k1_zmm_zmmmt,
		MVEX_Vcvtps2pd_zmm_k1_zmmmt,
		MVEX_Vcvtpd2ps_zmm_k1_zmmmt,
		MVEX_Vsubps_zmm_k1_zmm_zmmmt,
		MVEX_Vsubpd_zmm_k1_zmm_zmmmt,
		MVEX_Vpcmpgtd_kr_k1_zmm_zmmmt,
		MVEX_Vmovdqa32_zmm_k1_zmmmt,
		MVEX_Vmovdqa64_zmm_k1_zmmmt,
		MVEX_Vpshufd_zmm_k1_zmmmt_imm8,
		MVEX_Vpsrld_zmm_k1_zmmmt_imm8,
		MVEX_Vpsrad_zmm_k1_zmmmt_imm8,
		MVEX_Vpslld_zmm_k1_zmmmt_imm8,
		MVEX_Vpcmpeqd_kr_k1_zmm_zmmmt,
		MVEX_Vcvtudq2pd_zmm_k1_zmmmt,
		MVEX_Vmovdqa32_mt_k1_zmm,
		MVEX_Vmovdqa64_mt_k1_zmm,
		MVEX_Clevict1_m,
		MVEX_Clevict0_m,
		MVEX_Vcmpps_kr_k1_zmm_zmmmt_imm8,
		MVEX_Vcmppd_kr_k1_zmm_zmmmt_imm8,
		MVEX_Vpandd_zmm_k1_zmm_zmmmt,
		MVEX_Vpandq_zmm_k1_zmm_zmmmt,
		MVEX_Vpandnd_zmm_k1_zmm_zmmmt,
		MVEX_Vpandnq_zmm_k1_zmm_zmmmt,
		MVEX_Vcvtdq2pd_zmm_k1_zmmmt,
		MVEX_Vpord_zmm_k1_zmm_zmmmt,
		MVEX_Vporq_zmm_k1_zmm_zmmmt,
		MVEX_Vpxord_zmm_k1_zmm_zmmmt,
		MVEX_Vpxorq_zmm_k1_zmm_zmmmt,
		MVEX_Vpsubd_zmm_k1_zmm_zmmmt,
		MVEX_Vpaddd_zmm_k1_zmm_zmmmt,
		MVEX_Vbroadcastss_zmm_k1_mt,
		MVEX_Vbroadcastsd_zmm_k1_mt,
		MVEX_Vbroadcastf32x4_zmm_k1_mt,
		MVEX_Vbroadcastf64x4_zmm_k1_mt,
		MVEX_Vptestmd_kr_k1_zmm_zmmmt,
		MVEX_Vpermd_zmm_k1_zmm_zmmmt,
		MVEX_Vpminsd_zmm_k1_zmm_zmmmt,
		MVEX_Vpminud_zmm_k1_zmm_zmmmt,
		MVEX_Vpmaxsd_zmm_k1_zmm_zmmmt,
		MVEX_Vpmaxud_zmm_k1_zmm_zmmmt,
		MVEX_Vpmulld_zmm_k1_zmm_zmmmt,
		MVEX_Vgetexpps_zmm_k1_zmmmt,
		MVEX_Vgetexppd_zmm_k1_zmmmt,
		MVEX_Vpsrlvd_zmm_k1_zmm_zmmmt,
		MVEX_Vpsravd_zmm_k1_zmm_zmmmt,
		MVEX_Vpsllvd_zmm_k1_zmm_zmmmt,
		MVEX_Undoc_zmm_k1_zmmmt_512_66_0F38_W0_48,
		MVEX_Undoc_zmm_k1_zmmmt_512_66_0F38_W0_49,
		MVEX_Undoc_zmm_k1_zmmmt_512_66_0F38_W0_4A,
		MVEX_Undoc_zmm_k1_zmmmt_512_66_0F38_W0_4B,
		MVEX_Vaddnps_zmm_k1_zmm_zmmmt,
		MVEX_Vaddnpd_zmm_k1_zmm_zmmmt,
		MVEX_Vgmaxabsps_zmm_k1_zmm_zmmmt,
		MVEX_Vgminps_zmm_k1_zmm_zmmmt,
		MVEX_Vgminpd_zmm_k1_zmm_zmmmt,
		MVEX_Vgmaxps_zmm_k1_zmm_zmmmt,
		MVEX_Vgmaxpd_zmm_k1_zmm_zmmmt,
		MVEX_Undoc_zmm_k1_zmm_zmmmt_512_66_0F38_W0_54,
		MVEX_Vfixupnanps_zmm_k1_zmm_zmmmt,
		MVEX_Vfixupnanpd_zmm_k1_zmm_zmmmt,
		MVEX_Undoc_zmm_k1_zmm_zmmmt_512_66_0F38_W0_56,
		MVEX_Undoc_zmm_k1_zmm_zmmmt_512_66_0F38_W0_57,
		MVEX_Vpbroadcastd_zmm_k1_mt,
		MVEX_Vpbroadcastq_zmm_k1_mt,
		MVEX_Vbroadcasti32x4_zmm_k1_mt,
		MVEX_Vbroadcasti64x4_zmm_k1_mt,
		MVEX_Vpadcd_zmm_k1_kr_zmmmt,
		MVEX_Vpaddsetcd_zmm_k1_kr_zmmmt,
		MVEX_Vpsbbd_zmm_k1_kr_zmmmt,
		MVEX_Vpsubsetbd_zmm_k1_kr_zmmmt,
		MVEX_Vpblendmd_zmm_k1_zmm_zmmmt,
		MVEX_Vpblendmq_zmm_k1_zmm_zmmmt,
		MVEX_Vblendmps_zmm_k1_zmm_zmmmt,
		MVEX_Vblendmpd_zmm_k1_zmm_zmmmt,
		MVEX_Undoc_zmm_k1_zmm_zmmmt_512_66_0F38_W0_67,
		MVEX_Undoc_zmm_k1_zmmmt_512_66_0F38_W0_68,
		MVEX_Undoc_zmm_k1_zmmmt_512_66_0F38_W0_69,
		MVEX_Undoc_zmm_k1_zmmmt_512_66_0F38_W0_6A,
		MVEX_Undoc_zmm_k1_zmmmt_512_66_0F38_W0_6B,
		MVEX_Vpsubrd_zmm_k1_zmm_zmmmt,
		MVEX_Vsubrps_zmm_k1_zmm_zmmmt,
		MVEX_Vsubrpd_zmm_k1_zmm_zmmmt,
		MVEX_Vpsbbrd_zmm_k1_kr_zmmmt,
		MVEX_Vpsubrsetbd_zmm_k1_kr_zmmmt,
		MVEX_Undoc_zmm_k1_zmm_zmmmt_512_66_0F38_W0_70,
		MVEX_Undoc_zmm_k1_zmm_zmmmt_512_66_0F38_W0_71,
		MVEX_Undoc_zmm_k1_zmm_zmmmt_512_66_0F38_W0_72,
		MVEX_Undoc_zmm_k1_zmm_zmmmt_512_66_0F38_W0_73,
		MVEX_Vpcmpltd_kr_k1_zmm_zmmmt,
		MVEX_Vscaleps_zmm_k1_zmm_zmmmt,
		MVEX_Vpmulhud_zmm_k1_zmm_zmmmt,
		MVEX_Vpmulhd_zmm_k1_zmm_zmmmt,
		MVEX_Vpgatherdd_zmm_k1_mvt,
		MVEX_Vpgatherdq_zmm_k1_mvt,
		MVEX_Vgatherdps_zmm_k1_mvt,
		MVEX_Vgatherdpd_zmm_k1_mvt,
		MVEX_Undoc_zmm_k1_zmm_zmmmt_512_66_0F38_W0_94,
		MVEX_Undoc_zmm_k1_zmm_zmmmt_512_66_0F38_W1_94,
		MVEX_Vfmadd132ps_zmm_k1_zmm_zmmmt,
		MVEX_Vfmadd132pd_zmm_k1_zmm_zmmmt,
		MVEX_Vfmsub132ps_zmm_k1_zmm_zmmmt,
		MVEX_Vfmsub132pd_zmm_k1_zmm_zmmmt,
		MVEX_Vfnmadd132ps_zmm_k1_zmm_zmmmt,
		MVEX_Vfnmadd132pd_zmm_k1_zmm_zmmmt,
		MVEX_Vfnmsub132ps_zmm_k1_zmm_zmmmt,
		MVEX_Vfnmsub132pd_zmm_k1_zmm_zmmmt,
		MVEX_Vpscatterdd_mvt_k1_zmm,
		MVEX_Vpscatterdq_mvt_k1_zmm,
		MVEX_Vscatterdps_mvt_k1_zmm,
		MVEX_Vscatterdpd_mvt_k1_zmm,
		MVEX_Vfmadd233ps_zmm_k1_zmm_zmmmt,
		MVEX_Vfmadd213ps_zmm_k1_zmm_zmmmt,
		MVEX_Vfmadd213pd_zmm_k1_zmm_zmmmt,
		MVEX_Vfmsub213ps_zmm_k1_zmm_zmmmt,
		MVEX_Vfmsub213pd_zmm_k1_zmm_zmmmt,
		MVEX_Vfnmadd213ps_zmm_k1_zmm_zmmmt,
		MVEX_Vfnmadd213pd_zmm_k1_zmm_zmmmt,
		MVEX_Vfnmsub213ps_zmm_k1_zmm_zmmmt,
		MVEX_Vfnmsub213pd_zmm_k1_zmm_zmmmt,
		MVEX_Undoc_zmm_k1_mvt_512_66_0F38_W0_B0,
		MVEX_Undoc_zmm_k1_mvt_512_66_0F38_W0_B2,
		MVEX_Vpmadd233d_zmm_k1_zmm_zmmmt,
		MVEX_Vpmadd231d_zmm_k1_zmm_zmmmt,
		MVEX_Vfmadd231ps_zmm_k1_zmm_zmmmt,
		MVEX_Vfmadd231pd_zmm_k1_zmm_zmmmt,
		MVEX_Vfmsub231ps_zmm_k1_zmm_zmmmt,
		MVEX_Vfmsub231pd_zmm_k1_zmm_zmmmt,
		MVEX_Vfnmadd231ps_zmm_k1_zmm_zmmmt,
		MVEX_Vfnmadd231pd_zmm_k1_zmm_zmmmt,
		MVEX_Vfnmsub231ps_zmm_k1_zmm_zmmmt,
		MVEX_Vfnmsub231pd_zmm_k1_zmm_zmmmt,
		MVEX_Undoc_zmm_k1_mvt_512_66_0F38_W0_C0,
		MVEX_Vgatherpf0hintdps_mvt_k1,
		MVEX_Vgatherpf0hintdpd_mvt_k1,
		MVEX_Vgatherpf0dps_mvt_k1,
		MVEX_Vgatherpf1dps_mvt_k1,
		MVEX_Vscatterpf0hintdps_mvt_k1,
		MVEX_Vscatterpf0hintdpd_mvt_k1,
		MVEX_Vscatterpf0dps_mvt_k1,
		MVEX_Vscatterpf1dps_mvt_k1,
		MVEX_Vexp223ps_zmm_k1_zmmmt,
		MVEX_Vlog2ps_zmm_k1_zmmmt,
		MVEX_Vrcp23ps_zmm_k1_zmmmt,
		MVEX_Vrsqrt23ps_zmm_k1_zmmmt,
		MVEX_Vaddsetsps_zmm_k1_zmm_zmmmt,
		MVEX_Vpaddsetsd_zmm_k1_zmm_zmmmt,
		MVEX_Undoc_zmm_k1_zmm_zmmmt_512_66_0F38_W0_CE,
		MVEX_Undoc_zmm_k1_zmm_zmmmt_512_66_0F38_W1_CE,
		MVEX_Undoc_zmm_k1_zmm_zmmmt_512_66_0F38_W0_CF,
		MVEX_Vloadunpackld_zmm_k1_mt,
		MVEX_Vloadunpacklq_zmm_k1_mt,
		MVEX_Vpackstoreld_mt_k1_zmm,
		MVEX_Vpackstorelq_mt_k1_zmm,
		MVEX_Vloadunpacklps_zmm_k1_mt,
		MVEX_Vloadunpacklpd_zmm_k1_mt,
		MVEX_Vpackstorelps_mt_k1_zmm,
		MVEX_Vpackstorelpd_mt_k1_zmm,
		MVEX_Undoc_zmm_k1_zmmmt_512_0F38_W0_D2,
		MVEX_Undoc_zmm_k1_zmmmt_512_66_0F38_W0_D2,
		MVEX_Undoc_zmm_k1_zmmmt_512_0F38_W0_D3,
		MVEX_Vloadunpackhd_zmm_k1_mt,
		MVEX_Vloadunpackhq_zmm_k1_mt,
		MVEX_Vpackstorehd_mt_k1_zmm,
		MVEX_Vpackstorehq_mt_k1_zmm,
		MVEX_Vloadunpackhps_zmm_k1_mt,
		MVEX_Vloadunpackhpd_zmm_k1_mt,
		MVEX_Vpackstorehps_mt_k1_zmm,
		MVEX_Vpackstorehpd_mt_k1_zmm,
		MVEX_Undoc_zmm_k1_zmmmt_512_0F38_W0_D6,
		MVEX_Undoc_zmm_k1_zmmmt_512_66_0F38_W0_D6,
		MVEX_Undoc_zmm_k1_zmmmt_512_0F38_W0_D7,
		MVEX_Valignd_zmm_k1_zmm_zmmmt_imm8,
		MVEX_Vpermf32x4_zmm_k1_zmmmt_imm8,
		MVEX_Vpcmpud_kr_k1_zmm_zmmmt_imm8,
		MVEX_Vpcmpd_kr_k1_zmm_zmmmt_imm8,
		MVEX_Vgetmantps_zmm_k1_zmmmt_imm8,
		MVEX_Vgetmantpd_zmm_k1_zmmmt_imm8,
		MVEX_Vrndfxpntps_zmm_k1_zmmmt_imm8,
		MVEX_Vrndfxpntpd_zmm_k1_zmmmt_imm8,
		MVEX_Vcvtfxpntudq2ps_zmm_k1_zmmmt_imm8,
		MVEX_Vcvtfxpntps2udq_zmm_k1_zmmmt_imm8,
		MVEX_Vcvtfxpntpd2udq_zmm_k1_zmmmt_imm8,
		MVEX_Vcvtfxpntdq2ps_zmm_k1_zmmmt_imm8,
		MVEX_Vcvtfxpntps2dq_zmm_k1_zmmmt_imm8,
		MVEX_Undoc_zmm_k1_zmmmt_imm8_512_66_0F3A_W0_D0,
		MVEX_Undoc_zmm_k1_zmmmt_imm8_512_66_0F3A_W0_D1,
		MVEX_Vcvtfxpntpd2dq_zmm_k1_zmmmt_imm8,
		Via_undoc_F30FA6F0_16,
		Via_undoc_F30FA6F0_32,
		Via_undoc_F30FA6F0_64,
		Via_undoc_F30FA6F8_16,
		Via_undoc_F30FA6F8_32,
		Via_undoc_F30FA6F8_64,
		Xsha512_16,
		Xsha512_32,
		Xsha512_64,
		Xstore_alt_16,
		Xstore_alt_32,
		Xstore_alt_64,
		Xsha512_alt_16,
		Xsha512_alt_32,
		Xsha512_alt_64,
		[Comment("A zero-sized instruction. Can be used as a label.")]
		Zero_bytes,
	}

	[TypeGen(TypeGenOrders.CreatedInstructions)]
	sealed class CodeEnum {
		CodeEnum(GenTypes genTypes) {
			// It must have value 0
			if (genTypes[TypeIds.Code][nameof(Code.INVALID)].Value != 0)
				throw new InvalidOperationException();
		}
	}
}
